/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NLMOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NLM_2#**********************************/

//#define DEBUG 

#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <malloc.h>

#include <nlm_hal_fmn.h>
#include <nlm_hal_macros.h>
#include <nlm_hal_sae.h>
#include <nlm_crypto.h>
#include <nlm_crypto_api.h>
#include "nlm_crypto_data.h"

#include "mm.h"

/**
* @file_name crypto.c
*/
/**
* @defgroup crypto Crypto API
* @brief Description about the crypto apis
*/

#define printf(a, b...) printf(a, ##b)
#define shift_lower_bits(x, bitshift, numofbits) (((unsigned long long)(x) & ((1ULL << (numofbits)) - 1)) << (bitshift))
#define xtract_bits(x, bitpos, numofbits) ((x) >> (bitpos) & ((1ULL << (numofbits)) - 1))

static void *get_cache_aligned_mem(size_t size);
static void free_cache_aligned_mem(void *addr);

#define VC_MODE_ROUND_ROBIN 1
#define NUM_VC 16

/********************* H/W Counters  *******************************/

#if 0
struct crypto_hw_cntrs {
	uint64_t ingress_msg_cntr;
	uint64_t egress_msg_cntr;
	uint64_t egress_err_msg_cntr;
	uint64_t dma_rd_bytes[NUM_VC];
	uint64_t dma_wr_bytes[NUM_VC];
	uint64_t cipher_bytes[NUM_VC];
	uint64_t hash_bytes[NUM_VC];
	uint64_t op_cntr_0[NUM_VC];
	uint64_t op_cntr_1[NUM_VC];
	uint64_t op_cntr_2[NUM_VC];
	uint64_t op_cntr_3[NUM_VC];
};

static void
nlm_crypto_print_hw_cntrs(struct crypto_hw_cntrs *hw_cnt)
{
	int i;

	printf("***************** h/w counter dump start *****************n");
	printf("Ingress msg cntr: 0x%llx\n", (unsigned long long)
	       hw_cnt->ingress_msg_cntr);
	printf("Egress msg cntr: 0x%llx\n", (unsigned long long)
	       hw_cnt->egress_msg_cntr);
	printf("Egress error msg cntr: 0x%llx\n", (unsigned long long)
	       hw_cnt->egress_err_msg_cntr);

	for (i = 0; i < NUM_VC; i++) {
		printf("================ VC %d counter ===============\n", i);

		printf("DMA read bytes: 0x%llx\n", (unsigned long long)
		       hw_cnt->dma_rd_bytes[i]);
		printf("DMA write bytes: 0x%llx\n", (unsigned long long)
		       hw_cnt->dma_wr_bytes[i]);
		printf("Cipher bytes: 0x%llx\n", (unsigned long long)
		       hw_cnt->cipher_bytes[i]);
		printf("Hash bytes: 0x%llx\n", (unsigned long long)
		       hw_cnt->hash_bytes[i]);
		printf("Operation counter 0: 0x%llx\n", (unsigned long long)
		       hw_cnt->op_cntr_0[i]);
		printf("Operation counter 1: 0x%llx\n", (unsigned long long)
		       hw_cnt->op_cntr_1[i]);
		printf("Operation counter 2: 0x%llx\n", (unsigned long long)
		       hw_cnt->op_cntr_2[i]);
		printf("Operation counter 3: 0x%llx\n", (unsigned long long)
		       hw_cnt->op_cntr_3[i]);
	}
	printf("***************** h/w counter dump end *****************n");
}

static uint64_t
nlm_crypto_read_reg_lo_hi(int regnum_lo, int regnum_hi)
{
	return ((unsigned long long) read_crypto_reg(regnum_lo) |
		((unsigned long long) read_crypto_reg(regnum_hi) << 32));
}

static void
nlm_crypto_read_hw_counters(struct crypto_hw_cntrs *hw_cnt)
{
	int i;

	hw_cnt->ingress_msg_cntr =
	    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_ING_MSG_CNTR_LOW,
				      NLM_CRYPTO_REG_ING_MSG_CNTR_HIGH);

	hw_cnt->egress_msg_cntr =
	    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_EGR_MSG_CNTR_LOW,
				      NLM_CRYPTO_REG_EGR_MSG_CNTR_HIGH);

	hw_cnt->egress_err_msg_cntr =
	    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_EGR_ERR_MSG_CNTR_LOW,
				      NLM_CRYPTO_REG_EGR_ERR_MSG_CNTR_HIGH);

	for (i = 0; i < NUM_VC; i++) {
		hw_cnt->dma_rd_bytes[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_DMA_RD_LOW_BYTES
					      (i),
					      NLM_CRYPTO_REG_DMA_RD_HIGH_BYTES
					      (i));

		hw_cnt->dma_wr_bytes[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_DMA_WR_LOW_BYTES
					      (i),
					      NLM_CRYPTO_REG_DMA_WR_HIGH_BYTES
					      (i));

		hw_cnt->cipher_bytes[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_CIPHER_LOW_BYTES
					      (i),
					      NLM_CRYPTO_REG_CIPHER_HIGH_BYTES
					      (i));

		hw_cnt->hash_bytes[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_HASH_LOW_BYTES(i),
					      NLM_CRYPTO_REG_HASH_HIGH_BYTES
					      (i));

		hw_cnt->op_cntr_0[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_OP_CNTR_0_LOW_BYTES
					      (i),
					      NLM_CRYPTO_REG_OP_CNTR_0_HIGH_BYTES
					      (i));

		hw_cnt->op_cntr_1[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_OP_CNTR_1_LOW_BYTES
					      (i),
					      NLM_CRYPTO_REG_OP_CNTR_1_HIGH_BYTES
					      (i));

		hw_cnt->op_cntr_2[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_OP_CNTR_2_LOW_BYTES
					      (i),
					      NLM_CRYPTO_REG_OP_CNTR_2_HIGH_BYTES
					      (i));

		hw_cnt->op_cntr_3[i] =
		    nlm_crypto_read_reg_lo_hi(NLM_CRYPTO_REG_OP_CNTR_3_LOW_BYTES
					      (i),
					      NLM_CRYPTO_REG_OP_CNTR_3_HIGH_BYTES
					      (i));

	}
}

/********************* S/W Counters  *******************************/
uint64_t crypto_cntr_crypt_ops[NLM_CIPHER_MAX][NLM_CIPHER_MODE_MAX];
uint64_t crypto_cntr_crypt_bytes[NLM_CIPHER_MAX][NLM_CIPHER_MODE_MAX];
uint64_t crypto_cntr_auth_ops[NLM_AUTH_MAX][NLM_AUTH_MODE_MAX];
uint64_t crypto_cntr_auth_bytes[NLM_AUTH_MAX][NLM_AUTH_MODE_MAX];

static void
print_crypto_stats(void)
{
	// Take care of invalid modes...
}

static void
update_crypto_stats(int cipher_alg, int cipher_mode, int cipher_len,
		    int auth_alg, int auth_mode, int auth_len)
{
	if (cipher_alg) {
		crypto_cntr_crypt_ops[cipher_alg][cipher_mode]++;
		crypto_cntr_crypt_bytes[cipher_alg][cipher_mode] += cipher_len;
	}
	if (auth_alg) {
		crypto_cntr_auth_ops[auth_alg][auth_mode]++;
		crypto_cntr_auth_ops[auth_alg][auth_mode] += auth_len;
	}
}

static void
update_session_stats(struct nlm_crypto_op *cop)
{
	// session stats???
	update_crypto_stats(cop->session->cipher_alg,
			    cop->session->cipher_mode,
			    cop->cipher_len,
			    cop->session->auth_alg,
			    cop->session->auth_mode, cop->auth_len);
}

static void
reset_crypto_stats(void)
{
	int i, j;

	for (i = 0; i < NLM_CIPHER_MAX; i++) {
		for (j = 0; j < NLM_CIPHER_MODE_MAX; j++) {
			crypto_cntr_crypt_ops[i][j] = 0;
			crypto_cntr_crypt_bytes[i][j] = 0;
		}
	}

	for (i = 0; i < NLM_AUTH_MAX; i++) {
		for (j = 0; j < NLM_AUTH_MODE_MAX; j++) {
			crypto_cntr_auth_ops[i][j] = 0;
			crypto_cntr_auth_ops[i][j] = 0;
		}
	}
}
#endif

struct dev_data {
	void *crypto_vc;
	int crypto_fb_vc_base;
};

static struct dev_data dd;

static void
init_crypto_vc(void)
{
	dd.crypto_fb_vc_base = 3;
}

int
nlm_crypto_init(void)
{
	init_crypto_vc();

	return 0;
}

static int
crypto_get_vc(void)
{
	return NLM_CRYPTO_VC_BASE;
}

static int
crypto_get_fb_vc(void)
{
	int vc, cpu;

	vc = dd.crypto_fb_vc_base;

	cpu = processor_id();
	cpu = cpu * 4 + vc;

	return cpu;
}

/****************** Validate input, fill fields, alloc mem***************************/

//no checking for overflow
int
nlm_crypto_cipher_alg_has_mode(unsigned int cipher_alg)
{
	if (cipher_alg >= NLM_CIPHER_MAX)
		return 0;
	return cipher_alg_needs_mode[cipher_alg];
}

//no checking for overflow
int
nlm_crypto_auth_alg_has_mode(unsigned int auth_alg)
{
	if (auth_alg >= NLM_AUTH_MAX)
		return 0;
	return auth_alg_needs_mode[auth_alg];
}

static int
validate_cipher_alg_mode(struct crypto_session *session)
{
	if (session->cip) {
		if (session->cip->cipher_alg != NLM_CIPHER_BYPASS) {
			if (session->cip->cipher_alg >= NLM_CIPHER_MAX)
				return NLM_ERR_CIPHER_ALG_INVALID;

			if (nlm_crypto_cipher_alg_has_mode
			    (session->cip->cipher_alg)) {
				if ((session->cip->cipher_mode >=
				     NLM_CIPHER_MODE_MAX)
				    ||
				    !(cipher_mode_valid
				      [session->cip->cipher_alg]
				      [session->cip->cipher_mode]))

					return NLM_ERR_CIPHER_ALG_MODE_INVALID;
			}
		}
	}
	return 0;
}

static int
validate_auth_alg_mode(struct crypto_session *session)
{
	if (session->aip) {
		if (session->aip->auth_alg != NLM_AUTH_BYPASS) {
			if (session->aip->auth_alg >= NLM_AUTH_MAX)
				return NLM_ERR_AUTH_ALG_INVALID;
			if (nlm_crypto_auth_alg_has_mode
			    (session->aip->auth_alg)) {
				if ((session->aip->auth_mode >=
				     NLM_AUTH_MODE_MAX)
				    || !(auth_mode_valid[session->aip->auth_alg]
					 [session->aip->auth_mode]))

					return NLM_ERR_AUTH_ALG_MODE_INVALID;
			}
		}
	}
	return 0;
}

static int
validate_alg_mode_key_size(struct crypto_session *session)
{
	int ret = 0;

	if (session->cip) {
		ret = validate_cipher_alg_mode(session);
		if (!ret) {
			if (session->cip->cipher_key->iov_len !=
			    cipher_mode_key_len[session->cip->
						cipher_alg][session->
							    cip->cipher_mode])
				return NLM_ERR_CIPHER_KEY_LEN_INVALID;
		}
	}
	if (!ret && session->aip) {
		ret = validate_auth_alg_mode(session);
		if (!ret && session->aip->auth_key) {
			if (session->aip->auth_key->iov_len >
			    auth_mode_key_len[session->aip->auth_alg][session->
								      aip->auth_mode])
				return NLM_ERR_AUTH_KEY_LEN_INVALID;
		}
	}
	return ret;
}

#define L3_CACHELINE_SIZE 64
#define L3_CACHELINE_MASK 0x3fULL
#define NUM_64BIT_LOCATIONS 16
#define NUM_128BIT_LOCATIONS 8
#define NUM_256BIT_LOCATIONS 4

static int
crypto_fb_msg_print_err(uint64_t msg)
{
	uint32_t err, ret = 0;

	err = (uint32_t) msg;

	if (err & NLM_CRYPTO_FB_ERR_MASK)
		printf("\nERROR: Security Freeback Data Error: ");
	else
		return ret;

	if (err & NLM_CRYPTO_FB_DATA_ERR_FORCE_BYPASS) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_FORCE_BYPASS;
		printf("        Force bypass \n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_SW_DESC) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_SW_DESC;
		printf("        SW descriptor \n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_INSUFFICIENT_DATA_TO_AUTH) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_INSUFFICIENT_DATA_TO_CIPHER;
		printf("        Insufficient data to authenticate \n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_INSUFFICIENT_DATA_TO_CIPHER) {
		ret |= printf("        Insufficient data to cipher \n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_NO_PKT_DESC) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_NO_PKT_DESC;
		printf("        No pkt descriptor \n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_ECC_SRC_DST) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_ECC_SRC_DST;
		printf("        ECC error in src or dst\n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_ECC_DESC_0123) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_ECC_DESC_0123;
		printf("        ECC error in descriptor 0123\n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_ECC_AUTH_KEY) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_ECC_AUTH_KEY;
		printf("        ECC error in auth key\n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_ECC_CIPHER_KEY) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_ECC_CIPHER_KEY;
		printf("        ECC error in cipher key\n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_ECC_CNTL_DESC) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_ECC_CNTL_DESC;
		printf("        ECC error in CNTL descriptor\n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_ECC_DATA) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_ECC_DATA;
		printf("        ECC error in data\n");
	}
	if (err & NLM_CRYPTO_FB_DATA_ERR_ECC_DESIGNER_FB) {
		ret |= NLM_CRYPTO_FB_DATA_ERR_ECC_DESIGNER_FB;
		printf("        ECC error in designed freeback descriptor\n");
	}
	return ret;
}

static int
create_session(struct crypto_session *session)
{
	void *tmp_cntrl_desc = NULL;
	struct nlm_crypto_cipher_init_param cip, *tmp_cip = NULL;
	struct nlm_crypto_auth_init_param aip, *tmp_aip = NULL;
	unsigned int key_len;

	memset(&cip, 0, sizeof (struct nlm_crypto_cipher_init_param));
	memset(&aip, 0, sizeof (struct nlm_crypto_auth_init_param));

	if (!session->flag)
		session->cntrl_desc =
		    get_cache_aligned_mem(NLM_CRYPTO_SIZEOF_CNTRL_DESC);

	tmp_cntrl_desc = session->cntrl_desc;
	if (tmp_cntrl_desc == NULL) {
		printf("\nError: session->cntrl_desc is NULL");
		return NLM_ERR_INVALID_PARAM;
	}

	if (session->cip) {
		cip.cipher_alg = session->cip->cipher_alg;
		cip.cipher_mode = session->cip->cipher_mode;

		if (session->cip->cipher_key) {
			cip.cipher_key =
			    (struct nlm_iovec *) session->cip->cipher_key;
		}
		cip.arc4_cipher_key_len = session->cip->arc4_cipher_key_len;
		cip.arc4_key_init = session->cip->arc4_key_init;
		cip.cfb_mask = session->cip->cfb_mask;
		tmp_cip = &cip;
	}
	if (session->aip) {
		aip.auth_alg = session->aip->auth_alg;
		aip.auth_mode = session->aip->auth_mode;
		if (session->aip->auth_key) {
			aip.auth_key =
			    (struct nlm_iovec *)
			    malloc(sizeof (struct nlm_iovec));
			if (aip.auth_key == NULL) {
				printf("\n Malloc failed for aip.auth_key");
				return NLM_ERR_NO_MEM;
			}

			key_len =
			    auth_mode_key_len[session->aip->auth_alg][session->
								      aip->auth_mode];
			aip.auth_key->buf =
			    (unsigned char *) malloc(key_len + 1);
			if (aip.auth_key->buf == NULL) {
				printf
				    ("\n Malloc failed for aip.auth_key->buf");
				return NLM_ERR_NO_MEM;
			}
			memset(aip.auth_key->buf, 0, key_len);
			memcpy(aip.auth_key->buf, session->aip->auth_key->buf,
			       session->aip->auth_key->iov_len);
			aip.auth_key->iov_len = key_len;
			session->aip->auth_key->iov_len = key_len;
			aip.hmac = session->aip->hmac;
		}

		tmp_aip = &aip;
	}
	nlm_hal_crypto_preprocess_request((void *) tmp_cntrl_desc, tmp_cip,
					  tmp_aip);

	return 0;

}

/**
 * @brief This API sets up session for performing cipher and authentication
 * operation.
 *
 * @param[in] session_cntx User allocates memory for session context and passes
 * the pointer to this structure. This structure is internally filled in by
 * driver with the provided parameters. Also, validation checks are run on the
 * input parameters for correctness. Application passes this data structure to
 * subsequent cipher and auth apis.
 * @param[in] cipher_alg cipher algorithm
 * @param[in] cipher_mode cipher mode
 * @param[in] cipher_key cipher key
 * @param[in] additional_cipher_info additional cipher information. Valid for
 * ARC4 and other ciphers which need additional information.
 * @param[in] auth_alg authentication algorithm
 * @param[in] auth_mode authentication mode
 * @param[in] auth_key authentication key. NULL in case of MD5 and SHA without
 * hmac.
 * @param[in] tag_len length of the tag which is to be returned on tag
 * computation.
 *
 * @return Return value of 0 indicates success. 
 * Any other error code in case of failure.
 *
 * @sa (see also)
 *
 * @ingroup crypto
 */

int
crypto_setup_cipher_auth_session(struct crypto_session *session)
{
	int ret = 0;
	if (!session)
		return NLM_ERR_INVALID_PARAM;

	if (session->flag
	    && (session->cntrl_desc == NULL || session->pkt_desc == NULL))
		return NLM_ERR_INVALID_PARAM;

	if (session->aip) {
		if (session->aip->hmac
		    && (session->aip->auth_key == NULL
			|| session->aip->auth_alg == NLM_AUTH_BYPASS)) {
			crypto_cleanup_session(session);
			return NLM_ERR_INVALID_PARAM;
		}
	}

	ret = validate_alg_mode_key_size(session);
	if (ret) {
		crypto_cleanup_session(session);
		return ret;
	}

	return create_session(session);
}

/**
 * @brief This API resets key for an already configured session.
 *
 * @param[in] session Specifies the session which is setup by calling
 * one of the session setup apis..
 *
 * @param[in] cipher_key New cipher key. For auth only sessions , this parameter
 * is ignored.
 * @param[in] auth_key New auth key. For cipher only sessions , this parameter
 * is ignored.
 *
 * @return Return value of 0 indicates success. 
 * Any other error code in case of failure.
 *
 * @sa (see also)
 *
 * @ingroup crypto
 */
int
crypto_session_reset_key(struct crypto_session *session, unsigned char
			 *cipher_key, unsigned char *auth_key)
{
	if (cipher_key && session->cip) {
		memcpy(session->cip->cipher_key->buf, cipher_key,
		       session->cip->cipher_key->iov_len);
	}
	if (auth_key && session->aip) {
		memcpy(session->aip->auth_key->buf, auth_key,
		       session->aip->auth_key->iov_len);
	}
	return create_session(session);
}

/**
 * @brief This API cleans up internal driver memory associated with a session.
 *
 * @param[in] session Specifies the session which is setup by calling
 * one of the session setup apis..
 *
 * @return Return value of 0 indicates success. 
 * Any other error code in case of failure.
 *
 * @sa (see also)
 *
 * @ingroup crypto
 */
int
crypto_cleanup_session(struct crypto_session *session)
{
	if (!session)
		return NLM_ERR_INVALID_PARAM;
	if (session->flag == 0) {
		if (session->cntrl_desc)
			free_cache_aligned_mem(session->cntrl_desc);
		if (session->pkt_desc)
			free_cache_aligned_mem(session->pkt_desc);
	}
	return 0;
}

/******************** cipher auth api ****************************/
/*
// Only sessions are supported
User allocates cop, fills in the required fields and calls the API. 
In this API, first field of cop points to session which is already
established.
Return value of 0 indicates success. NLM_ERR_IN_PROGRESS in case the validation 
is successful and async operation is not yet completed.
Any other error code in case of failure.
*/

int
crypto_cipher_auth_op(struct crypto_session *session, struct crypto_param *cop)
{
	int ret = 0, i;
	struct crypto_session *sess = session;
	struct nlm_crypto_param cprm;
	uint32_t rx_vc;
	uint64_t err_msg = 0, tx_id;

	if (!cop)
		return NLM_ERR_CRYPTO_PARAM_NULL;

	if (!sess)
		return NLM_ERR_SESSION_NOT_SETUP;

	memset(&cprm, 0, sizeof (struct nlm_crypto_param));

	session->pkt_desc =
	    get_cache_aligned_mem(NLM_CRYPTO_SIZEOF_PKT_DESC(cop->nr_frags));

	cprm.enc = cop->enc;

	cprm.src_len = cop->src_len;

	cprm.src_phy =
	    (struct nlm_iovec *) malloc(sizeof (struct nlm_iovec) *
					cop->nr_frags);
	cprm.dst_phy =
	    (struct nlm_iovec *) malloc(sizeof (struct nlm_iovec) *
					cop->nr_frags);

	i = 0;
	while (i < cop->nr_frags) {
		cprm.src_phy[i].buf =
		    (unsigned char *) virt_to_phys(cop->src[i].buf);
		cprm.dst_phy[i].buf =
		    (unsigned char *) virt_to_phys(cop->dst[i].buf);
		cprm.src_phy[i].iov_len = cop->src[i].iov_len;
		cprm.dst_phy[i].iov_len = cop->dst[i].iov_len;

		i++;
	}

	if (session->cip) {
		if (cop->iv_len !=
		    cipher_mode_iv_len[session->cip->cipher_alg][session->
								 cip->cipher_mode])
			return NLM_ERR_CIPHER_IV_LEN_INVALID;
	}
	cprm.cipher_len = cop->cipher_len;
	cprm.iv_len = cop->iv_len;
	cprm.iv = cop->iv;

	cprm.iv_offset = cop->iv_offset;
	cprm.cipher_offset = cop->cipher_offset;
	cprm.cipher_bit_count = cop->cipher_bit_count;

	cprm.hash_bit_count = cop->hash_bit_count;
	cprm.tls_proto = cop->tls_proto;
	cprm.arc4_save_l3_alloc = cop->arc4_save_l3_alloc;
	cprm.arc4_save_state = cop->arc4_save_state;
	cprm.arc4_load_state = cop->arc4_load_state;
	cprm.hmac_external_pad_key = cop->hmac_external_pad_key;
	cprm.hash_clobber = cop->hash_clobber;
	cprm.cipher_clobber = cop->cipher_clobber;
	cprm.hash_output_l3_alloc = cop->hash_output_l3_alloc;
	cprm.cipher_output_l3_alloc = cop->cipher_output_l3_alloc;
	cprm.hash_source = cop->hash_source;
	cprm.send_designer_fb = cop->send_designer_fb;

	cprm.auth_len = cop->auth_len;
	cprm.auth_offset = cop->auth_offset;
	cprm.tag_len = cop->tag_len * 8;

	cprm.cipher_key_len = cop->cipher_key_len;
	cprm.auth_key_len = cop->auth_key_len;
	cprm.hash_dst_address_phy = virt_to_phys(cop->hash_dst_address);
	cprm.nr_frags = cop->nr_frags;

	cprm.designer_freeback_id = cop->designer_freeback_id;
	cprm.designer_freeback_len = cop->designer_freeback_len;
	cprm.designer_fb[0] = cop->designer_fb[0];
	cprm.designer_fb[1] = cop->designer_fb[1];
	cprm.designer_fb[2] = cop->designer_fb[2];
	cprm.designer_fb[3] = cop->designer_fb[3];

	rx_vc = crypto_get_fb_vc();
	tx_id = (uint64_t) cop;
	ret = nlm_hal_crypto_send_request(crypto_get_vc(), rx_vc,
					  (void *) session->cntrl_desc,
					  virt_to_phys(session->cntrl_desc),
					  (void *) session->pkt_desc,
					  virt_to_phys(session->pkt_desc),
					  &cprm, tx_id);

	if (!ret) {
		i = 0;
		while (tx_id != nlm_hal_crypto_receive_response(rx_vc, &err_msg)
		       && i < 10000)
			i++;
		if (i < 10000)
			ret = crypto_fb_msg_print_err(err_msg);
		else
		{
			ret = NLM_ERR_FREEBACK_NOT_RECEIVED;
			printf("\n Error _ FreeBack message is not received");
		}
	}

	return ret;
}

static void *
get_cache_aligned_mem(size_t size)
{
	void *addr, **tmp;
	unsigned long loc;
	addr = malloc(size + L3_CACHELINE_SIZE * 2);

	loc = (unsigned long) addr;
	loc = (loc + L3_CACHELINE_SIZE - 1) & ~L3_CACHELINE_MASK;

	tmp = (void **) loc;

	*tmp = addr;

	loc += L3_CACHELINE_SIZE;

	addr = (void *) loc;

	return addr;
}

static void
free_cache_aligned_mem(void *addr)
{
	unsigned long loc;

	loc = (unsigned long) addr;

	loc -= 64;

	addr = (void *) loc;

	addr = (void *) *(unsigned long *) addr;

	free(addr);
}
