/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "i2c.h"
#include "system.h"
#include "iomap.h"

static volatile int i2c_intr_flag[2];
volatile unsigned WIDTH *mmio_i2c[2];
static volatile spinlock_t i2c_lock[2];

irqreturn_t xlr_i2c_int(int irq, struct pt_regs * ptregs) {
#ifdef DEBUG
	{
		volatile unsigned int status = 0;
		int bus;

		if(irq == IRQ_PIC_I2C_0) bus = 0;
		else if(irq == IRQ_PIC_I2C_1) bus = 1;
	
		status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
		printf("I2C intr: status = 0x%x\n", status);
	}
#endif
	if(irq == IRQ_PIC_I2C_0)
		i2c_intr_flag[0] = 1;
	else if(irq == IRQ_PIC_I2C_1)
		i2c_intr_flag[1] = 1;
	return 0;
}

void xlr_i2c_init() {

    mmio_i2c[1] = phoenix_io_mmio(PHOENIX_IO_I2C_1_OFFSET);
    mmio_i2c[0] = phoenix_io_mmio(PHOENIX_IO_I2C_0_OFFSET);    

    phoenix_write_reg(mmio_i2c[0],I2C_CLKDIV, 0x14a);      // 100KHz operation
    phoenix_write_reg(mmio_i2c[0],I2C_HDSTATIM,0x107);    // stop to start time parameter
    phoenix_write_reg(mmio_i2c[1],I2C_CLKDIV, 0x14a);      // 100KHz operation
    phoenix_write_reg(mmio_i2c[1],I2C_HDSTATIM, 0x107);    // stop to start time parameter
}

int xlr_i2c_intr_init(int bus, unsigned int vcpu_mask) {

	int irq = 0;
	unsigned int phoenix_io_i2c_offset = 0;

	if(bus == 0) {
		irq = IRQ_PIC_I2C_0;
		phoenix_io_i2c_offset = PHOENIX_IO_I2C_0_OFFSET;
	}	
	else if(bus == 1) {
		irq = IRQ_PIC_I2C_1;
		phoenix_io_i2c_offset = PHOENIX_IO_I2C_1_OFFSET;
	}
	else {
		printf("%s: wrong bus no.\n", __FUNCTION__);
		return -1;
	}

	if(request_irq(irq, (void *)xlr_i2c_int, 0, vcpu_mask) != 0) {
		printf("request_irq failed for IRQ %d for I2C bus\n", irq);
		return -1;
	};

	mmio_i2c[bus] = phoenix_io_mmio(phoenix_io_i2c_offset);

	phoenix_write_reg(mmio_i2c[bus],I2C_CLKDIV, 0x14a);      // 100KHz operation
	phoenix_write_reg(mmio_i2c[bus],I2C_HDSTATIM, 0x107);    // stop to start time parameter
	phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 0x100);      // enable interrupt 


	/* used in interrupt mode */
	spin_lock_init(&i2c_lock[bus]);

	return 0;
}

/********************************************************************************
 * GENERAL NOTE:
 * Ackerr:	All writes should wait for sdoempty
 *		All reads should wait for datardy
 *		All transactions should wait for idle bit
 *		All startxfr commands should retry on lostarb and starterr
 *		- The above ensures that ACKERR is only due to a non-existant device
 *		On ACKERR, return error
 *	
 *******************************************************************************/


/*******************************************************************************
 * i2c_tx: 
 *          Writes a character to an i2c device
 * Inputs: 
 *          bus (0 or 1), dev_addr (device) 
 *          internal_addr(address within device) data(character to write)
 * Returns: 
 *          0 (success) -1 (failure)
 ******************************************************************************/

static int i2c_tx(unsigned bus, unsigned WIDTH dev_addr, unsigned WIDTH internal_addr, unsigned char data) {

    unsigned WIDTH status = 0;
    unsigned startxfr = 1;

#ifdef DEBUG
    printf("[%s:%d] Checking busy status\n", __FUNCTION__, __LINE__);
#endif
    // Wait for any previous transactions to complete
    while (I2C_BUSY(phoenix_read_reg(mmio_i2c[bus],I2C_STATUS)));

    /*  Interrupts off, 
     *  8 bit internal-device address, 
     *  7 bit i2c-compliant device address
     *  Both addresses enabled
     *  Config value = 0xf8
     */
    phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 0x00f8);

    /* Single Byte transfer */
    phoenix_write_reg(mmio_i2c[bus],I2C_BYTECNT, 0);

    phoenix_write_reg(mmio_i2c[bus],I2C_DEVADDR, dev_addr);
    phoenix_write_reg(mmio_i2c[bus],I2C_ADDR, internal_addr);
    phoenix_write_reg(mmio_i2c[bus],I2C_DATAOUT, (unsigned WIDTH)data);

    do {
        // initiate transfer
        if (startxfr) 
            phoenix_write_reg(mmio_i2c[bus],I2C_START, I2C_WRITE_XFR);
        status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
#ifdef DEBUG
        printf("[%s:%d]  status=%x\n", __FUNCTION__, __LINE__,status);
#endif
        // Failed to transmit
        if (I2C_ACKERR(status)) return -1;
        startxfr = I2C_RESTART_XFER(status); 
    }   while (!I2C_SDOEMPTY(status));
    return 0; 
}

/*******************************************************************************
 * i2c_rx: 
 *          Reads a character from an i2c device
 * Inputs: 
 *          bus (0 or 1), dev_addr (device) 
 *          internal_addr(address within device) 
 * Returns: 
 *          0 (success) -1 (failure). 
 *          On success, data contains the character read back
 ******************************************************************************/

static int i2c_rx(unsigned bus, unsigned WIDTH dev_addr, unsigned WIDTH internal_addr, unsigned char *data) {

    unsigned WIDTH status;
    unsigned startxfr = 1;

    int i = 0;
    
    // Wait for any previous transactions to complete
#ifdef DEBUG
    printf("[%s:%d] Checking busy status (init status = %x)\n", __FUNCTION__,
            __LINE__, phoenix_read_reg(mmio_i2c[bus],I2C_STATUS));
#endif
    while (I2C_BUSY(phoenix_read_reg(mmio_i2c[bus],I2C_STATUS)));

    if (I2C_DATARDY(phoenix_read_reg(mmio_i2c[bus],I2C_STATUS))) 
        *data = (unsigned char) phoenix_read_reg(mmio_i2c[bus],I2C_DATAIN);

    /*  Interrupts off
     *  8 bit internal address
     *  7 bit i2c device address followed by direction bit
     *  Transaction with no data
     *  both addresses enabled to register internal address with device
     *  Config value: 0xf8
     */

    phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 0xf8);
    phoenix_write_reg(mmio_i2c[bus],I2C_BYTECNT, 0);
    phoenix_write_reg(mmio_i2c[bus],I2C_DEVADDR,dev_addr);
    phoenix_write_reg(mmio_i2c[bus],I2C_ADDR, internal_addr);

    do {
        // initiate transfer
        if (startxfr) 
            phoenix_write_reg(mmio_i2c[bus],I2C_START, I2C_NODATA_XFR);
        status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
        // Failed to transmit
        if (I2C_ACKERR(status)) return -1;
        startxfr = I2C_RESTART_XFER(status);
    } while (I2C_BUSY(status));


    for (i=0;i<4000;i++);

    /* 
     *  Interrupts off
     *  8 bit internal address, not transmitted
     *  7 bit device address, transmitted
     *  Read single byte transaction
     *  Config value: 0xfa
     */
    phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 0x00fa);
    phoenix_write_reg(mmio_i2c[bus],I2C_BYTECNT, 0);
    startxfr = 1;
    do {
        // initiate transfer
        if (startxfr)
            phoenix_write_reg(mmio_i2c[bus],I2C_START, I2C_READ_XFR);
        status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
        // Failed to transmit
        if (I2C_ACKERR(status)) return -1;
        startxfr = I2C_RESTART_XFER(status);   
    } while (!I2C_DATARDY(status));

    *data = (unsigned char)phoenix_read_reg(mmio_i2c[bus],I2C_DATAIN);
    return 0;
}

/*******************************************************************************
 *
 * Write_board_eeprom:
 * Writes a value to an offset on the board eeprom
 * Inputs:  internal_addr(address within device) , data to write
 * Returns: 0 (success) -1 (failure). 
 *
 ******************************************************************************/
int write_board_eeprom(unsigned WIDTH offset, unsigned char data) {
#ifdef DEBUG
    printf("[%s:%d] Writing character %c (0x%x),to location %d\n", __FUNCTION__, __LINE__, data, data, offset);
#endif
    return(i2c_tx(ARIZONA_EEPROM_BUS,ARIZONA_EEPROM_ADDR,offset,data));
}

/*******************************************************************************
 *
 * read_board_eeprom
 * Reads a character from an offset in the board eeprom
 * Inputs: offset inside the eeprom
 * Returns: 0 (success) -1 (failure). On success, data contains the character read back
 *
 ******************************************************************************/
int read_board_eeprom(unsigned WIDTH offset, unsigned char *data) {
    return(i2c_rx(ARIZONA_EEPROM_BUS,ARIZONA_EEPROM_ADDR,offset,data));
}

/*******************************************************************************
 *
 * read_spd_eeprom
 * Reads a character from an offset in the spd eeprom
 * Inputs: offset inside the eeprom
 * Returns: 0 (success) -1 (failure). On success, data contains the character read back
 *
 ******************************************************************************/
int read_spd_eeprom(unsigned WIDTH offset, unsigned char *data) {
    return(i2c_rx(ARIZONA_SPD_BUS,ARIZONA_SPD_EEPROM0,offset,data));
}

/*******************************************************************************
 *
 * read_thermsensor
 * Reads current temperature
 * Inputs: none
 * Returns: 0 (success) -1 (failure). On success, data contains the temperature
 *
 ******************************************************************************/
int read_ext_tempsensor(unsigned char *data){
    int status;
    status = i2c_rx(ARIZONA_TEMPSENSOR_BUS,ARIZONA_TEMPSENSOR_ADDR,ARIZONA_TEMPSENSOR_EXT_TEMP,data);
    if (!status) 
        *data = *data & 0x7f;
    return status;
}

/*
  NETOS being single-task model, doesn't have a proper sleep/wakeup mechanism.
  xlr_i2c_sleep() and xlr_i2c_int() play a workaround for this.
*/
static void xlr_i2c_sleep(int bus) {
	int i;
	
	/* without this delay, STARTERR happens mostly */
    	for (i = 0; i < 0xfffff; i++) ;

	/* sleep till woken by the interrupt */
	while(!i2c_intr_flag[bus]) ;
	i2c_intr_flag[bus] = 0;
}

static int i2c_intr_rx(unsigned bus, unsigned WIDTH dev_addr, unsigned WIDTH internal_addr, unsigned char *data) {
	volatile unsigned WIDTH status;
	unsigned startxfr = 1;
	
	spin_lock((spinlock_t *) &i2c_lock[bus]);
#ifdef DEBUG
	printf("%s: initial status = 0x%x\n", __FUNCTION__, 
                phoenix_read_reg(mmio_i2c[bus],I2C_STATUS));
	printf("%s: initial config = 0x%x\n", __FUNCTION__, 
            phoenix_read_reg(mmio_i2c[bus],I2C_CONFIG));
	printf("%s: initial start = 0x%x\n", __FUNCTION__, 
            phoenix_read_reg(mmio_i2c[bus],I2C_START));
#endif


	/* Wait for any previous transactions to complete */
	status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
	if(I2C_BUSY(status))
	{
		/* this may work only for single vcpu;
                   multi-vcpu simultaneous should be avoided */
		spin_unlock((spinlock_t *) &i2c_lock[bus]);
		xlr_i2c_sleep(bus);
		spin_lock((spinlock_t *) &i2c_lock[bus]);
	}

	
	/*  Interrupts on
	*  8 bit internal address
	*  7 bit i2c device address followed by direction bit
	*  "No data" transfer
	*  both addresses enabled to register internal address with device
	*  Config value: 0xf8
	*/
	phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 
            phoenix_read_reg(mmio_i2c[bus], I2C_CONFIG) & 0x100);
	phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 
            phoenix_read_reg(mmio_i2c[bus], I2C_CONFIG) | 0xf8);
	phoenix_write_reg(mmio_i2c[bus],I2C_BYTECNT, 0); /* single byte xfer */
	phoenix_write_reg(mmio_i2c[bus],I2C_DEVADDR, dev_addr);
	phoenix_write_reg(mmio_i2c[bus],I2C_ADDR, internal_addr);

	do {
		/* start transfer */
		if (startxfr) 
			phoenix_write_reg(mmio_i2c[bus],I2C_START, I2C_NODATA_XFR);
		
		/* wait for interrupt */
		xlr_i2c_sleep(bus);

		/* decode the interrupt */
		status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
		if(I2C_ACKERR(status)) {
			spin_unlock((spinlock_t *) &i2c_lock[bus]);
			return -1; /* the device is non-existent ? */
		}
        	startxfr = I2C_RESTART_XFER(status);   
	} while (I2C_BUSY(status));

#ifdef DEBUG
	printf("%s: aft nodata xfer: I2C status = 0x%x\n", __FUNCTION__, 
            phoenix_read_reg(mmio_i2c[bus],I2C_STATUS));	
#endif

	/* 
	*  Interrupts on
	*  8 bit internal address, not transmitted
	*  7 bit device address, transmitted
	*  Read transfer
	*  Config value: 0xfa
	*/
	phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 
            (phoenix_read_reg(mmio_i2c[bus], I2C_CONFIG) & 0x100));
	phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 
        (phoenix_read_reg(mmio_i2c[bus], I2C_CONFIG) | 0xfa));
	phoenix_write_reg(mmio_i2c[bus],I2C_BYTECNT,  0);
	startxfr = 1;
	do {
		/* start transfer */
		if (startxfr)
			phoenix_write_reg(mmio_i2c[bus],I2C_START, I2C_READ_XFR);

		/* wait for interrupt */
		xlr_i2c_sleep(bus);

		/* decode the interrupt */
		status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
		if(I2C_ACKERR(status)) {
			spin_unlock((spinlock_t *) &i2c_lock[bus]);
			return -1; /* the device is non-existent ? */
		}
        	startxfr = I2C_RESTART_XFER(status);   
	} while (!I2C_DATARDY(status));

	*data = (unsigned char)phoenix_read_reg(mmio_i2c[bus],I2C_DATAIN);
#ifdef DEBUG
	printf("%s: aft read xfer: I2C status = 0x%x\n", __FUNCTION__,
            phoenix_read_reg(mmio_i2c[bus],I2C_STATUS));
#endif
	spin_unlock((spinlock_t *) &i2c_lock[bus]);

	return 0;
}

static int i2c_intr_tx(unsigned bus, unsigned WIDTH dev_addr, unsigned WIDTH internal_addr, unsigned char data) {
	unsigned WIDTH status = 0;
	unsigned startxfr = 1;
	int i;

	spin_lock((spinlock_t *) &i2c_lock[bus]);
#ifdef DEBUG
	printf("%s: initial status = 0x%x\n", __FUNCTION__,
            phoenix_read_reg(mmio_i2c[bus],I2C_STATUS));
	printf("%s: initial config = 0x%x\n", __FUNCTION__, 
                phoenix_read_reg(mmio_i2c[bus],I2C_CONFIG));
	printf("%s: initial start = 0x%x\n", __FUNCTION__, phoenix_read_reg(mmio_i2c[bus],I2C_START));
#endif

	/* Wait for any previous transactions to complete */
	status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);
	if(I2C_BUSY(status)) {
		/* this may work only for single vcpu;
                   multi-vcpu simultaneous should be avoided */
		spin_unlock((spinlock_t *) &i2c_lock[bus]);
		xlr_i2c_sleep(bus);
		spin_lock((spinlock_t *) &i2c_lock[bus]);
	}

	/*  Interrupts on, 
	*  8 bit internal-device address, 
	*  7 bit i2c-compliant device address
	*  Both addresses enabled
	*  Write transfer
	*  Config value = 0xf8
	*/

	phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 
            phoenix_read_reg(mmio_i2c[bus], I2C_CONFIG) & 0x100);
	phoenix_write_reg(mmio_i2c[bus],I2C_CONFIG, 
            phoenix_read_reg(mmio_i2c[bus], I2C_CONFIG) | 0xf8);
	phoenix_write_reg(mmio_i2c[bus],I2C_BYTECNT, 0); /* single byte xfer */
	phoenix_write_reg(mmio_i2c[bus],I2C_DEVADDR, dev_addr);
	phoenix_write_reg(mmio_i2c[bus],I2C_ADDR, internal_addr);
	phoenix_write_reg(mmio_i2c[bus],I2C_DATAOUT, (unsigned WIDTH)data);

	do {
		/* start transfer */
		if (startxfr) 
			phoenix_write_reg(mmio_i2c[bus],I2C_START,I2C_WRITE_XFR);

		/* wait for interrupt */
		xlr_i2c_sleep(bus);
		status = phoenix_read_reg(mmio_i2c[bus],I2C_STATUS);

		/* decode the interrupt */
		if(I2C_ACKERR(status)) {
			spin_unlock((spinlock_t *) &i2c_lock[bus]);
			return -1; /* the device is non-existent ? */
		}
        	startxfr = I2C_RESTART_XFER(status);   
	}   while (!I2C_SDOEMPTY(status));

	/* without this delay, a further read/write gives unpredictable behaviour */
	for (i = 0; i < 0xffffff; i++);
#ifdef DEBUG
	printf("%s: aft write xfer: I2C status = 0x%x\n", __FUNCTION__, 
            phoenix_read_reg(mmio_i2c[bus],I2C_STATUS));	
#endif
	spin_unlock((spinlock_t *) &i2c_lock[bus]);

	return 0; 
}

int i2c_intr_read (
	uint32_t bus, /* i2c bus number */
	uint32_t dev_addr,  /* slave device address on i2c bus */
	uint32_t internal_addr,  /* slave device memory location */
	uint8_t * buf, /* input buffer */
	int num_bytes /* no. of bytes to read */
)
{
        int i;
        for(i = 0; i < num_bytes; i++) {
                if(i2c_intr_rx(bus, dev_addr, internal_addr++, buf + i)) {
			printf("%s: ack error in read, returning without proceeding further\n", __FUNCTION__);
			return -1;
		}
	}
	return 0;
}

int i2c_intr_write
(
	uint32_t bus, /* i2c bus number */
	uint32_t dev_addr,  /* slave device address on i2c bus */
	uint32_t internal_addr,  /* slave device memory location */
	uint8_t * buf, /* output buffer */
	int num_bytes /* no. of bytes to write */
)
{
        int i;
        for(i = 0; i < num_bytes; i++)
        {
                if(i2c_intr_tx(bus, dev_addr, internal_addr++, buf[i])) {
			printf("%s: ack error in write, returning without proceeding further\n", __FUNCTION__);
			return -1;
		}
        }
	return 0;
}
