/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

/********************************************************************************
 gdbstub.c - A gdb debugger stub for NETOS running on xlr732-mips chip

 MODIFICATION HISTORY
 --------------------
 v1.0,16oct05,Suresh DS    single vCPU debugging through serial,
                           multicore debugging through network

 v1.1,25jan06,Suresh DS    single code base for serial and network modes,
                           support for run-time indication of serial/network mode,
                           multicore debugging through serial,
                           "ctrl-c stopping of execution" feature
                           and some other enhancements    

 v1.2,15mar06,Suresh DS    support for shared code segment (kuseg & kseg0),
                           "breakall" feature,
                           support for "initial break" as a run-time option

 v1.3,02jun06,Suresh DS    "break_t" feature,
                           "continueall" feature,
                           "vcpustate" feature,
                           "single-step+tbh" bug fix,
                           "host-has-attached+no-initial-break" condition handling,
                            code abstraction.

 v1.4,05jul06,Suresh DS    conversion of shared memory to circular FIFO,
                           removal of message ring dependancy,
                           addition of intelligence to allow the status message to
                           be sent only when gdb expects it,
                           changing the continueall packet from 'F' to '9'    

 v1.5,10jul06,Suresh DS    "qSymbol" packet support

 v1.6,23jan08,Suresh DS    pointers re-cast to long instead of unsigned long,
                           indentation,
                           cleanup

 v1.7,09feb09,Suresh DS    infrastructure changes
                           (debugger options taken at run-time instead of compile-time,
                           O3 support)

 v1.8,05mar09,Suresh DS    support for uart 0

 v2.0,15sep09,Suresh DS    Handling multiple vcpus from single session (SDK 1.7)

 v2.1,04dec09,Suresh DS    Converting the serial tunneler into a stream

 v2.2,08jan10,Suresh DS    Multi-session stub removed and single-session stub renamed as this file,
                           necessary support added

 v2.3,25jan10,Suresh DS    TBH cleaned up and optimized


 DESCRIPTION
 -----------
 This gdb stub is written for the xlr processor running NETOS. It operates in 
 two modes viz. serial and network. In serial mode, it uses UART1 of xlr to 
 communicate to the remote gdb host over a serial cable. In network mode, it 
 communicates to remote gdb host through TCP/IP network connection  provided by
 any other OS running in xlr in one of the virtual CPUs (given as compile-time 
 option; default : VCPU 0). This module is based on the stub for the Hitachi SH
 processor written by Ben Lee    and Steve Chamberlain and that for the MIPS 
 processor running rtems written by Ralf.  The former in turn "is originally 
 based on an m68k software stub written by Glenn Engel at HP, but has    changed 
 quite a bit". The changes for the R4600 were written by C.M.Heard at VVNET. 
 They were based in part on the Algorithmics R4000 version of Phil Bunce's 
 PMON program.

 Remote communication protocol:

 A debug packet whose contents are <data> is encapsulated for transmission in 
 the form:

 $ <data> # CSUM1 CSUM2

     <data> must be ASCII alphanumeric and cannot include characters
     '$' or '#'.  If <data> starts with two characters followed by
     ':', then the existing stubs interpret this as a sequence number.

     CSUM1 and CSUM2 are ascii hex representation of an 8-bit
     checksum of <data>, the most significant nibble is sent first.
     the hex digits 0-9,a-f are used.

 Receiver responds with:

     +    if CSUM is correct
     -    if CSUM is incorrect

 <data> is as follows.  All values are encoded in ascii hex digits.

     Request         Packet

     read registers  g
     reply           XX....X         Each byte of register data is described by 
                                     two hex digits. Registers are in the internal order
                                     for GDB, and the bytes in a register are in the same order 
                                     the machine uses.
                 or  ENN             for an error.

     write regs      GXX..XX         Each byte of register data is described by 
                                     two hex digits.
     reply           OK              for success
                 or  ENN             for an error

     read reg        pn...=r...      Read register n... with value r....
     reply           OK              for success
                 or  ENN             for an error

     read memory     mAA..AA,LLLL    AA..AA is address, LLLL is length.
     reply           XX..XX          XX..XX is mem contents
                                     Can be fewer bytes than requested
                                     if able to read only part of the data.
                 or  ENN             NN is errno

     write mem       MAA..AA,LLLL:XX..XX
                                     AA..AA is address,
                                     LLLL is number of bytes,
                                     XX..XX is data
     reply           OK              for success
                 or  ENN             for an error (this includes the case where 
                                     only part of the data was written).

     set thread      Hct             set thread t for subsequent operations
     reply           OK              for success
                 or  ENN             for an error

     query           q<str>          query for str
     reply           RES             RES is response for str 
                 or  ENN             for an error
 
     insert breakpoint
                     Ztype,addr,length
                                     Insert breakpoint of "type" at "addr" and covering the next 
                                     "length" bytes
     reply           OK              for success
                 or  ENN             for an error
 
     remove breakpoint
                     ztype,addr,length
                                     Remove breakpoint of "type" at "addr" and covering the next 
                                     "length" bytes
     reply           OK              for success
                 or  ENN             for an error
 
     cont            cAA..AA         AA..AA is address to resume
                                     If AA..AA is omitted, resume at same address.

     step            sAA..AA         AA..AA is address to resume
                                     If AA..AA is omitted, resume at same address.

     There is no immediate reply to step or cont.
     The reply comes when the machine stops.
     It is           SAA             AA is the "signal number"

     last signal     ?               Reply with the reason for stopping.
                                     This is the same reply as is generated for 
                                     step or cont: SAA where AA is the signal number.

     detach          D               Host is detaching.  Reply OK and end remote 
                                     debugging session.
 
     quit            k               Host is detaching.  Reply OK and end remote debugging 
                                     session.

     reserved        <other>         On other requests, the stub should ignore 
                                     the request and send an empty response ($#<checksum>).  This
                                     way we can extend the protocol and GDB can tell whether the 
                                     stub it is talking to uses the old or the new.

     Responses can be run-length encoded to save space.  A '*' means that
     the next character is an ASCII encoding giving a repeat count which
     stands for that many repetitions of the character preceding the '*'.
     The encoding is n+29, yielding a printable character when n >=3
     (which is where rle starts to win).  Don't use n > 99 since gdb
     masks each character is receives with 0x7f in order to strip off
     the parity bit.

     As an example, "0* " means the same thing as "0000".

 INCLUDES:
 gdbstub.h

********************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "gdbstub.h"

/* helper routine for L1 cache operations */
#define gdb_cache_op(op, base) __asm__ __volatile__ ("cache %0, 0(%1)\n" : : "i"(op), "r"(base))

#define barrier_i(addr) __asm__ __volatile__(".set mips4" "synci 0(%0)\n"::"r"(addr) :"memory")

/* external function declarations */
extern void phoenix_break_handler (void);
extern void phoenix_trap_handler (void);
extern int set_except_vector (int n, void *addr);
extern void trap_init (void);
extern void gdb_uart_init (void);
extern int32_t get_debug_char (void);
extern int32_t put_debug_char (int8_t c);

/* local functions declarations */
static uint64_t get_target_addr (uint64_t pc, struct pt_regs *ptregs);
static void gdb_status (uint8_t mode, uint16_t sigval, struct pt_regs *ptregs);
static void gdb_send_ipi (uint32_t vector, uint32_t broadcast, uint32_t nmi, uint32_t thr_id, uint32_t cpu_id);
static void update_online_map (int which, int what, int vcpu_id);

/* external variable declarations */
extern uint8_t _loadaddr[];
extern uint8_t _endaddr[]; /* from linker script (to decide branch possibility) */

static volatile struct gdb_bpnode gdb_breakpoint[TOTAL_VCPU][TOTAL_BREAKPOINTS] __shared_memory__;
static volatile struct gdb_ssnode gdb_instr_buffer[TOTAL_VCPU] __shared_memory__;
static volatile struct gdb_tbhnode gdb_tbhbreakpoint __shared_memory__;

static volatile uint8_t gdb_ack_allowed[TOTAL_VCPU] __shared_memory__;
static volatile uint8_t gdb_status_allowed[TOTAL_VCPU] __shared_memory__;

extern int gdb_connection_mode;
extern int gdb_uart;
extern int gdb_initial_break_val;

/* some flags */
static volatile int32_t host_has_attached[TOTAL_VCPU];	/* indicates if host has detached */
static volatile int32_t gdb_initial_break[TOTAL_VCPU];	/* indicates the initial break caused by gdbstub_init() */
static volatile int32_t xlr_binary[TOTAL_VCPU];	/* for later use */
static volatile int8_t gdblock[TOTAL_VCPU];
static volatile int8_t ctrlc_ipi[TOTAL_VCPU];
static volatile int gdb_vcont_vcpus __shared_memory_bss__ = 0;
static volatile int gdb_vcpu_c __shared_memory_bss__ = -1;
static volatile int gdb_vcpu_g __shared_memory_bss__ = -1;
static volatile int gdb_current_vcpu __shared_memory_bss__ = -1;

/* --- Multi-session related stuff --- */
/* Variables referring to the shared memory segment in kseg0 */
static volatile uint32_t *online_map = (uint32_t *)(long)(OS_NETOS_CPU_ONLINE_MAP);
static volatile spinlock_t *online_map_lock; /* and it's associated lock */
static volatile spinlock_t *tbh_lock = (spinlock_t *)(long)(OS_NETOS_TBH_LOCK); /* and it's associated lock */
static volatile uint8_t *gdb_tx_pkt[TOTAL_VCPU]; /* points to Tx pkt (to gdb) in shared memory  */
static volatile struct fifo *gdb_tx_pkt_fifo[TOTAL_VCPU];
static volatile unsigned int *gdb_tx_pkt_len[TOTAL_VCPU];
static volatile struct fifo *gdb_tx_pkt_len_fifo[TOTAL_VCPU]; /* points to Tx pkt len in shared memory */
static volatile uint8_t *gdb_rx_pkt[TOTAL_VCPU]; /* points to Rx pkt (from gdb) in shared memory  */
static volatile struct fifo *gdb_rx_pkt_fifo[TOTAL_VCPU];

static volatile int8_t gdb_tbh __shared_memory_bss__ = 0;
static volatile uint32_t tbh_vcpu __shared_memory_bss__ = -1;
static volatile uint8_t gdb_vcont_buf[TOTAL_VCPU][OS_NETOS_TX_BUF_SIZE] __shared_memory__;
static volatile int gdb_vcont_buf_len[TOTAL_VCPU] __shared_memory__;
static volatile int read_from_vcont_buf[TOTAL_VCPU] __shared_memory__;
static volatile int gdb_stopped_vcpus[TOTAL_VCPU] __shared_memory__;
static volatile int gdb_started_vcpus[TOTAL_VCPU] __shared_memory__;
void gdb_exception(struct pt_regs *ptregs);

/* Set the Break - Exception Handler (Expt Code 9) */
static void gdb_break_init(void)
{
    set_except_vector(9, gdb_exception);
}

/* Set the Trap - Exception Handler (Expt Code 13) */
/*
   NOTE: This function is not named trap_init() as there is already a trap_init()
     in NETOS even though it doesn't actually set the trap handler. (This 
     is linux legacy stuff)
 */
static void xlr_trap_init(void)
{
    set_except_vector(13, gdb_exception);
}

/* Function to call a break instruction */
static void __attribute__ ((noinline)) gdb_break(void)
{
#if 0
    if (code >= 0xfffff) {
        printf("%s: Unknown code\n", __FUNCTION__);
        return;
    }
    __asm__ __volatile__("break %0\n"::"r"(code));
#endif
    __asm__ __volatile__(".word 0x8d\n\t");
}

static void gdb_mdelay(uint32_t msec)
{
    uint64_t i = 0;
    uint64_t mips_counter_frequency = 1200 * 1000 * 1000;
    uint64_t loop_count = (mips_counter_frequency / 1000) * msec;
    
    for (i = 0; i < loop_count; i++) ;
}

static void gdb_flush_icache_addr (uint64_t base)
{
    gdb_cache_op (Hit_Writeback_Inv_D, base);
    gdb_cache_op (Hit_Invalidate_I, base);
}

static __inline__ void gdb_lock(int vcpu_id)
{
    spin_lock((spinlock_t *) tbh_lock);
    gdblock[vcpu_id] = 1;
}

static __inline__ void gdb_unlock(int vcpu_id)
{
    if (gdblock[vcpu_id]) {
        gdblock[vcpu_id] = 0;
        
        spin_unlock((spinlock_t *) tbh_lock);
    }
    /*barrier();*/
}

/* check if breakpoint is still needed at this address */
static uint8_t gdb_break_needed(uint64_t * addr)
{
#if 0
    volatile struct gdb_bpnode *z0 = NULL; /* pointer to list of normal breakpoints */
#endif
    int vcpu_id = processor_id();
    int id;
    /*int32_t bp;*/
    
    for (id = 0; id < TOTAL_VCPU; id++) {
#if 0
        if (!(*online_map & (1 << id)))
            continue;
        
        /* index into vcpu id */
        z0 = gdb_breakpoint[id];
        
        /* search the user breakpoint nodes */
        for (bp = 0; bp < TOTAL_BREAKPOINTS; bp++) {
#ifdef GDB_DEBUG_MSG
            printf("%u:bn():vcpu=%u,bp=%d\n", vcpu_id, id, bp);
#endif
            if (!(z0[bp].addr))
                break;
            if (id == vcpu_id) {
                if ((addr == z0[bp].addr) && (z0[bp].valid >= 2))
                    return 1;
            }
            else if ((addr == z0[bp].addr) && z0[bp].valid)
                return 1;
        }
        
        if (id == vcpu_id)
            continue;
#endif
#ifdef GDB_DEBUG_MSG
        printf("%d:bn():vcpu=%d,ssb\n", vcpu_id, id);
#endif
        if (!(*online_map & (1 << id)) || id == vcpu_id)
            continue;

        /* search the single-step breakpoint node */
        if (gdb_instr_buffer[id].targetAddr && (addr == gdb_instr_buffer[id].targetAddr))
            return 1;
    }
    
    /* break not needed */
    return 0;
}

/* Function to undo the single step breakpoint */
/* If a step breakpoint was planted, restore the saved instruction */
static void undo_single_step(void)
{
    int vcpu_id = processor_id();
    
    /* If no other vcpu needs this break, replace it with original instruction, flush i-cache, and sync */
    if (!gdb_break_needed ((uint64_t *) (gdb_instr_buffer[vcpu_id].targetAddr)))
        *(uint32_t *)(gdb_instr_buffer[vcpu_id].targetAddr) = gdb_instr_buffer[vcpu_id].savedInstr;
    /*printf("%d:uss:Ins %x At %llx\n", vcpu_id, *(uint32_t *)(gdb_instr_buffer[vcpu_id].targetAddr), (int64_t)(gdb_instr_buffer[vcpu_id].targetAddr));*/
    gdb_flush_icache_addr ((int64_t)(long)(gdb_instr_buffer[vcpu_id].targetAddr));
    barrier();

    /* Update the single step nodes with defaults */
    gdb_instr_buffer[vcpu_id].targetAddr = NULL;
    gdb_instr_buffer[vcpu_id].savedInstr = BREAK_INSTR;
}

/* get original instruction for current pc */
static uint32_t get_orig_instr(uint64_t * addr)
{
#if 0
    volatile struct gdb_bpnode *z0 = NULL; /* pointer to list of normal breakpoints */
#endif
    uint32_t id;
    /*int bp;*/
    
    for (id = 0; id < TOTAL_VCPU; id++) {

        if (!(*online_map & (1 << id)))
            continue;
        
        /* Index into the vcpu id */
#if 0
        z0 = gdb_breakpoint[id];
        
        /* Search the normal breakpoint nodes */

        for (bp = 0; bp < TOTAL_BREAKPOINTS; bp++) {

            if (!(z0[bp].addr)) break;
            /* TODO: The 2nd is ideally not needed; remove it in the end */
            if ((addr == z0[bp].addr) && (z0[bp].instr != BREAK_INSTR))
                return z0[bp].instr;
        }
#endif
        
        /* Search the single-step breakpoint node */
        if (gdb_instr_buffer[id].targetAddr && (gdb_instr_buffer[id].targetAddr == addr) && (gdb_instr_buffer[id].savedInstr != BREAK_INSTR))
            return gdb_instr_buffer[id].savedInstr;
    }
    
    /* Could not get original instruction */
    return BREAK_INSTR;
}

static void print_listinfo(const char *arg0, char *arg1)
{
#ifdef GDB_DEBUG_MSG
    /* local declarations */
    volatile struct gdb_bpnode *z0 = NULL;	/* now multiple breakpoint setting is supported */
    uint32_t id;
    int32_t bp;
    
    printf("%u:%s:%s\n", processor_id(), arg0, arg1);
    printf("----------------------\n");
    for (id = 0; id < TOTAL_VCPU; id++) {
        if (!(*online_map & (1 << id)))
            continue;
        printf("ListInfoFor %u:\n", id);
        z0 = gdb_breakpoint[id];
        for (bp = 0; bp < TOTAL_BREAKPOINTS; bp++) {
            if (!(z0[bp].addr))
                break;
            printf("node %d: addr=0x%llx,instr=%x,valid=%d\n ", bp, (int64_t) (long)(z0[bp].addr), z0[bp].instr, z0[bp].valid);
        }
    }
#endif
}

#if 0
/* TODO: Function to print debug messages */
static void debug_printf(char *str, ...)
{
}
#endif

/* Global variable to hold the (optional) address following the 'c' command */
static volatile uint64_t gnext_addr[TOTAL_VCPU];

/* Transmit/Receive buffer to hold data from / to shared memory */
static volatile uint8_t gdb_rxtx_buf[TOTAL_VCPU][OS_NETOS_TX_BUF_SIZE];

/* Used by slave vcpu to intimate master vcpu to send Tx data */
static void inform_tunneler(uint32_t len)
{
    int32_t indx;
    
    while ((indx = get_fifo_insert((struct fifo *)gdb_tx_pkt_len_fifo[0])) == FIFO_FULL) barrier();
    *(gdb_tx_pkt_len[0] + indx) = len;
    fifo_insert ((struct fifo *)gdb_tx_pkt_len_fifo[0]);
}

/* Slave-specific function to send packet to gdb without needing ack */
static void gdb_putpacket_noack_slave(uint8_t * buf, int len)
{
    
    int count; /* no.of chars in packet data */
    uint8_t checksum; /* calculated checksum of packet data */
    int vcpu_id = processor_id();
    int indx;
    int k;
    
    /* If this is NOT the CURRENT THREAD, then copy the pkt to the CURRENT THREAD's global buffer */
    if (gdb_vcpu_g == vcpu_id && vcpu_id != gdb_current_vcpu) {
        gdb_vcont_buf_len[gdb_current_vcpu] = len;
        memcpy ((uint8_t *) gdb_vcont_buf [gdb_current_vcpu], (uint8_t *) buf, len);
        barrier ();
        read_from_vcont_buf[gdb_current_vcpu] = 1;
        return;
    }
/* TODO: checksum calculation need not be done separately, it can be integrated with the code below */
    /* Calculate the checksum */
    for (count = 0, checksum = 0; count < len; count++)
        checksum += buf[count];
    
    /* Construct and send the packet to this vcpu's shared memory region */
    while ((indx = get_fifo_insert((struct fifo *)gdb_tx_pkt_fifo[0])) == FIFO_FULL) barrier();
    *(gdb_tx_pkt[0] + indx) = '$';
    fifo_insert ((struct fifo *)gdb_tx_pkt_fifo[0]);

    for (k = 0; k < len; k++) {
        while ((indx = get_fifo_insert((struct fifo *)gdb_tx_pkt_fifo[0])) == FIFO_FULL) barrier();
        *(gdb_tx_pkt[0] + indx) = buf[k];
        fifo_insert ((struct fifo *)gdb_tx_pkt_fifo[0]);
    }

    while ((indx = get_fifo_insert((struct fifo *)gdb_tx_pkt_fifo[0])) == FIFO_FULL) barrier();
    *(gdb_tx_pkt[0] + indx) = '#';
    fifo_insert ((struct fifo *)gdb_tx_pkt_fifo[0]);

    while ((indx = get_fifo_insert((struct fifo *)gdb_tx_pkt_fifo[0])) == FIFO_FULL) barrier();
    *(gdb_tx_pkt[0] + indx) = uchar_to_hex (checksum >> 4);
    fifo_insert ((struct fifo *)gdb_tx_pkt_fifo[0]);

    while ((indx = get_fifo_insert((struct fifo *)gdb_tx_pkt_fifo[0])) == FIFO_FULL) barrier();
    *(gdb_tx_pkt[0] + indx) = uchar_to_hex (checksum & 0xf);
    fifo_insert ((struct fifo *)gdb_tx_pkt_fifo[0]);
    
    barrier ();
    
    /* Inform the tunneler thro' shared memory */
    inform_tunneler (len + 4);
    /*printf("%d:L%dFor'%c'\n", vcpu_id, len, buf[0]);*/
}

/* Function to send no acknowledgement packet to gdb */
static void gdb_putpacket_noack(uint8_t * buf, int len)
{
    gdb_putpacket_noack_slave (buf, len);
    return;
}

/* Transmits buf to gdb, and waits for an acknowledgement */
static uint32_t gdb_putpacket(uint8_t * buf, int len)
{
    uint32_t try_again = MAX_GDB_PUTPACKET_TRIES;
    uint8_t c;		/* ack/nack character received from host  */
    /*int vcpu_id = processor_id();*/
    int indx;
#ifdef GDB_DEBUG_MSG
    printf("%d entered %s\n", vcpu_id, __FUNCTION__);
#endif
    
    /* Retry a few times if the message doesn't go */
    while (try_again--) {
        
        /* Send the packet */
        gdb_putpacket_noack (buf, len);
        if (processor_id() != gdb_current_vcpu) return 0;
        
        /* Get response from receiver, ignore garbage packets */
        do {
            
            /* Clear Rx shared memory contents after getting ack/nack character */
            while ((indx = get_fifo_remove((struct fifo *)gdb_rx_pkt_fifo[0])) == FIFO_EMPTY) barrier();
            c = *(gdb_rx_pkt[0] + indx);
            /*printf("%d:gp():GA'%c'\n", vcpu_id, c); */
            fifo_remove((struct fifo *)gdb_rx_pkt_fifo[0]);
            barrier();
        } while (c != '+' && c != '-');
#if 0   /* finite wait */
        /* get response from receiver, ignore garbage packets */
        for (i = 0; i < 20; i++) {
           if ((indx = get_fifo_remove((struct fifo *)f)) != FIFO_EMPTY)
               break;
           gdb_mdelay(1);
           barrier();
        }
        if (i == 20)
            break; /* no response from gdb, quit */
        c = *(gdb_rx_pkt[vcpu_id] + indx);
        if (c == '$')
            break; /* a new pkt, handle it */
        printf("%u:gp():GA'%c'\n", vcpu_id, c);
        fifo_remove((struct fifo *)f);
        barrier();
#endif
        
        /* if message was received ok, we're done */
        if (c == '+') break;
    } /* while(try_again--) */
    
#ifdef GDB_DEBUG_MSG
    printf("%d leaving %s\n", vcpu_id, __FUNCTION__);
#endif
    return 0;
}

/* Function to extract a complete packet from shared memory. This returns the 
   no. of (data) bytes received excluding RSP */
/* TODO: Change the return value to int, and at all the places where this is called */
static int gdb_getpacket_secondary (uint8_t * rxbuf, uint32_t maxlen)
{
    uint8_t checksum;	/* of transmitted data of a single packet */
    uint8_t rx_checksum;	/* of received data of a single packet */
    uint32_t count;		/* no. of bytes of data received */
    int vcpu_id = processor_id();
    int8_t ack;
    int indx;
    int k;
    
    /*printf("%d:entered ggs\n", vcpu_id);*/
    
get_packet:
    
    /* Start counting bytes */
    checksum = 0;
    count = 0;
    k = 0;
    
    /* Non-current threads pick their packets from global buffer populated by 
       CURRENT THREAD */
    if (!gdb_status_allowed[vcpu_id])
        while (!read_from_vcont_buf[vcpu_id]) barrier();

    if (read_from_vcont_buf[vcpu_id]) {
        indx = 0;
        while (gdb_vcont_buf[vcpu_id][indx++] != '#') count++;
        memcpy ((uint8_t *) rxbuf, (uint8_t *) gdb_vcont_buf[vcpu_id] + 1, count + 1);
        memset ((uint8_t *) gdb_vcont_buf[vcpu_id], 0, sizeof (gdb_vcont_buf[vcpu_id]));
        read_from_vcont_buf[vcpu_id] = 0;
        return count;
    }

    /* Copy the received packet(after checking it is available) from Tx shared 
       memory to internal buffer and clear Tx shared memory contents */
    while ((indx = get_fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0])) == FIFO_EMPTY) barrier();
    rxbuf[k] = *(gdb_rx_pkt[0] + indx);
    fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0]);
    
    /* If the packet does not start with the correct delimiter '$', drop the packet.
    * One instance where this happens is during the initial connection when the 
    * host gdb prefixes an ack character (as part of initial handshake?) to the 1st
    * packet it sends; we handle that here in that we bypass the ack character; 
    * but, second time onwards, we expect ack/nack character which is handled in gdb_putpacket() 
    */
    if (!(rxbuf[k] == '$' || rxbuf[k] == 3)) goto get_packet;
    if (rxbuf[k] == 3) return 1;

    k++;
    do {
        while ((indx = get_fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0])) == FIFO_EMPTY) barrier();
        rxbuf[k] = *(gdb_rx_pkt[0] + indx);
        fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0]);
    } while (rxbuf[k++] != '#');

    while ((indx = get_fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0])) == FIFO_EMPTY) barrier();
    rxbuf[k++] = *(gdb_rx_pkt[0] + indx);
    fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0]);

    while ((indx = get_fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0])) == FIFO_EMPTY) barrier();
    rxbuf[k++] = *(gdb_rx_pkt[0] + indx);
    fifo_remove ((struct fifo *)gdb_rx_pkt_fifo[0]);

    barrier();
    
    /* Calculate the checksum */
    while (rxbuf[count + 1] != '#')
        checksum += rxbuf[++count];
    
    /* Compare received checksum with calculated checksum and send +ve ack or -ve ack */
    rx_checksum = hex_to_uchar (rxbuf[count + 2]) << 4;
    rx_checksum += hex_to_uchar (rxbuf[count + 3]);
    ack = (checksum != rx_checksum) ? '-' : '+';
    
    /*printf("%d:ggs:ack=%c\n", vcpu_id, ack);*/

    if (gdb_ack_allowed[vcpu_id]) {

        /* TODO: Use gdb_pupacket_noack () */
        /* Update Tx shared memory with ack */
    
        while ((indx = get_fifo_insert ((struct fifo *)gdb_tx_pkt_fifo[0])) == FIFO_FULL) barrier();
        *(gdb_tx_pkt[0] + indx) = ack;
        fifo_insert ((struct fifo *)gdb_tx_pkt_fifo[0]);
    
        barrier ();
            
        /* Inform the tunneler thro' shared memory */
        inform_tunneler (1); /* message length is 1 */
        /*printf("\n%d:L%dFor'%c'\n", vcpu_id, 1, rxbuf[1]);*/
    }
    gdb_ack_allowed[vcpu_id] = 1;
    
    /* if +ve ack, strip the '$' from rxbuf */
    if (ack == '+')
        memmove ((uint8_t *) rxbuf, (uint8_t *) rxbuf + 1, count + 1);
    else
        goto get_packet;
    
    barrier ();

    /*printf("%u:leaving ggs\n", vcpu_id);*/
    return count;
}

void tx_to_gdb (int vcpu, int len)
{
    uint8_t temp[OS_NETOS_TX_BUF_SIZE];
    volatile int indx;
    int k;

    /* TODO: Allocate OS_NETOS_TX_BUF_SIZE+1 in shared memory and make this comparision as '>'. Do this everywhere */
    if (len >= OS_NETOS_TX_BUF_SIZE) {
#ifdef GDB_DEBUG_MSG
        printf ("packet length (%d) from vcpu %d exceeded maximum length\n",len, vcpu);
#endif
        return;
    }

    /* TODO: Merge the foll. with sending to UART */
    for(k = 0; k < len; k++) {
        while ((indx = get_fifo_remove ((struct fifo *)gdb_tx_pkt_fifo[0])) == FIFO_EMPTY) barrier ();
        temp[k] = *(gdb_tx_pkt[0] + indx);
        fifo_remove ((struct fifo *)gdb_tx_pkt_fifo[0]);
    }
    temp[k] = 0;

    /* Write onto the UART */
#ifdef GDB_DEBUG_MSG
    printf ("sending packet [%s] of length %d to host from vcpu %d\n", temp, len, vcpu);
#endif
    for (k = 0; k < len; k++)
        put_debug_char (temp[k]);
}

void handle_rx_from_gdb (void)
{
    int n; /* value received from UART1 */
    char outbyte;
    int i = -1;
    int indx;

    if ((n = get_debug_char()) == -1) return;
    outbyte = (int8_t) n;
    for (i = 0; i < TOTAL_VCPU; i++)
    if (*online_map & (1 << i)) {
        gdb_send_ipi (IRQ_STOP_GDB_IPI, 0, 0, (i & 3), (i >> 2));
        barrier ();
    }
    
    /* Update Rx shared memory */
    while ((indx = get_fifo_insert ((struct fifo *)gdb_rx_pkt_fifo[0])) == FIFO_FULL) barrier ();
    *(gdb_rx_pkt[0] + indx) = outbyte;
    fifo_insert ((struct fifo *)gdb_rx_pkt_fifo[0]);
}

void handle_tx_to_gdb (void)
{
    int indx;
    int len;

    if ((indx = get_fifo_remove ((struct fifo *)gdb_tx_pkt_len_fifo[0])) != FIFO_EMPTY) {
        len = *(gdb_tx_pkt_len[0] + indx);
        fifo_remove ((struct fifo *)gdb_tx_pkt_len_fifo[0]);
        tx_to_gdb (0, len);
    }
}

/* Function to extract a complete packet from shared memory */
static void gdb_handle_tx_rx (void)
{
#ifdef GDB_DEBUG_MSG
    printf("%u:entered ggm\n", TUNNELER_VCPU);
#endif

    /* If no netos vcpus are brought up, exit */
    if (!*online_map) {
        printf ("netos is not brought up on any vcpus, tunneler exits\n");
    }

    while (1) {
        handle_tx_to_gdb ();
        handle_rx_from_gdb ();
    }
}

/*
 * Function to read register contents.
 * Reads from register 'regnum', placing the results into hbuf.
 * Informs the caller of the number of bytes added to hbuf by adding to newhbuf.
 * Returns -1 if an illegal regnum is provided.

 * regnum is translated in the following order:
 * (taken from appendix D.5 in the gdb 6.4 manual)

 *    general purpose registers:
 *    R0, R1, R2, R3, R4, R5, R6, R7, R8, R9, R10, R11, R12, R13, R14, R15, R16,
 *    R17, R18, R19, R20, R21, R22, R23, R24, R25, R26, R27, R28, R29, R30, R31
 *
 *    special purpose registers: (lo, hi, COP0 registers)
 *    sr, lo, hi, bad, cause, epc (we cannot read pc, so we read epc)

 *    floating point registers:
 *    currently not implemented in xlr732-mips processor 
 */
static uint32_t
gdb_read_register(uint32_t regnum, uint8_t * hbuf, uint8_t ** newhbuf,
		  struct pt_regs *ptregs)
{
    uint64_t regval = 0;	/* register value */
    
    /* fetch the register value based on register number */
    switch (regnum) {
    
        /* read from general & special purpose registers */
        case 0: /* gpr R0 */
            regval = ptregs->regs[regnum];
            break;
        case 1: /* gpr R1 */
            regval = ptregs->regs[regnum];
            break;
        case 2: /* gpr R2 */
            regval = ptregs->regs[regnum];
            break;
        case 3: /* gpr R3 */
            regval = ptregs->regs[regnum];
            break;
        case 4: /* gpr R4 */
            regval = ptregs->regs[regnum];
            break;
        case 5: /* gpr R5 */
            regval = ptregs->regs[regnum];
            break;
        case 6: /* gpr R6 */
            regval = ptregs->regs[regnum];
            break;
        case 7: /* gpr R7 */
            regval = ptregs->regs[regnum];
            break;
        case 8: /* gpr R8 */
            regval = ptregs->regs[regnum];
            break;
        case 9: /* gpr R9 */
            regval = ptregs->regs[regnum];
            break;
        case 10: /* gpr R10 */
            regval = ptregs->regs[regnum];
            break;
        case 11: /* gpr R11 */
            regval = ptregs->regs[regnum];
            break;
        case 12: /* gpr R12 */
            regval = ptregs->regs[regnum];
            break;
        case 13: /* gpr R13 */
            regval = ptregs->regs[regnum];
            break;
        case 14: /* gpr R14 */
            regval = ptregs->regs[regnum];
            break;
        case 15: /* gpr R15 */
            regval = ptregs->regs[regnum];
            break;
        case 16: /* gpr R16 */
            regval = ptregs->regs[regnum];
            break;
        case 17: /* gpr R17 */
            regval = ptregs->regs[regnum];
            break;
        case 18: /* gpr R18 */
            regval = ptregs->regs[regnum];
            break;
        case 19: /* gpr R19 */
            regval = ptregs->regs[regnum];
            break;
        case 20: /* gpr R20 */
            regval = ptregs->regs[regnum];
            break;
        case 21: /* gpr R21 */
            regval = ptregs->regs[regnum];
            break;
        case 22: /* gpr R22 */
            regval = ptregs->regs[regnum];
            break;
        case 23: /* gpr R23 */
            regval = ptregs->regs[regnum];
            break;
        case 24: /* gpr R24 */
            regval = ptregs->regs[regnum];
            break;
        case 25: /* gpr R25 */
            regval = ptregs->regs[regnum];
            break;
        case 26: /* gpr R26 */
            regval = ptregs->regs[regnum];
            break;
        case 27: /* gpr R27 */
            regval = ptregs->regs[regnum];
            break;
        case 28: /* gpr R28 */
            regval = ptregs->regs[regnum];
            break;
        case 29: /* gpr R29 */
            regval = ptregs->regs[regnum];
            break;
        case 30: /* gpr R30 */
            regval = ptregs->regs[regnum];
            break;
        case 31: /* gpr R31 */
            regval = ptregs->regs[regnum];
            break;
        case 32: /* status */
            regval = ptregs->cp0_status;
            break;
        case 33: /* lo */
            regval = ptregs->lo;
            break;
        case 34: /* hi */
            regval = ptregs->hi;
            break;
        case 35: /* bad */
            regval = ptregs->cp0_badvaddr;
            break;
        case 36: /* cause */
            regval = ptregs->cp0_cause;
            break;
        case 37: /* epc */
            regval = ptregs->cp0_epc;
            break;
        
        /* read from floating point registers, and, fsr, fir, fp */
        case 38:
        case 39:
        case 40:
        case 41:
        case 42:
        case 43:
        case 44:
        case 45:
        case 46:
        case 47:
        case 48:
        case 49:
        case 50:
        case 51:
        case 52:
        case 53:
        case 54:
        case 55:
        case 56:
        case 57:
        case 58:
        case 59:
        case 60:
        case 61:
        case 62:
        case 63:
        case 64:
        case 65:
        case 66:
        case 67:
        case 68:
        case 69:
        case 70:
        case 71:
        case 72:
        
            /* currently xlr732-mips doesn't support floating point registers */
            regval = 0;
            break;
            
        default:
        
            return -1;
            break;
    }

    /* convert the result (regval) to hex format */
    ulonglong_to_hexbuf(regval, hbuf);
    
    /* add output length to newhbuf */
    if (newhbuf)
        *newhbuf += sizeof(uint64_t) * 2;
    
    return 0;
}

/* Function to write into register */
static uint32_t
gdb_write_register(uint32_t regnum, uint8_t * hbuf, uint8_t ** newhbuf,
		   struct pt_regs *ptregs)
{
    uint64_t regval;	/* register value */
    
    /* extract the regval, update newhbuf */
    regval = hexbuf_to_ulonglong(hbuf);
    if (newhbuf)
        *newhbuf += sizeof(uint64_t) * 2;
    
    /* based on register number, write into register */
    switch (regnum) {
    
        /* write to general & special purpose registers */
        case 0: /* gpr R0 */
            ptregs->regs[regnum] = regval;
            break;
        case 1: /* gpr R1 */
            ptregs->regs[regnum] = regval;
            break;
        case 2: /* gpr R2*/
            ptregs->regs[regnum] = regval;
            break;
        case 3: /* gpr R3 */
            ptregs->regs[regnum] = regval;
            break;
        case 4: /* gpr R4 */
            ptregs->regs[regnum] = regval;
            break;
        case 5: /* gpr R5 */
            ptregs->regs[regnum] = regval;
            break;
        case 6: /* gpr R6 */
            ptregs->regs[regnum] = regval;
            break;
        case 7: /* gpr R7 */
            ptregs->regs[regnum] = regval;
            break;
        case 8: /* gpr R8 */
            ptregs->regs[regnum] = regval;
            break;
        case 9: /* gpr R9 */
            ptregs->regs[regnum] = regval;
            break;
        case 10: /* gpr R10 */
            ptregs->regs[regnum] = regval;
            break;
        case 11: /* gpr R11 */
            ptregs->regs[regnum] = regval;
            break;
        case 12: /* gpr R12 */
            ptregs->regs[regnum] = regval;
            break;
        case 13: /* gpr R13 */
            ptregs->regs[regnum] = regval;
            break;
        case 14: /* gpr R14 */
            ptregs->regs[regnum] = regval;
            break;
        case 15: /* gpr R15 */
            ptregs->regs[regnum] = regval;
            break;
        case 16: /* gpr R16 */
            ptregs->regs[regnum] = regval;
            break;
        case 17: /* gpr R17 */
            ptregs->regs[regnum] = regval;
            break;
        case 18: /* gpr R18 */
            ptregs->regs[regnum] = regval;
            break;
        case 19: /* gpr R19 */
            ptregs->regs[regnum] = regval;
            break;
        case 20: /* gpr R20 */
            ptregs->regs[regnum] = regval;
            break;
        case 21: /* gpr R21 */
            ptregs->regs[regnum] = regval;
            break;
        case 22: /* gpr R22 */
            ptregs->regs[regnum] = regval;
            break;
        case 23: /* gpr R23 */
            ptregs->regs[regnum] = regval;
            break;
        case 24: /* gpr R24 */
            ptregs->regs[regnum] = regval;
            break;
        case 25: /* gpr R25 */
            ptregs->regs[regnum] = regval;
            break;
        case 26: /* gpr R26 */
            ptregs->regs[regnum] = regval;
            break;
        case 27: /* gpr R27 */
            ptregs->regs[regnum] = regval;
            break;
        case 28: /* gpr R28 */
            ptregs->regs[regnum] = regval;
            break;
        case 29: /* gpr R29 */
            ptregs->regs[regnum] = regval;
            break;
        case 30: /* gpr R30 */
            ptregs->regs[regnum] = regval;
            break;
        case 31: /* gpr R31 */
            ptregs->regs[regnum] = regval;
            break;
        case 32: /* status */
            ptregs->cp0_status = regval;
            break;
        case 33: /* lo */
            ptregs->lo = regval;
            break;
        case 34: /* hi */
            ptregs->hi = regval;
            break;
        case 35: /* bad */
            ptregs->cp0_badvaddr = regval;
            break;
        case 36: /* cause */
            ptregs->cp0_cause = regval;
            break;
        case 37: /* epc */
            ptregs->cp0_epc = regval;
            break;
        
        /* write to floating point registers, and, fsr, fir, fp */
        case 38:
        case 39:
        case 40:
        case 41:
        case 42:
        case 43:
        case 44:
        case 45:
        case 46:
        case 47:
        case 48:
        case 49:
        case 50:
        case 51:
        case 52:
        case 53:
        case 54:
        case 55:
        case 56:
        case 57:
        case 58:
        case 59:
        case 60:
        case 61:
        case 62:
        case 63:
        case 64:
        case 65:
        case 66:
        case 67:
        case 68:
        case 69:
        case 70:
        case 71:
        case 72:
        
        /* currently xlr732-mips doesn't support floating point registers */
        regval = 0;
            break;
            
        default:
        
            return -1;
            break;
        }
    
    return 0;
}

/* Function to read from memory at a specified address */
static void gdb_read_memory(uint64_t addr, uint32_t len, uint8_t * buf)
{
    volatile union mem_u mem;	/* value to be retrieved from addr */
    
    while (len) {
        
        /* Read in 64-bit units where we can (this is important if the target is a 
        * peripheral register or other memory location that only supports 64-bit 
        * accesses) 
        */
        /*
        * TODO: Handle TLB mapped address later -- presently we can dump memory only
        * from kseg0 or kseg1
        */
        while (len >= sizeof(uint64_t) && (addr % sizeof(uint64_t) == 0)) {
            
            mem.val64 = *(uint64_t *) (long)addr;
            
            *buf++ = uchar_to_hex(mem.byte[0].nibble0);
            *buf++ = uchar_to_hex(mem.byte[0].nibble1);
            *buf++ = uchar_to_hex(mem.byte[1].nibble0);
            *buf++ = uchar_to_hex(mem.byte[1].nibble1);
            *buf++ = uchar_to_hex(mem.byte[2].nibble0);
            *buf++ = uchar_to_hex(mem.byte[2].nibble1);
            *buf++ = uchar_to_hex(mem.byte[3].nibble0);
            *buf++ = uchar_to_hex(mem.byte[3].nibble1);
            *buf++ = uchar_to_hex(mem.byte[4].nibble0);
            *buf++ = uchar_to_hex(mem.byte[4].nibble1);
            *buf++ = uchar_to_hex(mem.byte[5].nibble0);
            *buf++ = uchar_to_hex(mem.byte[5].nibble1);
            *buf++ = uchar_to_hex(mem.byte[6].nibble0);
            *buf++ = uchar_to_hex(mem.byte[6].nibble1);
            *buf++ = uchar_to_hex(mem.byte[7].nibble0);
            *buf++ = uchar_to_hex(mem.byte[7].nibble1);
            
            addr += sizeof(uint64_t);
            len -= sizeof(uint64_t);
        }
        
        /*
        * Read in 32-bit units where we can (this is important if the target is a peripheral register or other memory
        * location that only supports 32-bit accesses) 
        */
        /*
        * TODO: Handle TLB mapped address later -- presently we can dump memory only from
        * kseg0 or kseg1
        */
        while (len >= sizeof(uint32_t) && (addr % sizeof(uint32_t) == 0)) {
            mem.val32 = *(uint32_t *) (long)addr;
            
            *buf++ = uchar_to_hex(mem.byte[0].nibble0);
            *buf++ = uchar_to_hex(mem.byte[0].nibble1);
            *buf++ = uchar_to_hex(mem.byte[1].nibble0);
            *buf++ = uchar_to_hex(mem.byte[1].nibble1);
            *buf++ = uchar_to_hex(mem.byte[2].nibble0);
            *buf++ = uchar_to_hex(mem.byte[2].nibble1);
            *buf++ = uchar_to_hex(mem.byte[3].nibble0);
            *buf++ = uchar_to_hex(mem.byte[3].nibble1);
            
            addr += sizeof(uint32_t);
            len -= sizeof(uint32_t);
        }
        
        /*
        * TODO: Handle TLB mapped address later -- presently we can dump memory only from
        * kseg0 or kseg1
        */
        /* Read in 16-bit units where we can (same reasons as above) */
        if (len >= sizeof(uint16_t) && (addr % sizeof(uint16_t) == 0)) {
            
            mem.val16 = *(uint16_t *) (long)addr;
            
            *buf++ = uchar_to_hex(mem.byte[0].nibble0);
            *buf++ = uchar_to_hex(mem.byte[0].nibble1);
            *buf++ = uchar_to_hex(mem.byte[1].nibble0);
            *buf++ = uchar_to_hex(mem.byte[1].nibble1);
            
            addr += sizeof(uint16_t);
            len -= sizeof(uint16_t);
        }
        
        /*
        * TODO: Handle TLB mapped address later -- presently we can dump memory only from
        * kseg0 or kseg1
        */
        if (len == 1 || (len && (addr % sizeof(uint16_t) != 0))) {
            
            /* request is totally misaligned; read a byte, and hope for the best */
            
            mem.val8 = *(uint8_t *) (long)addr;
            
            *buf++ = uchar_to_hex(mem.byte[0].nibble0);
            *buf++ = uchar_to_hex(mem.byte[0].nibble1);
            
            addr += sizeof(uint8_t);
            len -= sizeof(uint8_t);
        }
    }
    return;
}

/* Function to write to memory at a specified address */
static void gdb_write_memory(const uint8_t * buf, volatile uint64_t addr, uint32_t len)
{
    volatile uint64_t mem; /* value to be stored into addr */
    
    while (len) {
        /* Write in 64-bit units where we can (this is important if the target is a 
           peripheral register or other memory location that only supports 32-bit 
           accesses) */
        while (len >= sizeof(uint64_t) && addr % sizeof(uint64_t) == 0) {
            mem = hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            
            /**(uint64_t*)addr = mem; *//* didn't work */
            *(uint64_t *)(long)addr = mem;
            
            addr += sizeof (uint64_t);
            len -= sizeof (uint64_t);
        }
        
        /* Write in 32-bit units where we can (this is important if the target is a 
           peripheral register or other memory location that only supports 32-bit 
           accesses) */
        while (len >= sizeof(uint32_t) && addr % sizeof(uint32_t) == 0) {
            mem = hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            
            *(uint32_t *)(long)addr = mem;
            
            addr += sizeof (uint32_t);
            len -= sizeof (uint32_t);
        }
        
        /* Write in 16-bit units where we can (same reasons as above) */
        if (len >= sizeof (uint16_t) && addr % sizeof (uint16_t) == 0) {
            mem = hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            
            *(uint64_t *) (long)addr = mem;
            
            addr += sizeof(uint16_t);
            len -= sizeof(uint16_t);
        }
        
        if (len == 1 || (len && addr % sizeof(uint16_t) != 0)) {
            /* Request is totally misaligned; write a byte, and hope for the best */
            
            mem = hex_to_uchar(*buf++);
            mem = (mem << 4) + hex_to_uchar(*buf++);
            
            *(uint64_t *) (long)addr = mem;
            
            addr += sizeof(uint8_t);
            len -= sizeof(uint8_t);
        }
    }
    return;
}

/* Function to parse the zbreak (breakpoint) command (sent by gdb host) */
static void parse_zbreak (const uint8_t * pbuf, uint32_t * ptype, uint64_t * paddr, int *plen)
{
	/* Get the breakpoint type */
	*ptype = hex_to_uchar(*pbuf++);

	/* Get the breakpoint address */
	*paddr = 0;
	pbuf++;
	while (*pbuf != ',') {
		*paddr <<= 4;
		*paddr += hex_to_uchar(*pbuf++);
	}

	/* Get the breakpoint length */
	*plen = 0;
	pbuf++;
	while (*pbuf != '#') {
		*plen <<= 4;
		*plen += hex_to_uchar(*pbuf++);
	}

} /* parse_zbreak() */

/* Handlers specific initialization */
static void handler_init(void)
{
	/* Initialize the general exception handler */
	trap_init();

	/* Initialize the break handler */
	gdb_break_init();

	/* Initialize the trap handler */
	xlr_trap_init();
}

/* Master specific initialization */
static void gdbstub_init_primary (void)
{
    int id;

    /* Initialize TBH lock */
    spin_lock_init (tbh_lock);
    
    /* UART initialization */
    gdb_uart_init ();
    
    /* Initialize Tx and Rx data structures */

    for (id = 0; id < TOTAL_VCPU; id++) {
        gdb_tx_pkt_fifo[id] = (struct fifo *)((long)OS_NETOS_TXFIFO + 12 * id);
    
        gdb_tx_pkt[id] = (uint8_t *) ((long)OS_NETOS_TX_BUF_BASE + (long)OS_NETOS_TX_BUF_SIZE * id);
    
        gdb_tx_pkt_len_fifo[id] = (struct fifo *)((long)OS_NETOS_TX_PKTLEN_FIFO + 12 * id);
    
        gdb_tx_pkt_len[id] = (unsigned int *)((long)OS_NETOS_TX_PKTLEN + 4 * (long)OS_NETOS_TX_PKTLEN_NUMBER * id);
    
        gdb_rx_pkt_fifo[id] = (struct fifo *)((long)OS_NETOS_RXFIFO + 12 * id);
    
        gdb_rx_pkt[id] = (uint8_t *)((long)OS_NETOS_RX_BUF_BASE + (long)OS_NETOS_RX_BUF_SIZE * id);
    }

#if 0 /* TODO: for future */
    /* initialize the gdb online map with 0 */
    online_map = (uint32_t *)(long)(OS_NETOS_GDB_ONLINE_MAP);
    *online_map = 0;
#endif
    
    /* Initialize the cpu online map with 0 */
    *online_map = 0;
    
    /* Initialize lock init variable so that other VCPUs can proceed */
    *(uint32_t *)(long)(OS_NETOS_CPU_ONLINE_MAP_LOCK_INIT_DONE) = (LOCK_INIT_DONE);
    gdb_mdelay (2);
    barrier ();
}

/* Secondary specific initialization */
static void gdbstub_init_secondary (void)
{
    int vcpu_id = processor_id();
    volatile struct gdb_bpnode *z0 = NULL;
    volatile int bp;
    
    /* Initialize all global stuff */
    host_has_attached[vcpu_id] = 0; /* indicates if host has detached */
    gdb_initial_break[vcpu_id] = 1; /* indicates the initial break caused by gdbstub_init() */
    xlr_binary[vcpu_id] = 0; /* for later use */
    gdblock[vcpu_id] = 0;
    ctrlc_ipi[vcpu_id] = 0;
    gnext_addr[vcpu_id] = 0x0;
    
    /* Initialize Tx and Rx data structures and shared memory for this vcpu */
    gdb_tx_pkt_fifo[0] = (struct fifo *)((long)OS_NETOS_TXFIFO + 12 * 0);
    fifo_init ((struct fifo *)gdb_tx_pkt_fifo[0], OS_NETOS_TX_BUF_SIZE);

    gdb_tx_pkt[0] = (uint8_t *) ((long)OS_NETOS_TX_BUF_BASE + (long)OS_NETOS_TX_BUF_SIZE * 0);
    *(gdb_tx_pkt[0] + 0) = 0;

    gdb_tx_pkt_len_fifo[0] = (struct fifo *)((long)OS_NETOS_TX_PKTLEN_FIFO + 12 * 0);
    fifo_init ((struct fifo *)gdb_tx_pkt_len_fifo[0], OS_NETOS_TX_PKTLEN_NUMBER);

    gdb_tx_pkt_len[0] = (unsigned int *)((long)OS_NETOS_TX_PKTLEN + 4 * (long)OS_NETOS_TX_PKTLEN_NUMBER * 0);
    *(gdb_tx_pkt_len[0] + 0) = 0;

    gdb_rx_pkt_fifo[0] = (struct fifo *)((long)OS_NETOS_RXFIFO + 12 * 0);
    fifo_init ((struct fifo *)gdb_rx_pkt_fifo[0], OS_NETOS_RX_BUF_SIZE);

    gdb_rx_pkt[0] = (uint8_t *)((long)OS_NETOS_RX_BUF_BASE + (long)OS_NETOS_RX_BUF_SIZE * 0);
    *(gdb_rx_pkt[0] + 0) = 0;

    gdb_tx_pkt_fifo[vcpu_id] = (struct fifo *)((long)OS_NETOS_TXFIFO + 12 * vcpu_id);
    fifo_init ((struct fifo *)gdb_tx_pkt_fifo[vcpu_id], OS_NETOS_TX_BUF_SIZE);

    gdb_tx_pkt[vcpu_id] = (uint8_t *) ((long)OS_NETOS_TX_BUF_BASE + (long)OS_NETOS_TX_BUF_SIZE * vcpu_id);
    *(gdb_tx_pkt[vcpu_id] + 0) = 0;

    gdb_tx_pkt_len_fifo[vcpu_id] = (struct fifo *)((long)OS_NETOS_TX_PKTLEN_FIFO + 12 * vcpu_id);
    fifo_init ((struct fifo *)gdb_tx_pkt_len_fifo[vcpu_id], OS_NETOS_TX_PKTLEN_NUMBER);

    gdb_tx_pkt_len[vcpu_id] = (unsigned int *)((long)OS_NETOS_TX_PKTLEN + 4 * (long)OS_NETOS_TX_PKTLEN_NUMBER * vcpu_id);
    *(gdb_tx_pkt_len[vcpu_id] + 0) = 0;

    gdb_rx_pkt_fifo[vcpu_id] = (struct fifo *)((long)OS_NETOS_RXFIFO + 12 * vcpu_id);
    fifo_init ((struct fifo *)gdb_rx_pkt_fifo[vcpu_id], OS_NETOS_RX_BUF_SIZE);

    gdb_rx_pkt[vcpu_id] = (uint8_t *)((long)OS_NETOS_RX_BUF_BASE + (long)OS_NETOS_RX_BUF_SIZE * vcpu_id);
    *(gdb_rx_pkt[vcpu_id] + 0) = 0;

    barrier();
    
    /* Initialize the break-array for this vcpu */
    z0 = gdb_breakpoint[vcpu_id];
    for (bp = 0; bp < TOTAL_BREAKPOINTS; bp++) {
        z0[bp].addr = NULL; /* address at which to break */
        z0[bp].instr = BREAK_INSTR; /* original instruction at addr, which will be replaced by the break instruction (during breakpoint setting) */
        z0[bp].valid = 0;
        /*memset((void *)(z0[bp].fun_name), (int)'\0', 256); *//* TODO: check why this didn't work */
    }
    
    /* Initialize the single-step node for this vcpu */
    gdb_instr_buffer[vcpu_id].targetAddr = NULL;
    gdb_instr_buffer[vcpu_id].savedInstr = BREAK_INSTR;
    
    /* Set some flags */
    gdb_ack_allowed[vcpu_id] = 1;
    gdb_status_allowed[vcpu_id] = 1;
    memset ((uint8_t *) gdb_vcont_buf[vcpu_id], 0, sizeof (gdb_vcont_buf[vcpu_id]));
    gdb_vcont_buf_len[vcpu_id] = 0;
    read_from_vcont_buf[vcpu_id] = 0;
    gdb_stopped_vcpus[vcpu_id] = 0;
    
    /* Wait till master vcpu is up */
    while (*(uint32_t *)(long)(OS_NETOS_CPU_ONLINE_MAP_LOCK_INIT_DONE) != (LOCK_INIT_DONE))
        barrier();
    
    /* Add gdb and VCPU entry to online map */
    update_online_map (CPU_MAP, ADD_MAP_ENTRY, vcpu_id);
#if 0
    /* TODO: for future */
    update_online_map(gdb, add_entry);
#endif
}

/* Function to initialize and start the gdb stub */
void gdbstub_init (void)
{
    int vcpu_id = processor_id ();
    
    /* Check the connection mode and initial break options */
    if (gdb_connection_mode < 0 || gdb_initial_break_val < 0)
        printf ("%s: incorrect values for gdb_connection_mode/gdb_uart\n", __FUNCTION__);	
    
#ifdef GDB_DEBUG_MSG
    printf ("%d entered %s\n", vcpu_id, __FUNCTION__);
    printf ("%s: connection = %d, uart = %d, initial break = %d\n", __FUNCTION__, gdb_connection_mode, gdb_uart, gdb_initial_break_val);
#endif
    /* Initialize the various handlers */
    handler_init ();
    
    init_irq (); 
    
    /* If primary vcpu ... (applicable only in serial mode ) */
    if (vcpu_id == TUNNELER_VCPU) gdbstub_init_primary ();
    
    /* If secondary VCPU ... */
    else gdbstub_init_secondary ();
    
    /* Enable global interrupt */
    sti ();
    
    if (vcpu_id == TUNNELER_VCPU) gdb_handle_tx_rx ();

    /* Call break, if initial break is needed or if it is tunneler vcpu */
    if (gdb_initial_break_val) gdb_break ();
    
#ifdef GDB_DEBUG_MSG
    printf("%d leaving %s\n", vcpu_id, __FUNCTION__);
#endif
}

/* Function to remove gdb/cpu entry from the corresponding online map */
static void update_online_map(int map, int operation, int vcpu_id)
{

    /* Removing the cpu entry indicates that this vcpu has completed and cannot be 
       accesses again unless it is restarted. Removing the gdb entry indicates that 
       this gdb session is closed and this vcpu is out of the exception continuing 
       the 'main' program; This VCPU cannot be accessed again unless it is connected
       to by a new gdb session in which case the online_map will be updated by this 
       gdb entry, provided the VCPU is still running at that time. */
    if (map == GDB_MAP) {
#if 0 /* for future */
        online_map_lock = (spinlock_t *) (OS_NETOS_GDB_ONLINE_MAP_LOCK); /* get lock for online_map */
        online_map = (uint32_t *) (uint32_t) (OS_NETOS_GDB_ONLINE_MAP);	/* get online_map shared area */
#endif
    }
    if (map == CPU_MAP) {
        online_map_lock = (spinlock_t *)(long)(OS_NETOS_CPU_ONLINE_MAP_LOCK);
        /*online_map = (uint32_t *) (long) (OS_NETOS_CPU_ONLINE_MAP); *//* get online_map shared area */
    }
    spin_lock((spinlock_t *)online_map_lock);
    if (operation == ADD_MAP_ENTRY)
        *online_map |= (1 << vcpu_id);
    else
        *online_map &= ~(1 << vcpu_id);
    spin_unlock((spinlock_t *)online_map_lock);
}

/* Function called when a VCPU exits */
void gdbstub_exit(void)
{
    /* Remove gdb and vcpu entry from online map */
#if 0
    /* TODO: for future */
    update_online_map(gdb, remove_entry);
#endif
    update_online_map(CPU_MAP, REMOVE_MAP_ENTRY, processor_id());

    /* Inform the host that the target has exited */
    gdb_status('W', 15, NULL);
}

/* Function to identify the command sent thro' monitor by host gdb */
static int parse_cmd(const int8_t * buf)
{
    return -1;
}

/* Function to handle '?' pkt */
static void handle_questionmark(uint32_t sigval)
{
	gdb_status('T'/* S */, sigval, NULL);
}

/* Function to handle 'c' pkt */
static void handle_c(char *bufptr, struct pt_regs *ptregs)
{
	volatile uint64_t addr;	/* address of instruction/data */
	int vcpu_id = processor_id();

        /* Clear the previous single-step break, if any */
        if (gdb_instr_buffer[vcpu_id].targetAddr)
            undo_single_step ();

	/* Parse address, if it's there */
	addr = 0;
	while (*bufptr != '#') {
		addr <<= 4;
		addr += hex_to_uchar(*bufptr++);
	}

	/* There are three ways to jump to address 'addr'.
	   1) Casting 'addr' to a function pointer and calling the function
	   (check the bootloader's code for this syntax)
	   2) Through an unconditional jump instruction (in assembly)
	   3) Returning from gdb_monitor(), effectively giving the control
	   back to either gdb_exception() where EPC is updated with 'addr'
	   passed via the global variable gnext_addr, or int_handler()

	   The first, and second method (used by sh2 and given below in #if 0...#endif
	   block) do not ensure that control is passed back to either gdb_exception() or 
	   int_handler() from where gdb_monitor() is called.  Hence the third 
	   method is used here. Note that gdb_monitor() can be called from either 
	   gdb_exception(), in case a break instruction is hit, or from the
	   int_handler(),when it receives packet(s) from the remote gdb */

	/* second method */
#if 0
	if (valid_addr(addr)) {
		__asm__ __volatile__("j %0\t\n"::"r"(addr));
	}

	/* exit back to interrupted code */
#endif

	/* Third method */
	gnext_addr[vcpu_id] = addr;

	if (gdb_initial_break[vcpu_id]) {
		gdb_initial_break[vcpu_id] = 0;

		/* This is the first break-exception and caused by by gdbstub_init();
		 * now execution should start from the next address after EPC, 
		 * which is passed thro' gnext_addr
		 */
		addr = ptregs->cp0_epc;
		gnext_addr[vcpu_id] = addr + 4;	/* the next instruction is stored 4 bytes after addr */
	}
}

/* Function to handle 'C' pkt */
static void handle_C(char *bufptr, struct pt_regs *ptregs)
{
	volatile uint64_t addr;	/* address of instruction/data */
	uint32_t sigval;	/* extracted signal value */
	int vcpu_id = processor_id();

        /* Clear the previous single-step break, if any */
        if (gdb_instr_buffer[vcpu_id].targetAddr)
            undo_single_step ();

	/* Parse signal */
	sigval = 0;
	while (*bufptr != ';' && *bufptr != '#') {
		sigval <<= 4;
		sigval += hex_to_uchar(*bufptr++);
	}

	/* Parse address, if it's there */
	addr = 0;
	if (*bufptr == ';') {
		bufptr++;
		while (*bufptr != '#') {
			addr <<= 4;
			addr += hex_to_uchar(*bufptr++);
		}
	}

	/* There are three ways to jump to address 'addr'.
	   1) Casting 'addr' to a function pointer and calling the function
	   (check the bootloader's code for this syntax)
	   2) Through an unconditional jump instruction (in assembly)
	   3) Returning from gdb_monitor(), effectively giving the control
	   back to either gdb_exception() where EPC is updated with 'addr'
	   passed via the global variable gnext_addr, or int_handler()

	   The first, and second method (used by sh2 and given below in #if 0...#endif
	   block) do not ensure that control is passed back to either gdb_exception() or 
	   int_handler() from where gdb_monitor() is called.  Hence the third 
	   method is used here. Note that gdb_monitor() can be called from either 
	   gdb_exception(), in case a break instruction is hit, or from the
	   int_handler(),when it receives packet(s) from the remote gdb */

	/* Second method */
#if 0
	if (valid_addr(addr)) {
		__asm__ __volatile__("j %0\t\n"::"r"(addr));
	}

	/* Exit back to interrupted code */
#endif

	/* Third method */
	gnext_addr[vcpu_id] = addr;

	if (gdb_initial_break[vcpu_id]) {
		gdb_initial_break[vcpu_id] = 0;

		/*
		 * this is the first break-exception and caused by by gdbstub_init();
		 * now execution should start from the next address after EPC, 
		 * which is passed thro' gnext_addr
		 */
		addr = ptregs->cp0_epc;
		gnext_addr[vcpu_id] = addr + 4;	/* the next instruction is stored 4 bytes after addr */
	}
}

/* Function to handle 'D' pkt */
static void handle_D(void)
{
	int vcpu_id = processor_id();

	/* Indicate host has detached */
	host_has_attached[vcpu_id] = 0;

	gdb_putpacket((uint8_t *)"OK", 2);
}

/* Function to handle 'g' pkt */
static void handle_g(uint8_t * bufptr, struct pt_regs *ptregs)
{
	int vcpu_id = processor_id();
	int regnum;

	/* Read all the registers that gdb_read_register will give us */
	for (regnum = 0, bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];
	     gdb_read_register(regnum, bufptr, &bufptr, ptregs) != -1;
	     regnum++) ;

	gdb_putpacket((uint8_t *) gdb_rxtx_buf[vcpu_id],
		      bufptr - gdb_rxtx_buf[vcpu_id]);
}

/* Function to handle 'G' pkt */
static void handle_G (uint8_t * bufptr, struct pt_regs *ptregs)
{
	uint32_t regnum;

	/* Pass each register's value to gdb_write_register */
	for (regnum = 0; *bufptr != '#'; regnum++)
		gdb_write_register(regnum, bufptr, &bufptr, ptregs);

	gdb_putpacket((uint8_t *)"OK", 2);
}

/* Function to handle 'H' pkt */
static void handle_H (uint8_t * bufptr)
{
    /* Detenlmne the target vcpu:
    There are 3 sets of packets that detenlmne/alter the target vcpu, namely,

    1) set thread (H)
    -- sets the thread for subsequent operations (step 2 or 3 below)

    2) continue/step (c, C, s, S)
    -- should be applied only to the thread(s) set for these operations

    3) others (the rest of the above)
    -- should be applied only to the thread(s) set for these operations */
    
    int i;
    char tempbuf[8];
    int tempvcpu; 
        
    gdb_vcpu_c = gdb_vcpu_g = processor_id ();

    memset (tempbuf, 0, sizeof (tempbuf));
    for (i = 1; bufptr[i] != '#'; i++)
        tempbuf[i - 1] = bufptr[i];
    tempvcpu = atoi (tempbuf);
    if (tempvcpu >= TOTAL_VCPU)
        printf ("vcpu (thread) beyond the allowed (1-31) range\n");

    /* We consider only thread id's that are neither -1 nor 0; gdb sends -1 for all threads and 0 for any thread */
    else if (!(tempvcpu == -1 || tempvcpu == 0)) {
        if (bufptr[0] == 'c' || bufptr[0] == 'C' || bufptr[0] == 's' || bufptr[0] == 'S')
            gdb_vcpu_c = tempvcpu;
        else
            gdb_vcpu_g = tempvcpu;
    }

    /*printf("%u:gdb_vcpu_c=%u\n",processor_id(),gdb_vcpu_c);*/
    /*printf("%u:gdb_vcpu_g=%u\n",processor_id(),gdb_vcpu_g);*/

    gdb_putpacket((uint8_t *)"OK", 2);
}

/* Function to handle 'k' pkt */
static void handle_k(void)
{

	/* This flag should not be set, otherwise it'll not allow new session from
	 * the same gdb device
	 */
	/*host_has_attached[vcpu_id]  = 0; */

#if 0
	/* TODO: for future */
	/* remove gdb entry from online map */
	update_online_map(gdb, remove_entry);
#endif

#if 0
	/*
	 * In the case of serial connection, gdb_putpacket_noack("OK", 2) will 
	 * give a kernel panic in the host as the tty layer is already closed there; 
	 * In the case of network connection, gdb_putpacket_noack("OK", 2) will 
	 * give a write error as the socket is already closed on the host side; 
	 * hence we don't send any packet here
	 */

	gdb_putpacket_noack("OK", 2);
#endif

	/* Currently we don't kill the program, but allow it to continue */
	/*gdb_kill(); */
}

/* Function to handle 'm' pkt */
static void handle_m(uint8_t * bufptr)
{
	volatile uint64_t addr;	/* address of instruction/data */
	uint32_t len;
	int vcpu_id = processor_id();

	/* TODO: add bus error handling to gdb_handle_exception */

	/* Parse the address */
	addr = 0;
	while (*bufptr != ',') {
		addr <<= 4;
		addr += hex_to_uchar(*bufptr++);
	}

	/* Parse the length */
	len = 0;
	bufptr++;
	while (*bufptr != '#') {
		len <<= 4;
		len += hex_to_uchar(*bufptr++);
	}

	/* Read the memory (gdb will correctly interpret our byte ordering) */
	gdb_read_memory(addr, len, (uint8_t *) gdb_rxtx_buf[vcpu_id]);

	gdb_putpacket((uint8_t *) gdb_rxtx_buf[vcpu_id], len * 2);
}

/* Function to handle 'M' pkt */
static void handle_M(uint8_t * bufptr)
{
	volatile uint64_t addr;	/* address of instruction/data */
	uint32_t len;

	/* TODO: add bus error handling to gdb_handle_exception */

	/* Read address */
	addr = 0;
	while (*bufptr != ',') {
		addr <<= 4;
		addr += hex_to_uchar(*bufptr++);
	}

	/* Read length */
	len = 0;
	bufptr++;
	while (*bufptr != ':') {
		len <<= 4;
		len += hex_to_uchar(*bufptr++);
	}

	/* Write the memory (in the order in which gdb has sent) */
	gdb_write_memory(bufptr + 1, addr, len);

	gdb_putpacket((uint8_t *)"OK", 2);
}

/* Function to handle 'p' pkt */
static void handle_p(uint8_t * bufptr, struct pt_regs *ptregs)
{
	uint32_t regnum;
	int vcpu_id = processor_id();

	/* Parse register number */
	regnum = 0;
	while (*bufptr != '#') {
		regnum <<= 4;
		regnum += hex_to_uchar(*bufptr++);
	}

	/* Read the register regnum through gdb_read_register */
	bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];
	gdb_read_register(regnum, bufptr, &bufptr, ptregs);

	gdb_putpacket((uint8_t *) gdb_rxtx_buf[vcpu_id],
		      bufptr - gdb_rxtx_buf[vcpu_id]);
}

/* Function to handle 'P' pkt */
static void handle_P(uint8_t * bufptr, struct pt_regs *ptregs)
{
	uint32_t regnum;

	/* Parse register number */
	regnum = 0;
	while (*bufptr != '=') {
		regnum <<= 4;
		regnum += hex_to_uchar(*bufptr++);
	}

	/* Send the value */
	gdb_write_register(regnum, bufptr + 1, 0, ptregs);

	gdb_putpacket((uint8_t *)"OK", 2);
}

/* Respond to "R" query; this is a remote (ie. monitor) command, so give an appropriate reply */
static void handle_q_R(uint8_t * bufptr)
{
    uint32_t i = 0;
    uint32_t mon_cmd; /* Actual monitor command */
    uint8_t cmd_buf[256]; /* The extracted de-hexed command */
    uint8_t cmd_char[2]; /* Individual characters of the command ('2' stands for 2 hex bytes representing a character) */
    uint8_t *tempbuf; /* to hold the (monitor) output console packet */
    
    /* Go to the start of the command */
    bufptr += 5;
    tempbuf = bufptr;
    memset(cmd_buf, 0, sizeof(cmd_buf));
    
    /* Extract the monitor command from the gdb host packet */
    while (*bufptr != '#') {
        hexbuf_to_ucharbuf(cmd_char, bufptr, 1);
        cmd_buf[i++] = cmd_char[0];
        bufptr += 2; /* go past 2 hex bytes representing 1 actual character */
    }
    
    /* Convert the monitor command into an integer format so that it can be used in switch..case */
    mon_cmd = parse_cmd((signed char *)cmd_buf);

    switch (mon_cmd) {
        default: /* Send empty response */
            gdb_putpacket((uint8_t *)"", 0);
            break;
    }
}

static void handle_q_C(uint8_t * bufptr)
{
    /* Respond to "C" query ie. current active thread; currently we have not 
       implemented identifying the four hardware threads inside a vcpu; once 
       that is done, this will be modified to return the current hardware thread
    */
    int vcpu_id = processor_id();
    uint8_t tempbuf[3];

    /* Get the output buffer */
    bufptr = (uint8_t *)gdb_rxtx_buf[vcpu_id];

    strcpy ((char *)bufptr, "QC ");
    bufptr += strlen ("QC ");
    memset(tempbuf, 0, sizeof(tempbuf));
    sprintf ((char *)tempbuf, "%x", vcpu_id);
    strcpy ((char *)bufptr, (char *)tempbuf);
    bufptr += strlen ((char *)tempbuf);
    gdb_putpacket ((uint8_t *) gdb_rxtx_buf[vcpu_id], bufptr - gdb_rxtx_buf[vcpu_id]);
}

static void handle_q_O(uint8_t * bufptr)
{
    /* Respond to "O" query ie. Offsets */
    gdb_putpacket((uint8_t *)"Text=0;Data=0;Bss=0", 19);
}

static void handle_q_S(uint8_t * bufptr)
{
    /* Respond to "S" query ie. Symbol */
    /* TODO */
    /*gdb_putpacket((uint8_t *)"OK", 2);*/
    gdb_putpacket((uint8_t *)"", 0);
}

static void handle_q_f(uint8_t * bufptr)
{
    /* Respond to "f" query ie. first (thread) */

    int vcpu_id = processor_id();
    uint8_t tempbuf[256], i; /* TODO: change tempbuf to just 3 chars */
    memset(tempbuf, 0, sizeof(tempbuf));

    /* TODO: extract the command fully */

    /* Get the output buffer */
    bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];

    *bufptr++ = 'm';
    *bufptr++ = ' ';

    /* Extract the individual vcpu id's */
    for (i = 0; i < 32; i++)
        if (*online_map & (1 << i)) {
            /* TODO: sprintf & strcpy can both be merged into one */
            sprintf ((char *)tempbuf, "%x", i);
            strcpy ((char *)bufptr, (char *)tempbuf);
            bufptr = bufptr + strlen ((char *)tempbuf);
            *bufptr++ = ',';
        }
    *bufptr++ = 'l';

    gdb_putpacket ((uint8_t *) gdb_rxtx_buf[vcpu_id], bufptr - gdb_rxtx_buf[vcpu_id]);
}

static void handle_q_s(uint8_t * bufptr)
{
    /* Respond to "s" query ie. second (thread) */

    int vcpu_id = processor_id();

    /* TODO: extract the command fully */

    /* Get the output buffer */
    bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];

    /* Since we return all the thread id's in handle_q_f, we use this to just tenlmnate further queries from gdb */
    *bufptr++ = 'l';

    gdb_putpacket ((uint8_t *) gdb_rxtx_buf[vcpu_id], bufptr - gdb_rxtx_buf[vcpu_id]);
}

static void handle_q_T(uint8_t * bufptr)
{
    /* Respond to "T" query ie. further thread info */

    int vcpu_id = processor_id();
    uint8_t tempbuf[256];
    memset(tempbuf, 0, sizeof(tempbuf));

    /* TODO: extract the command fully, since thread arg is also passed */

    /* TODO: Give the correct status; This will replace vcpustate */
    sprintf ((char *)tempbuf, "running");

    /* Get the output buffer */
    bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];

    ucharbuf_to_hbuf(bufptr, tempbuf, strlen((char *)tempbuf));
    bufptr = bufptr + strlen((char *)tempbuf) * 2;

    gdb_putpacket ((uint8_t *) gdb_rxtx_buf[vcpu_id], bufptr - gdb_rxtx_buf[vcpu_id]);
}

/* Function to handle 'q' pkt */
static void handle_q(uint8_t * bufptr)
{
	/* Currently we are extracting only the first character after 'q'.
	 * TODO: extract the full string and act 
	 */
	/*if(bufptr[0] == 'R' && bufptr[1] == 'c' && bufptr[2] == 'm' && bufptr[3] == 'd' && bufptr[4] == ',') */
	if (*bufptr == 'R')
		handle_q_R(bufptr);
	if (*bufptr == 'C')
		handle_q_C(bufptr);
	if (*bufptr == 'O')
		handle_q_O(bufptr);
	if (*bufptr == 'S')
		handle_q_S(bufptr);
	if (*bufptr == 'f')
		handle_q_f(bufptr);
	if (*bufptr == 's')
		handle_q_s(bufptr);
	if (*bufptr == 'T')
		handle_q_T(bufptr);
}

/* Function to handle 's' pkt */
static void handle_s(uint8_t * bufptr, struct pt_regs *ptregs)
{
    volatile uint64_t addr; /* address of instruction/data */
    volatile uint64_t next_addr = 0; /* address of next instruction */
    int vcpu_id = processor_id();
    
    /* Clear the previous single-step break, if any */
    if (gdb_instr_buffer[vcpu_id].targetAddr)
        undo_single_step ();

    /* parse address, if it's there */
    addr = 0;
    while (*bufptr != '#') {
        addr <<= 4;
        addr += hex_to_uchar(*bufptr++);
    }
    
    /* The address of the next instruction may not be the one in sequence if the
       current instruction is a branch; Hence gdb will not know this address in 
       advance and it is the job of the stub to find this */
    next_addr = (addr == 0) ? get_target_addr (ptregs->cp0_epc, ptregs) : get_target_addr (addr, ptregs);
    
    /* Save the next address */
    gdb_instr_buffer[vcpu_id].targetAddr = (uint64_t *)(long)next_addr;

    /* Save the next instruction. If the next instruction is already break, find
       the original instruction; This can happen if some other vcpu has put a 
       single-step break, but unable to proceed */
    if (*(uint32_t *)(long)next_addr == BREAK_INSTR)
        gdb_instr_buffer[vcpu_id].savedInstr = get_orig_instr ((uint64_t *)(long)next_addr);
    else
        gdb_instr_buffer[vcpu_id].savedInstr = *(uint32_t *)(long)next_addr;

    /* Replace the instruction at next addr with break, flush i-cache and sync */
    *(uint32_t *)(long)next_addr = BREAK_INSTR;
    gdb_flush_icache_addr ((int64_t)(long)(next_addr));
    barrier();
    
    /*printf("%u:hs():Ins %x At %llx\n", processor_id(), *(uint32_t *)(long)next_addr, (int64_t)next_addr);*/

    /* This is executed only once */
    if (gdb_initial_break[vcpu_id]) {

        gdb_initial_break[vcpu_id] = 0;
        
        /* This vcpu is in the initial break now; Execution should start from the
           next address (gnext_addr) after EPC */
        gnext_addr[vcpu_id] = ptregs->cp0_epc + 4; /* the next instruction after EPC is: EPC + 4 */
    }
    
    /* Third method (explained in case 'c') */
    if (addr != 0)
        gnext_addr[vcpu_id] = addr;
}

/* Function to handle 'S' pkt */
static void handle_S(uint8_t * bufptr, struct pt_regs *ptregs)
{
    volatile uint64_t addr; /* address of instruction/data */
    volatile uint64_t next_addr = 0; /* address of next instruction */
    int vcpu_id = processor_id();
    uint32_t sigval; /* extracted signal value */
    
    /* Clear the previous single-step break, if any */
    if (gdb_instr_buffer[vcpu_id].targetAddr)
        undo_single_step ();

    /* Parse signal */
    sigval = 0;
    while (*bufptr != ';' && *bufptr != '#') {
        sigval <<= 4;
        sigval += hex_to_uchar(*bufptr++);
    }
    
    /* Parse address, if it's there */
    addr = 0;
    if (*bufptr == ';') {
        bufptr++;
        while (*bufptr != '#') {
            addr <<= 4;
            addr += hex_to_uchar(*bufptr++);
        }
    }
    
    /* The address of the next instruction may not be the one in sequence if the
       current instruction is a branch; Hence gdb will not know this address in 
       advance and it is the job of the stub to find this */
    next_addr = (addr == 0) ? get_target_addr (ptregs->cp0_epc, ptregs) : get_target_addr (addr, ptregs);
    
    /* Store the next address */
    gdb_instr_buffer[vcpu_id].targetAddr = (uint64_t *)(long)next_addr;

    /* Store the next instruction. If the next instruction is already break, find
       the original instruction; This can happen if some other vcpu has put a 
       single-step break, but unable to proceed */
    if (*(uint32_t *)(long)next_addr == BREAK_INSTR)
        gdb_instr_buffer[vcpu_id].savedInstr = get_orig_instr ((uint64_t *)(long)next_addr);
    else
        gdb_instr_buffer[vcpu_id].savedInstr = *(uint32_t *)(long)next_addr;

    /* Replace the instruction at next addr with break, flush i-cache and sync */
    *(uint32_t *)(long)next_addr = BREAK_INSTR;
    gdb_flush_icache_addr ((int64_t)(long)(next_addr));
    barrier();

    /*printf("%u:hs():Ins %x At %llx\n", processor_id(), *(uint32_t *)(long)next_addr, (int64_t)next_addr);*/

    
    /* This is executed only once */
    if (gdb_initial_break[vcpu_id]) {

        gdb_initial_break[vcpu_id] = 0;
    
        /*
         * This vcpu is in the initial break now; Execution should start from the
         * next address (gnext_addr) after EPC
         */
        gnext_addr[vcpu_id] = ptregs->cp0_epc + 4; /* the next instruction after EPC is: EPC + 4 */
    }
    
    /* Third method (explained in case 'c') */
    if (addr != 0)
        gnext_addr[vcpu_id] = addr;
}

/* Function to handle 'W' pkt */
static void handle_W(void)
{
    gdb_putpacket((uint8_t *)"OK", 2);
}

/* Function to handle 'X' pkt */
static void handle_X(void)
{
    int vcpu_id = processor_id();
    
    xlr_binary[vcpu_id] = 1;
    
    gdb_putpacket((uint8_t *)"", 0);
}

/* Function to handle 'Z' pkt */
static void handle_Z(uint8_t * bufptr)
{
    volatile uint64_t addr; /* address of instruction/data */
    uint32_t len;
    uint32_t type;
    /*int vcpu_id = processor_id();*/
    int bp;
    volatile uint8_t bp_set = 0;
    volatile struct gdb_bpnode * z0 = NULL; /* now multiple breakpoint setting is supported */
    
    /* Parse the zbreak (breakpoint) command (sent by the host) */
    parse_zbreak(bufptr, &type, (uint64_t *) & addr, (int *)&len);
    
#if 0
    /* if parse_zbreak() has any return value, then this is useful */
    if (!ret) {
        gdb_putpacket("E01", 3);
        break;
    }
#endif
    
    /* check the breakpoint type:
    0 --> software breakpoint
    1 --> hardware breakpoint
    2 --> write watchpoint
    3 --> read watchpoint
    4 --> access watchpoint
    */
    
    if (type != 0) {
    
        /* We support only software break points so far; send error response */
        gdb_putpacket((uint8_t *)"E02", 3);
        return;
    }
#if 0
    /* TODO: check R_SZ value from rtems code / gdb manual (not a priority item) */
    if (len != R_SZ) {	/* was 1 */
    gdb_putpacket("E03", 3);
    return;
    }
#endif
    
    /* Let us check whether this break point is already set */

    z0 = gdb_breakpoint[0];
    for (bp = 0; bp < TOTAL_BREAKPOINTS; bp++) {

        if (!(z0[bp].addr)) break;
        if ((z0[bp].addr == (uint64_t *)(long)addr) && z0[bp].valid) {
            bp_set = 1;
            break;
        }

    }

    if (bp_set) {

        /* We already have a breakpoint for this address. But sending error 
           response crashes gdb to seg fault */
        bp_set = 0;
        gdb_putpacket((uint8_t *)"OK", 2);
        return;

    }
    
    if (bp == TOTAL_BREAKPOINTS) {
    
        /* Memory could not be allocated; send error response */
        /* No guarantee that gdb recovers from this; gdb may have to re-started
         */
        gdb_putpacket((uint8_t *)"E05", 3);
        return;
    }
    
    /* Store the breakpoint address */
    z0[bp].addr = (uint64_t *)(long)addr;
    
    /* Store the instruction at the breakpoint address. If that instruction is 
       already break, find the original instruction; This can happen if some 
       other vcpu has put a single-step break, but unable to proceed */
    if (*(uint32_t *)(long)addr == BREAK_INSTR)
        z0[bp].instr = get_orig_instr ((uint64_t *)(long)addr);
    else
        z0[bp].instr = *(uint32_t *)(long)addr;

    /* Insert break at the breakpoint address, flush i-cache, and sync */
    *(uint32_t *)(z0[bp].addr) = BREAK_INSTR;
    gdb_flush_icache_addr ((int64_t)(long)(z0[bp].addr));
    barrier();
    
    /* Make this breakpoint valid */
    z0[bp].valid = 1;

    /*printf("%u:Z:Ins %x At %llx\n", processor_id(), *(uint32_t *)(z0[bp].addr), (int64_t)(long)(z0[bp].addr));*/
    print_listinfo((char *)__FUNCTION__, "Z");
    
    gdb_putpacket((uint8_t *)"OK", 2);
}

/* Function to handle 'z' pkt */
static void handle_z(uint8_t * bufptr)
{
    volatile uint64_t addr;	/* address of instruction/data */
    uint32_t len;
    uint32_t type;
    /*int vcpu_id = processor_id();*/
    int bp;
    volatile uint8_t bp_set = 0;
    volatile struct gdb_bpnode *z0 = NULL; /* now multiple breakpoint setting is supported */
    
    /* Parse the zbreak (breakpoint) command (sent by the host) */
    parse_zbreak(bufptr, &type, (uint64_t *) & addr, (int *)&len);
    
#if 0
    /* if parse_zbreak() has any return value, then this is useful */
    if (!ret) {
        gdb_putpacket("E01", 3);
        pu_idreak;
    }
#endif
    
    /* check the breakpoint type:
    0 --> software breakpoint
    1 --> hardware breakpoint
    2 --> write watchpoint
    3 --> read watchpoint
    4 --> access watchpoint
    */
    
    if (type != 0) {
    
        /* We support only software break points so far; send error response */
        gdb_putpacket((uint8_t *)"E01", 3);
        return;
    }
#if 0
    /* TODO: check R_SZ value from rtems code / gdb manual (not a priority item) */
    if (len != R_SZ) {	/* was 1 */
        gdb_putpacket("E02", 3);
        return;
    }
#endif
    
    /* Let us check whether this break point is already set */

    z0 = gdb_breakpoint[0];
    for (bp = 0; bp < TOTAL_BREAKPOINTS; bp++) {
        if (!(z0[bp].addr)) break;
        if ((z0[bp].addr == (uint64_t *)(long)addr) && z0[bp].valid) {
            bp_set = 1;
            break;
        }
    }

    if (!(z0[bp].addr) || !bp_set) {
        /* We don't have a breakpoint for this address. But sending error 
           response crashes gdb to seg fault */
        gdb_putpacket((uint8_t *)"OK", 2);
        return;
    }

    if (bp == TOTAL_BREAKPOINTS) {
        /* Breakpoints limit is reached, send error response */
        /* No guarantee that gdb recovers from this; gdb may have to re-started
         */
        gdb_putpacket((uint8_t *)"E04", 3);
        return;
    }
    
    bp_set = 0;
    
    /* We have a breakpoint for this address, so remove it. Put the old 
       instruction back, flush i-cache, and sync */
    *(uint32_t *) (long)addr = z0[bp].instr;
    gdb_flush_icache_addr ((int64_t)(long)(addr));
    barrier();
    
    /* Invalidate the node in the list */
    /* TODO: Invalidating the addr member should be enough */
    z0[bp].instr = BREAK_INSTR;
    z0[bp].valid = 0;
    z0[bp].addr = NULL;
    while (bp < (TOTAL_BREAKPOINTS - 1)) {
        z0[bp] = z0[bp + 1];
        bp++;
    }
    z0[bp].instr = BREAK_INSTR;
    z0[bp].valid = 0;
    z0[bp].addr = NULL;
        
    /*printf("%u:gm():'z':Ins %x AtAddr %llx\n", processor_id(), *(uint32_t *)(long)addr, (int64_t)addr);*/
    print_listinfo((char *)__FUNCTION__, "z");
    
    gdb_putpacket((uint8_t *)"OK", 2);
}

/* Function to handle 'T' pkt */
static void handle_T(uint8_t * bufptr)
{
    uint32_t thread = 0; /* extracted thread */
    
    /* parse thread */
    while (*bufptr != '#') {
        thread <<= 4;
        thread += hex_to_uchar(*bufptr++);
    }

    /* if thread is alive, send positive response */
    if (*online_map & (1 << thread))
	gdb_putpacket((uint8_t *)"OK", 2);
    else
	gdb_putpacket((uint8_t *)"E00", 3);
}

/* Function to handle 'vC' pkt */
static void handle_v_C(uint8_t * bufptr)
{
    int vcpu_id = processor_id();

    if (strncmp ((char *)bufptr, "Cont", 4) == 0) {

        bufptr += 4;

        if (*bufptr == '?') {

            /* Reply with supported commands */
            char * tempbuf = "vCont;c;C;s;S;t;T";

            /* Get the output buffer */
            bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];

            strcpy ((char *)bufptr, tempbuf);
            bufptr += strlen ((char *)tempbuf);

            gdb_putpacket ((uint8_t *) gdb_rxtx_buf[vcpu_id], bufptr - gdb_rxtx_buf[vcpu_id]);
        }

        else if (*bufptr++ == ';') {

            /* Take appropriate actions on various threads */
            int target_vcpu_id;
            int vcpu;
            char temp_buf[OS_NETOS_TX_BUF_SIZE]; /* entire pkt (till '#') sent by gdb */
            char * str = NULL;
            char * str0 = NULL;
            char temp_cmd_with_thr[8]; /* each (command + thread id) in temp_buf */
            char temp_thr[8];
            char cmd[TOTAL_VCPU][8];
            char default_cmd[8]; /* command with no thread id specified */
            uint8_t checksum;
            int indx = 0;
            int k;

            memset (cmd, 0, sizeof (cmd));
            memset (default_cmd, 0, sizeof (default_cmd));
            memset (temp_buf, 0, sizeof (temp_buf));
            strncpy ((char *)temp_buf, (char *)bufptr, (uint8_t *)strstr((char *)bufptr, "#") - bufptr);
            str = strtok ((char *)temp_buf, ";");
            while (str) {
                target_vcpu_id = -1;
                memset (temp_cmd_with_thr, 0, sizeof (temp_cmd_with_thr));

                /* Copy the entire token (command + thread id ) */
                strcpy ((char *)temp_cmd_with_thr, str);

                /* If no delimiter ':' is found, then the entire token is the command; else extract the thread id from the token */
                str0 = strstr ((char *)temp_cmd_with_thr, ":"); 
                if (str0) {

                    /* Extract the thread id */
                    memset (temp_thr, 0, sizeof (temp_thr));
                    strcpy (temp_thr, str0 + 1);
                    target_vcpu_id = hexbuf_to_uint ((unsigned char *) temp_thr);

                    /* Remove the thread id from cmd */
                    *str0 = '\0'; 
                }

                /*printf("cmd=%s\n",temp_cmd_with_thr);*/
                /*printf("target_vcpu_id=%d\n",target_vcpu_id);*/

                if (target_vcpu_id != -1)

                    /* This command is for a particular thread */
                    strcpy (cmd[target_vcpu_id], temp_cmd_with_thr);

                else {

                    /* This command is for all the threads which do not have any action specified;
                       note the command and update the threads later */ 
                    memset (default_cmd, 0, sizeof (default_cmd));
                    strcpy (default_cmd, temp_cmd_with_thr);
                }

                /* Get the next token */
                str = strtok (NULL, ";");
            }

            /*printf ("default cmd = %s\n", default_cmd);*/

            /* Update vcpus that do not have any action specified, with default command */
            for (vcpu = 0; vcpu < TOTAL_VCPU; vcpu++)
                if ((*online_map & (1 << vcpu)) && (strcmp (cmd[vcpu], "\0") == 0))
                        strcpy (cmd[vcpu], default_cmd);

#ifdef GDB_DEBUG_MSG
            for (vcpu = 0; vcpu < TOTAL_VCPU; vcpu++)
                if (*online_map & (1 << vcpu))
                    printf ("cmd[%d] = %s\n", vcpu, cmd[vcpu]);
#endif

            /* Find out which vcpus have received command thro' vCont */
            for (vcpu = 0; vcpu < TOTAL_VCPU; vcpu++)
                if ((*online_map & (1 << vcpu)) && (strcmp (cmd[vcpu], "\0") != 0))
                    gdb_vcont_vcpus |= (1 << vcpu);

            /*printf("%d:vCont_vcpus = 0x%x\n", vcpu_id, gdb_vcont_vcpus);*/

            /* Insert commands (with RSP) into shared memory of the vCont vcpus */
            for (vcpu = 0; vcpu < TOTAL_VCPU; vcpu++)
                if ((*online_map & (1 << vcpu)) && (gdb_vcont_vcpus & (1 << vcpu))) {

                    /* The individual vcpus should not ack these commands, as the ack for all these commands has already been sent as ack for vCont */
                    gdb_ack_allowed[vcpu] = 0;
                    barrier ();
                    checksum = 0;

                    indx = 0;
                    gdb_vcont_buf[vcpu][indx++] = '$';
                    for (k = 0; k < strlen (cmd[vcpu]); k++) {
        
                        gdb_vcont_buf[vcpu][indx++] = cmd[vcpu][k];
        
                        /* Calculate the checksum */
                        checksum += cmd[vcpu][k];
                    }
        
                    gdb_vcont_buf[vcpu][indx++] = '#';
        
                    gdb_vcont_buf[vcpu][indx++] = uchar_to_hex (checksum >> 4);
        
                    gdb_vcont_buf[vcpu][indx++] = uchar_to_hex (checksum & 0xf);

                    read_from_vcont_buf[vcpu] = 1;

                    barrier ();
                }
        }
    }
}

/* Function to handle 'v' pkt */
static void handle_v(uint8_t * bufptr)
{
   if (*bufptr == 'C')
       handle_v_C (bufptr); 
}

void sync_with_normal_case (void)
{
    int vcpu_id = processor_id();

    write_phnx_eirr ((uint64_t) 1 << IRQ_BA_GDB_IPI);
    write_phnx_eirr ((uint64_t) 1 << IRQ_STOP_GDB_IPI);
    gdb_status_allowed[vcpu_id] = 1;
    gdb_vcont_vcpus &= ~(1 << vcpu_id);
    memset ((uint8_t *) gdb_vcont_buf[vcpu_id], 0, sizeof (gdb_vcont_buf[vcpu_id]));
    read_from_vcont_buf[vcpu_id] = 0;

    gdb_unlock (vcpu_id);
    
    gdb_stopped_vcpus[vcpu_id] = 0;
    
    /* Wait for other vcpus to come out */
    while (gdb_vcont_vcpus) barrier ();
}

int handle_command_to_thread (uint8_t *bufptr)
{
    int vcpu_id = processor_id();
    char pkt_type = 0;
    int target_vcpu = -1; 

    pkt_type = (*bufptr == 'c' || *bufptr == 'C' || *bufptr == 's' || *bufptr == 'S') ? 'c' : 'g'; 
    if ((pkt_type == 'c' && gdb_vcpu_c == vcpu_id) || (pkt_type == 'g' && gdb_vcpu_g == vcpu_id)) return 0;

    /* Copy pkt to target vcpu */

    target_vcpu = (pkt_type == 'c') ? gdb_vcpu_c : gdb_vcpu_g;
    gdb_vcont_buf[target_vcpu][0] = '$';
    memcpy ((uint8_t *) gdb_vcont_buf[target_vcpu] + 1, (uint8_t *) gdb_rxtx_buf[vcpu_id], sizeof (gdb_vcont_buf[target_vcpu]));
    barrier ();
    read_from_vcont_buf[target_vcpu] = 1;
    /* While 'c' pkt does not elicit response from target vcpu, 'g' pkt does */
    if (pkt_type == 'g') {
        memset ((uint8_t *) bufptr, 0, sizeof (gdb_rxtx_buf[vcpu_id]));
        while (!read_from_vcont_buf[vcpu_id]) barrier ();
        memcpy ((uint8_t *) bufptr, (uint8_t *) gdb_vcont_buf[vcpu_id], gdb_vcont_buf_len[vcpu_id]);
        gdb_putpacket ((uint8_t *)bufptr, gdb_vcont_buf_len[vcpu_id]);
        read_from_vcont_buf[vcpu_id] = 0;
    }
    return 1;
}
 
int gdb_packet_processor (int sigval, uint8_t *bufptr, struct pt_regs *ptregs)
{
    int vcpu_id = processor_id ();

    gdb_lock(vcpu_id);
        switch (*bufptr++) {
            
            case 3: /* ^c: stop other vcpus */
                if (!gdb_status_allowed[vcpu_id]) break;
                gdb_ack_allowed[vcpu_id] = 1;
                if (gdb_status_allowed[vcpu_id])
                    gdb_status ('T', 15, ptregs);
                break;
    
            case '?': /* Send last signal */
                handle_questionmark (sigval);
                break;
            
            case 'c': /* Continue (address optional) */
                handle_c ((char *)bufptr, ptregs);
                sync_with_normal_case (); /* gdb_unlock happens inside this */
                return 1;
            
            case 'C': /* Continue with signal (address optional) */
                handle_C ((char *)bufptr, ptregs);
                sync_with_normal_case (); /* gdb_unlock happens inside this */
                return 1;
            
            /* TODO: Find the use for this; not a priority item, to be done later */
            case 'd': /* Toggle debug flag */
                break;
            
            case 'D': /* Remote system is detaching - return OK and exit from debugger */
                handle_D ();
                return 1;
            
            case 'g': /* Read register values */
                handle_g (bufptr, ptregs);
                break;
            
            case 'G': /* Write registers */
                handle_G (bufptr, ptregs);
                break;
            
            case 'H': /* Set thread */
                handle_H (bufptr);
                break;
            
            case 'k': /* Kill program */
                handle_k ();
                break;
            
            case 'm': /* Read memory */
                handle_m (bufptr);
                break;
            
            case 'M': /* Write memory */
                handle_M (bufptr);
                break;
            
            /* TODO: Currently this is not used */
            case 'O': /* Output to console (ignore, as we have no console) */
                break;
            
            case 'p': /* Read register */
                handle_p (bufptr, ptregs);
                break;
            
            /* TODO: check this case 'P' on XLR */
            case 'P': /* write register */
                handle_P (bufptr, ptregs);
                break;
            
            case 'q': /* Query */
                handle_q (bufptr);
                break;
            
            case 's': /* Step (address optional) */
                handle_s (bufptr, ptregs);
                sync_with_normal_case (); /* gdb_unlock happens inside this */
                return 1;
            
            case 'S': /* Step with signal (address optional) */
                handle_S (bufptr, ptregs);
                sync_with_normal_case (); /* gdb_unlock happens inside this */
                return 1;
    
            case 'T': /* Thread id */
                handle_T (bufptr);
                break;
            
            case 'v': /* Verbose ie. verbose-resume, verbose-file etc. */
                handle_v (bufptr);
                break;
    
            case 'W': /* MAA..AA,LLLL:  Write LLLL bytes at address AA..AA - return OK */
                handle_W ();
                break;
            
            case 'X': /* XAA..AA,LLLL:<binary data>#cs */
                handle_X ();
                break;
            
            case 'Z': /* Add breakpoint */
                handle_Z (bufptr);
                break;
            
            case 'z': /* Remove breakpoint */
                handle_z (bufptr);
                break;
            
            default:
                /* Unknown command -- send empty response as per gdb spec */
                gdb_putpacket ((uint8_t *)"", 0);
                break;
        }
    gdb_unlock (vcpu_id);
    return 0;
}

/* The gdb command processor */
static void gdb_monitor (int sigval, struct pt_regs *ptregs)
{
    uint8_t *bufptr;
    int vcpu_id = processor_id();
    volatile int cnt = 0;

    /* Continuously poll for host command and take action; control will go out 
       of the while loop only when the host has either detached, or issued 
       continue or step commands */
    while (1) {
#ifdef GDB_DEBUG_MSG
        printf("%d:gm:while\n", vcpu_id);
#endif
        
        /* Find out what the host wants us to do */
        while (!(cnt = gdb_getpacket_secondary ((uint8_t *) gdb_rxtx_buf[vcpu_id], sizeof (gdb_rxtx_buf[vcpu_id])))) ;

#ifdef GDB_DEBUG_MSG
        /*printf("\n%d:gm:RP'%s'\n", vcpu_id, gdb_rxtx_buf[vcpu_id]);*/
        printf("\n%d:gm:RP'%c'\n", vcpu_id, gdb_rxtx_buf[vcpu_id][0]);
#endif
        
        if (!host_has_attached[vcpu_id])
            host_has_attached[vcpu_id] = 1;
        
        bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];
        
        /* Handle command to vcpu based on previous set-thread ('H') pkt */
        if (vcpu_id == gdb_current_vcpu && *bufptr != 'H')
            if (handle_command_to_thread (bufptr)) continue;

        /* Depending on the pkt, we may have to exit expt/intr handler */
        if (gdb_packet_processor (sigval, bufptr, ptregs)) return;
    }
}

/* Translates our exception number into something that gdb will recognize  */
static int gdb_sigval (int exc_code)
{
    int retval = 0;
    
    /* TODO: Take care of other exceptions in future */
    /* TODO: Give meaningful names to the signal numbers */
    switch (exc_code) {
    
        case 32:
            /* SIGSOMETHING */
            retval = 5;
            break;
        
        case 13:
            /* SIGTRAP */
            retval = 5;
            break;
        
        case 9:
            /* SIGBREAK */
            retval = 5;
            break;
        
        case 0:
            /* SIGTERM */
            retval = 15;
            break;
        
        default:
            retval = 0;
            break;
    }
    
    return retval;
}

/* Function which gives the current status of the program execution to the gdb host */
static void gdb_status(uint8_t mode, uint16_t sigval, struct pt_regs *ptregs)
{
    int vcpu_id = processor_id();
    uint8_t tx_buf[OS_NETOS_TX_BUF_SIZE];
    /*uint8_t regval[16];*/ /* the index 16 indicates the no. of hex bytes required to represent 8-byte ie. 64-bit value */
    uint8_t *bufptr = tx_buf;
    /*uint16_t regnum;*/ /* register number */
    uint8_t tempbuf[3];
    
    /*printf("%u entered %s\n", processor_id(), __FUNCTION__); */
    /*printf("host_has_attached=%d\n", host_has_attached[processor_id()]);*/

    /* If host is not attached, don't proceed further */
    if (!host_has_attached[vcpu_id]) return;
    
    /*
    Currently four modes are supported. 
    'S': signal mode
    'W': exit mode
    'X': tenlmnation mode
    'T': thread mode
    */
    /* Inform the host the mode that the target is in */
    *bufptr++ = mode;
    
    /* Send the signal number */
    *bufptr++ = uchar_to_hex(sigval >> 4);
    *bufptr++ = uchar_to_hex(sigval & 0xf);
    
    if (mode == 'T') {

#if 0
        /* Read all the registers that gdb_read_register will give us */
        for (regnum = 0; gdb_read_register(regnum, regval, 0, ptregs) != -1; regnum++) {
            
            /* Stuff in the most-significant nibble of the register number, if nonzero */
            if (regnum & 0xf0)
                *bufptr++ = uchar_to_hex(regnum >> 4);
            
            /* Send the least-significant nibble of regnum */
            *bufptr++ = uchar_to_hex(regnum);
            *bufptr++ = ':';
            
            /* Send the register value */
            memcpy(bufptr, regval, sizeof(regval));
            bufptr += sizeof(regval);
            
            /* Tenlmnate field */
            *bufptr++ = ';';
            
            /* Don't overrun tx_buf */
            if (bufptr > (tx_buf + sizeof(tx_buf)) - 14)
                break;
        }
#endif
        /*strcpy (bufptr, "thread:");*/
        strcpy ((char *)bufptr, "thread:");
        bufptr += strlen ("thread:");
        memset (tempbuf, 0, sizeof (tempbuf));
        sprintf ((char *)tempbuf, "%x", vcpu_id);
        strcpy ((char *)bufptr, (char *)tempbuf);
        bufptr += strlen ((char *)tempbuf);
        *bufptr++ = ';';
    }
    
    /* if(mode == 'T') */
    gdb_putpacket (tx_buf, bufptr - tx_buf);
    /*printf("%u leaving %s\n", processor_id(), __FUNCTION__); */
}

void sync_with_tbh_case (void)
{
    int vcpu_id = processor_id();
    int id;

    gdb_started_vcpus[vcpu_id] = 0;
    gdb_stopped_vcpus[vcpu_id] = 1;
    
    /* Wait for other vcpus to stop */
    for (id = 0; id < TOTAL_VCPU; id++)
        if ((id != vcpu_id) && (*online_map & (1 << id)))
            while (!gdb_stopped_vcpus[id]) barrier ();
    
    /* Wait for TBH to unset */ 
    while (tbh_vcpu != -1) barrier ();
    gdb_status_allowed[vcpu_id] = 1;
    write_phnx_eirr ((uint64_t) 1 << IRQ_BA_GDB_IPI);
    
    gdb_stopped_vcpus[vcpu_id] = 0;
    gdb_started_vcpus[vcpu_id] = 1;
    
    /* Wait for other vcpus to start */
    for (id = 0; id < TOTAL_VCPU; id++)
        if ((id != vcpu_id) && (*online_map & (1 << id)))
            while (!gdb_started_vcpus[id]) barrier ();
    
    return ;
}

int gdb_breakpoint_type (struct pt_regs *ptregs)
{
    volatile struct gdb_bpnode *z0 = NULL; /* pointer to list of normal breakpoints */
    int bp;
    int vcpu_id = processor_id ();

    /* See if its one of our zbreaks */
    z0 = gdb_breakpoint[0];
    for (bp = 0; bp < TOTAL_BREAKPOINTS; bp++)
        if ((z0[bp].addr == (uint64_t *)(long)(ptregs->cp0_epc)) && z0[bp].valid)
            return NORMAL_BKPT;
    
    if (gdb_instr_buffer[vcpu_id].targetAddr && ((gdb_instr_buffer[vcpu_id].targetAddr) == (volatile uint64_t *)(long)(ptregs->cp0_epc)))
        return SINGLESTEP_BKPT;

    if (*(uint32_t *)(long)(ptregs->cp0_epc) == INITIAL_BREAK_INSTR)
        return INITIAL_BKPT;

    return TBH_BKPT;
}

int gdb_handle_single_step (struct pt_regs *ptregs)
{
    volatile uint64_t addr; /* next address to break */
    int vcpu_id = processor_id ();
    volatile uint32_t temp_instr = 0; /* branch delay flag */

    /* Restore the saved instruction at the single-step target address */
    undo_single_step ();
        
    /* If this exception is due to break instruction at branch delay slot... */
    if (ptregs->cp0_cause & (1 << 31)) {
#ifdef GDB_DEBUG_MSG
        printf("%d:%s: Branch delay slot\n", vcpu_id, __FUNCTION__);
#endif
        /* a break instruction is put at the next address after the delay slot 
           (which means the address of the branch instruction is not part of the
           loaded addresses); this "next address" is obtained as:
           The current EPC still contains the address just above delay slot, 
           hence "next address" is obtained by adding 8 to the address in EPC */

        /* Read the EPC (exception program counter) register) */
        addr = ptregs->cp0_epc;
            
        /* Increment the addr by 2 instructions (ie. 8 bytes) */
         addr = addr + 8;
            
        /* Fill in the single step structure */
        gdb_instr_buffer[vcpu_id].targetAddr = (uint64_t *)(long)addr;
        if (*(uint32_t *) (long)addr == BREAK_INSTR) {
            temp_instr = get_orig_instr((uint64_t *)(long)addr);
            if (temp_instr == BREAK_INSTR)
                gdb_instr_buffer[vcpu_id].targetAddr = NULL;
            else
                gdb_instr_buffer[vcpu_id].savedInstr = temp_instr;
        }
        else {
            gdb_instr_buffer[vcpu_id].savedInstr = *(uint32_t *) (long)addr;
              
            /* Replace the instruction at addr with the break instruction so 
               that the break is put at the next address after the delay slot */
            *(uint32_t *)(long)addr = BREAK_INSTR;
            gdb_flush_icache_addr ((int64_t)(long)addr);
        }
        barrier();
        return -1;
    }
    return 0;
}

void gdb_handle_tbh (struct pt_regs *ptregs)
{
    int vcpu_id = processor_id ();
    int id;

    /* TBH - part 1 */
    if (tbh_vcpu == -1) {
        /*printf("%d:ge:TBH-p1\n", vcpu_id);*/

        gdb_started_vcpus[vcpu_id] = 0;
        gdb_stopped_vcpus[vcpu_id] = 1;

        /* Set TBH operation */
        tbh_vcpu = vcpu_id;

        gdb_unlock (vcpu_id);

        /* Wait for other vcpus to stop */
        for (id = 0; id < TOTAL_VCPU; id++)
            if ((id != vcpu_id) && (*online_map & (1 << id)))
                while (!gdb_stopped_vcpus[id]) barrier ();

        /* Save the current address */
        gdb_tbhbreakpoint.addr = (uint64_t *)(long)(ptregs->cp0_epc);

        /* Remove the current single-step break */
        *(uint32_t *)(long)(ptregs->cp0_epc) = get_orig_instr ((uint64_t *)(long)(ptregs->cp0_epc));
        gdb_flush_icache_addr ((int64_t)(long)(ptregs->cp0_epc));

        /* Save the next address */
        gdb_tbhbreakpoint.targetAddr = (uint64_t *)(long)(get_target_addr (ptregs->cp0_epc, ptregs));

        /* Save the next instruction */
        gdb_tbhbreakpoint.savedInstr = *(uint32_t *)(gdb_tbhbreakpoint.targetAddr);

        /* Put break at next address */
        *(uint32_t *)(gdb_tbhbreakpoint.targetAddr) = BREAK_INSTR;
        gdb_flush_icache_addr ((int64_t)(long)(gdb_tbhbreakpoint.targetAddr));

        barrier ();

        /* TBH - part 1 is over, exit to proceed to TBH - part 2 */
    }

    /* TBH - part 2 */
    else if (tbh_vcpu == vcpu_id) {
        /*printf("%d:ge:TBH-p2\n", vcpu_id);*/

        /* Put break at previous address */
        *(uint32_t *)(gdb_tbhbreakpoint.addr) = BREAK_INSTR;
        gdb_flush_icache_addr ((int64_t)(long)(gdb_tbhbreakpoint.addr));

        /* Put original instruction at current address */
        *(uint32_t *)(gdb_tbhbreakpoint.targetAddr) = gdb_tbhbreakpoint.savedInstr;
        gdb_flush_icache_addr ((int64_t)(long)(gdb_tbhbreakpoint.targetAddr));

        barrier ();

        /* Invalidate the TBH node */
        gdb_tbhbreakpoint.addr = NULL;
        gdb_tbhbreakpoint.instr = BREAK_INSTR;
        gdb_tbhbreakpoint.targetAddr = NULL;
        gdb_tbhbreakpoint.savedInstr = BREAK_INSTR;

        /* Send pending ctrl-c, if any, to itself */
        if (ctrlc_ipi[vcpu_id]) {

            ctrlc_ipi[vcpu_id] = 0;
            /*printf("%d:IPI%dTo%d\n", vcpu_id, IRQ_STOP_GDB_IPI, vcpu_id); */
            gdb_send_ipi (IRQ_STOP_GDB_IPI, 0, 0, vcpu_id & 3, vcpu_id >> 2);

        }

        /* TBH - part 2 is over; Unset TBH operation */
        tbh_vcpu = -1;

        gdb_unlock (vcpu_id);

        gdb_stopped_vcpus[vcpu_id] = 0;
        gdb_started_vcpus[vcpu_id] = 1;

        /* Wait for other vcpus to start */
        for (id = 0; id < TOTAL_VCPU; id++)
            if ((id != vcpu_id) && (*online_map & (1 << id)))
                while (!gdb_started_vcpus[id]) barrier ();
    }
}

/* The assembly language exception handlers, break and trap, call this function.
   These assembly handlers save and restore registers, on entry and exit, 
   respectively; hence no need to handle that in this function */
void gdb_exception(struct pt_regs *ptregs)
{
    volatile uint32_t expt_code; /* exception code */
    int sigval = 0; /* signal number corresponding to the exception code */
    int vcpu_id = processor_id();
    int id;
    int bkpt_type = -1;
#ifdef GDB_DEBUG_MSG
    printf ("%d entered %s with epc 0x%llx\n", vcpu_id, __FUNCTION__, ptregs->cp0_epc);
#endif
    
    gdb_lock (vcpu_id);
    /*printf("%d:ge():AL\n", vcpu_id);*/

        /* non-CURRENT vcpus */
        if (!gdb_status_allowed[vcpu_id]) {
            /*printf("%d:ge():RL0\n", vcpu_id);*/
            gdb_unlock (vcpu_id);
            if (tbh_vcpu != -1) {
    
                /* TBH is set; Sync with it */
                sync_with_tbh_case ();
                return;
            }
        }

        /* CURRENT vcpu */
        else {
            gdb_current_vcpu = vcpu_id;
            /*printf("current vcpu = %u\n", gdb_current_vcpu);*/
        
            gdb_vcpu_c = vcpu_id;
            gdb_vcpu_g = vcpu_id;
        
            /* Stop other vcpus */
            for (id = 0; id < TOTAL_VCPU; id++)
                if ((id != vcpu_id) && (*online_map & (1 << id))) {
#ifdef GDB_DEBUG_MSG
                    printf ("%d:IPI%dTo%d\n", vcpu_id, IRQ_STOP_GDB_IPI, id);
#endif
                    gdb_status_allowed[id] = 0;
                    gdb_send_ipi (IRQ_BA_GDB_IPI, 0, 0, (id & 3), (id >> 2));
                    barrier ();
                }
        
            bkpt_type = gdb_breakpoint_type (ptregs);
            
            switch (bkpt_type) {
                case NORMAL_BKPT:
                    break;
                case SINGLESTEP_BKPT:
                    if (gdb_handle_single_step (ptregs)) return;
                    break;
                case INITIAL_BKPT:
                    break;
                case TBH_BKPT:
                    gdb_handle_tbh (ptregs);
                    return;
                default:
                    break;
            }
    
            /*printf("%d:ge():RL1\n", vcpu_id);*/
            gdb_unlock (vcpu_id);
        
            gdb_ack_allowed[vcpu_id] = 1;
        
            /* Get the exception code and translate it to a gdb signal*/
            /* TODO: use macros to define these hardcoded values */
            expt_code = (ptregs->cp0_cause & 0x7c) >> 2;
            sigval = gdb_sigval (expt_code);
        
            /* Tell the host why we're here */
            if (gdb_status_allowed[vcpu_id])
               gdb_status('T'/*'S'*/, sigval, ptregs);
        }
    
    /* Ask gdb what to do next */
    gdb_monitor (sigval, ptregs);
    
    /* This is executed when 'c', 'C', 's', 'S' packet contains an address */
    if (gnext_addr[vcpu_id] != 0) {
        
        /* Update the epc with gnext_addr and clear gnext_addr */
        ptregs->cp0_epc = gnext_addr[vcpu_id];
        gnext_addr[vcpu_id] = 0;
    }
    
#ifdef GDB_DEBUG_MSG
    printf("%d leaving %s\n", vcpu_id, __FUNCTION__);
#endif
}

/* Function to fetch the target address where a breakpoint will be inserted in 
   single-step condition */
static uint64_t get_target_addr(uint64_t pc, struct pt_regs *ptregs)
{
    InstFmt inst;		/* instruction */
    uint64_t ret_addr = pc + 4;	/* the target address by default */
    
    /*printf("%u entered %s\n", processor_id(), __FUNCTION__); */

    inst.word = *(uint32_t *) (long)pc;	/* read the next instruction  */
    if (inst.word == BREAK_INSTR)
        inst.word = get_orig_instr((uint64_t *) (long)pc);
    if (inst.word == NOP_INSTR || inst.word == BREAK_INSTR) {
    
        /* The instruction can be a valid NOP also, not associated with any branch; In 
        such cases, this printf() gives impression that an error condition is 
        reached, so it is commented out
        */
        /*printf("%u:gta():OrigInstr %x AtAddr %llx; Strange situation\n", processor_id(), inst.word, pc); */
        return ret_addr;
    }
    switch (inst.RType.op) { /* override default if branch */

        case OP_SPECIAL:

            switch (inst.RType.func) {

                case OP_JR:
                case OP_JALR:
                ret_addr = ptregs->regs[inst.RType.rs];
                break;

            };
            break; /* case OP_SPECIAL */
        
        case OP_REGIMM:

            switch (inst.IType.rt) {

                case OP_BLTZ:
                case OP_BLTZL:
                case OP_BLTZAL:
                case OP_BLTZALL:
                    if (ptregs->regs[inst.IType.rs] < 0)
                        ret_addr = (signed short)(inst.IType.imm << 2) + (pc + 4);
                    else
                        ret_addr = (pc + 8);
                    break;
        
                case OP_BGEZ:
                case OP_BGEZL:
                case OP_BGEZAL:
                case OP_BGEZALL:
                if (ptregs->regs[inst.IType.rs] >= 0)
                    ret_addr = (signed short)(inst.IType.imm << 2) + (pc + 4);
                else
                    ret_addr = (pc + 8);
                break;
                };
            break; /* case OP_REGIMM */
        
        case OP_J:

        case OP_JAL:

            ret_addr = ((inst.JType.target << 2) + ((pc + 4) & 0xfffffffff0000000ULL));
            break; /* case OP_JAL */
        
        case OP_BEQ:

        case OP_BEQL:

            if (ptregs->regs[inst.IType.rs] == ptregs->regs[inst.IType.rt])
                ret_addr = (signed short)(inst.IType.imm << 2) + (pc + 4);
            else
                ret_addr = (pc + 8);
            break; /* case OP_BEQL */
        
        case OP_BNE:

        case OP_BNEL:

            if (ptregs->regs[inst.IType.rs] != ptregs->regs[inst.IType.rt])
                ret_addr = (signed short)(inst.IType.imm << 2) + (pc + 4);
            else
                ret_addr = (pc + 8);
            break; /* case OP_BNEL */
        
        case OP_BLEZ:

        case OP_BLEZL:
            if (ptregs->regs[inst.IType.rs] <= 0)
                ret_addr = (signed short)(inst.IType.imm << 2) + (pc + 4);
            else
                ret_addr = (pc + 8);
            break; /* case OP_BLEZL */
        
        case OP_BGTZ:

        case OP_BGTZL:

            if (ptregs->regs[inst.IType.rs] > 0)
                ret_addr = (signed short)(inst.IType.imm << 2) + (pc + 4);
            else
                ret_addr = (pc + 8);
            break; /* case OP_BGTZL */
        
    } /* switch (inst.RType.op) */
    
    /* The target address should be part of the loaded addresses; */
    if (((uint8_t *)(long)ret_addr >= _loadaddr) && ((uint8_t *)(long)ret_addr <= _endaddr))
        return ret_addr;

    /* Otherwise, it means that the target address is a branched address which is 
    * not part of the loaded addresses and which we cannot track down, hence make 
    * the target address as one 8 bytes past the current address (pc) since the 
    * address after pc (ie. pc + 4) is definitely a branch delay slot which will 
    * be executed along with the branch instruction (at pc)
    */
    else
        return pc + 8;
    /*printf("%u leaving %s\n", processor_id(), __FUNCTION__); */
}

/* IPI handler for stop ipi due to ctrl-c / stop from CURRENT vcpu */
irqreturn_t gdb_stop_ipi_handler(int irq, struct pt_regs *ptregs)
{
    int vcpu_id = processor_id ();
    int id;
#ifdef GDB_DEBUG_MSG
    printf ("%d entered %s with irq %d and epc %llx\n\n", vcpu_id, __FUNCTION__, irq, ptregs->cp0_epc);
#endif
    /* Acknowledge the IPI interrupt from PIC; done at the CPU level */
    write_phnx_eirr ((uint64_t) 1 << irq);

    if (((uint8_t *) (long)ptregs->cp0_epc >= _loadaddr) && ((uint8_t *) (long)ptregs->cp0_epc <= _endaddr)) {
        /* Tell the host why we're here */
        switch (irq) {
            case IRQ_STOP_GDB_IPI:
                gdb_lock (vcpu_id);
                /*printf("%u:gsih:AL\n", vcpu_id);*/

                    if (!gdb_status_allowed[vcpu_id]) {
                        /*printf("%d:gsih:RL\n", vcpu_id);*/
                        gdb_unlock (vcpu_id);
                        if (tbh_vcpu != -1) {
    
                            /* TBH is set... */
                            if (tbh_vcpu == vcpu_id) {
    
                                /* ...by this vcpu itself; Save the ^c and exit */
                                ctrlc_ipi[vcpu_id] = 1;
                            }
                            else {
    
                                /* ...by some other vcpu; Sync with it */
                                sync_with_tbh_case ();
                            
                                /* Send this ctrl-c to itself again so that it is not missed */
                                gdb_send_ipi (IRQ_STOP_GDB_IPI, 0, 0, (vcpu_id & 3), (vcpu_id >> 2));
                            }
                            return 0;
                        }
                        break;
                    }
    
                    gdb_current_vcpu = vcpu_id;
                    /*printf("gsih:current vcpu = %u\n", gdb_current_vcpu);*/
    
                    gdb_vcpu_c = vcpu_id;
                    gdb_vcpu_g = vcpu_id;
    
                    /* Stop other vcpus */
                    for (id = 0; id < TOTAL_VCPU; id++)
                        if ((id != vcpu_id) && (*online_map & (1 << id))) {
                            gdb_status_allowed[id] = 0;
#ifdef GDB_DEBUG_MSG
                            /*printf ("%d:IPI%dTo%u\n", vcpu_id, IRQ_STOP_GDB_IPI, id);*/
#endif
                            /*gdb_send_ipi (IRQ_BA_GDB_IPI, 0, 0, (id & 3), (id >> 2));*/
                            barrier ();
                    }

                /*printf("%d:gsih:RL0\n", vcpu_id);*/
                gdb_unlock (vcpu_id);
                break;


                /* TODO: If any vcpu can act as tunneler, this needs to be taken care of */
                if (tbh_vcpu == vcpu_id) {

                    /* ^c is received for the TBH vcpu itself, save it, and exit */
                    ctrlc_ipi[vcpu_id] = 1;
                    return 0;

                }

                break;

            case IRQ_BA_GDB_IPI:
                gdb_lock (vcpu_id);
                /*printf("%d:gsti():AL\n", vcpu_id);*/
                
                if (!gdb_status_allowed[vcpu_id]) {
                    /*printf("%d:gsti():RL\n", vcpu_id);*/
                    gdb_unlock (vcpu_id);
                    if (tbh_vcpu != -1) {
                    
                        /* TBH is set; Sync with it */
                        sync_with_tbh_case ();
                        return 0;
                    }
                }
                break;

            default:
                break;
        }

        /*printf("%d:gsti():RL\n", vcpu_id);*/
        gdb_unlock (vcpu_id);

        gdb_monitor(15, ptregs);
        
        /* This is executed when 'c', 'C', 's', 'S' packet contains an address */
        if (gnext_addr[vcpu_id] != 0) {
        
            /* Update the EPC with gnext_addr and initialize gnext_addr to default value */
            ptregs->cp0_epc = gnext_addr[vcpu_id];
            gnext_addr[vcpu_id] = 0;
        }
    }

    else {
        uint8_t ibuf[] = "Trying to stop at an unloaded address...disallowed...try again\r\n";
        
        /* Get the output buffer */
        uint8_t *bufptr = (uint8_t *) gdb_rxtx_buf[vcpu_id];
        *bufptr++ = 'O';
        
        /* Convert into hex format */
        ucharbuf_to_hbuf(bufptr, ibuf, strlen((char *)ibuf));
        
        gdb_putpacket((uint8_t *) gdb_rxtx_buf[vcpu_id], strlen((char *)ibuf) * 2 + 1);
        
        gdb_putpacket((uint8_t *)"OK", 2);
    }
    
#ifdef GDB_DEBUG_MSG
    printf("%d leaving %s\n", vcpu_id, __FUNCTION__);
#endif
    return 0;
}

/* Send ipi to the vcpu defined by cpuId and thrId */
static void gdb_send_ipi(uint32_t vector, uint32_t broadcast, uint32_t nmi, uint32_t thr, uint32_t vcpu)
{
    uint32_t ipi = 0x0;
    
#ifdef GDB_DEBUG_MSG
    printf("%u entered %s\n", processor_id(), __FUNCTION__);
#endif
    if (vector >= 64)
        printf("%s: Wrong arg (vector = %d)\n", __FUNCTION__, vector);
    if (broadcast != 0 && broadcast != 1)
        printf("%s: Wrong arg (broadcast = %d)\n", __FUNCTION__,
    broadcast);
    if (nmi != 0 && nmi != 1)
        printf("%s: Wrong arg (nmi = %d)\n", __FUNCTION__, nmi);
    if (thr >= 4)
        printf("%s: Wrong arg (thrId = %d)\n", __FUNCTION__, thr);
    if (vcpu >= 8)
        printf("%s: Wrong arg (cpuId = %d)\n", __FUNCTION__, vcpu);
    
    ipi = ipi | (vector << 0);
    ipi = ipi | (broadcast << 7);
    ipi = ipi | (nmi << 8);
    ipi = ipi | (thr << 16);
    ipi = ipi | (vcpu << 20);
    
    pic_send_ipi(ipi);
#ifdef GDB_DEBUG_MSG
    printf("%u leaving %s\n", processor_id(), __FUNCTION__);
#endif
}
