/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

/********************************************************************************
 gdbuart.c - UART1 management by the gdb stub

 MODIFICATION HISTORY
 --------------------
 v1.0,16oct05,Suresh DS initial version
 v1.1,25jan06,Suresh DS	error handling, changing UART1 routines for Rx & Tx to 
                        handle only 1 byte and inclusion of gdb stub's own fifo
                        handling file
 v1.2,15mar06,Suresh DS	Pre-compilation code to support integration with VxWorks
                        for shared code segment (kseg0) support
 v1.3,05mar09,Suresh DS	support for uart 0
********************************************************************************/

#ifdef XLR_VXWORKS
#include "wakeup_printk.h"
#else
#include <stdio.h>
#include <stdlib.h>
#endif

#include "uart.h"
#include "byteorder.h"
#include "pt_regs.h"
#include "fifo.h"
#include "pic.h"
#include "mips-exts.h"
#include "iomap.h"

#ifndef UART1_REG_BITS
/* UART1 related definitions */
#define UART_IER_RDAI 0x01
#define UART_IER_RLSI 0x04
#define UART_LSR_DRI 0x01
#define UART_LSR_OEI 0x02
#define UART_LSR_PEI 0x03
#define UART_LSR_FEI 0x04
#define UART_LSR_BEI 0x05
#define UART_LSR_TFE 0x06
#define UART_LSR_TEI 0x07
#define UART_LSR_ERR 0x80
#endif

extern int gdb_uart;
volatile uint32_t * volatile gdb_uart_mmio = NULL;
static struct fifo tx_fifo;
static struct fifo rx_fifo;

#define TX_BURST_SIZE 1
/*#define TX_BURST_SIZE 3*/

/* Function to get 1 byte from UART1 */
int32_t get_debug_char (void)
{
    int32_t data = 0;
    volatile int32_t lsr = 0;
    
    lsr = be32_to_cpu (phoenix_read_reg (gdb_uart_mmio, UART_LSR));
    if (lsr & 0x02) {
        /* fifo overrun */
        printf ("0:gdc():hw RO\n");
    }
    if (lsr & 0x01) {
        /* Rx Data */
        data = be32_to_cpu (phoenix_read_reg (gdb_uart_mmio, UART_RHR));
        if (!fifo_enqueue (&rx_fifo, data))
            printf ("0:gdc():drv RO\n");
    }
    if (lsr & 0x80) {
        /* parity/frame/break - push a 0! */
        if (!fifo_enqueue (&rx_fifo, data))
            printf ("0:gdc():drv RO\n");
    }
    
    if (fifo_dequeue (&rx_fifo, &data))
        return data;
    else
        return -1;
}

/* Function to put 1 byte to UART1 */
int32_t put_debug_char (int8_t c)
{
    /* local declarations */
    int32_t data = 0;
    volatile int32_t lsr = 0;
    int32_t i = 0;
    
    if (c == '\n') {
        int8_t cr = '\r';
        if (!fifo_enqueue (&tx_fifo, (int32_t) cr)) {
            /* Too bad tx fifo is full, drop the character */
            printf ("0:gdc():drv TO\n");
            return 1;
        }
    }
    if (!fifo_enqueue (&tx_fifo, (int32_t) c)) {
        /* Too bad tx fifo is full, drop the character */
        printf("0:gdc():drv TO\n");
        return 1;
    }
    
    /* accumalate some more characters */
    /*if((fifo_size(&tx_fifo) < TX_BURST_SIZE)&&(c!='\n')&&(c!='\r')) return 0;*/
    
    for (;;) {
        
        if (fifo_empty (&tx_fifo))
            break;
        
        /* wait for Tx empty indication */
        for (;;) {
            
            lsr = be32_to_cpu (phoenix_read_reg(gdb_uart_mmio, UART_LSR));
            
            if (lsr & 0x02) {
                /* fifo overrun */
                printf ("0:pdc():hw RO\n");
            }
            if (lsr & 0x80) {
                /* parity/frame/break - push a 0! */
                if (!fifo_enqueue (&rx_fifo, 0))
                    printf ("0:pdc():parerr-drv RO\n");
            }
            if (lsr & 0x01) {
                /* Rx Data */
                data = be32_to_cpu (phoenix_read_reg(gdb_uart_mmio, UART_RHR));
                if (!fifo_enqueue (&rx_fifo, data))
                    printf ("0:pdc():drv RO\n");
            }
            /* Tx Fifo empty */
            if (lsr & 0x20)
            break;
        }
        
        /* transmit upto TX_BURST_SIZE char */
        for (i = 0; i < TX_BURST_SIZE && fifo_dequeue (&tx_fifo, &data); i++)
        
            /* Tx Data */
            phoenix_write_reg (gdb_uart_mmio, UART_THR, cpu_to_be32(data));
    }
    return 0;
}/* put_debug_char()*/

void gdb_uart_init (void)
{
    if (gdb_uart == 1)
        gdb_uart_mmio = (uint32_t *) phoenix_io_mmio (PHOENIX_IO_UART_1_OFFSET);
    else if (gdb_uart == 0)
        gdb_uart_mmio = (uint32_t *) phoenix_io_mmio (PHOENIX_IO_UART_0_OFFSET);
#ifdef GDB_DEBUG_MSG
    printf ("%u entered %s\n", processor_id (), __FUNCTION__);
#endif
    
    /* Initialize the Tx & Rx FIFOs (software buffers) */
    fifo_init (&tx_fifo);
    fifo_init (&rx_fifo);
    
    /* Set up the baud rate */
    phoenix_write_reg (gdb_uart_mmio, UART_LCR, cpu_to_be32 (be32_to_cpu (phoenix_read_reg (gdb_uart_mmio, UART_LCR)) | (1 << 7)));
    /* 38400 bps */
    phoenix_write_reg (gdb_uart_mmio, UART_DLB_2, cpu_to_be32 (UART_BR_DLB2));
    phoenix_write_reg (gdb_uart_mmio, UART_DLB_1, cpu_to_be32 (UART_BR_DLB1));
#if 0 /* 9600 bps */
    phoenix_write_reg (gdb_uart_mmio, UART_DLB_2, cpu_to_be32 (UART_BR9600_DLB2));
    phoenix_write_reg (gdb_uart_mmio, UART_DLB_1, cpu_to_be32 (UART_BR9600_DLB1));
#endif
    phoenix_write_reg (gdb_uart_mmio, UART_LCR, cpu_to_be32 (be32_to_cpu (phoenix_read_reg (gdb_uart_mmio, UART_LCR)) & ~(1 << 7)));
#ifdef GDB_DEBUG_MSG
    printf ("%u leaving %s\n", processor_id (), __FUNCTION__);
#endif
}
