/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _IRQ_H
#define _IRQ_H

#define NR_IRQS 64

#include "pt_regs.h"
#include "spinlock.h"
/* IRQ line status. */

#define IRQ_INPROGRESS     1    // IRQ handler active - do not enter!
#define IRQ_DISABLED       2    // IRQ disabled - do not enter!
#define IRQ_PENDING        4    // IRQ pending - replay on enable
#define IRQ_REPLAY         8    // IRQ has been replayed but not acked yet
#define IRQ_AUTODETECT    16    // IRQ is being autodetected
#define IRQ_WAITING       32  	// IRQ not yet seen - for autodetection
#define IRQ_LEVEL         64  	// IRQ level triggered
#define IRQ_MASKED       128 	// IRQ masked - shouldn't be seen again
#define IRQ_PER_CPU      256 	// IRQ is per CPU

#define IRQ_UART                  2
#define IRQ_IPI_SMP_FUNCTION      3
#define IRQ_IPI_SMP_RESCHEDULE    4
#define IRQ_REMOTE_DEBUG          5
#define IRQ_MAC                   6
#define IRQ_TIMER                 7

#define IRQ_PIC_I2C_0            19
#define IRQ_PIC_I2C_1            20


#define IRQ_IPI_CRF_MGMT_IPI	NLM_MANAGEMENT_IPI /* */
#define IRQ_IPI_CRF_EVENTQ_IPI  NLM_EVENTQ_IPI

#define IRQ_STOP_GDB_IPI      40
#define IRQ_BA_GDB_IPI        41

#define IRQ_STOP_CPU_IPI 51
#define IRQ_WAKEUP_CPU_IPI 52
#define IRQ_JOIN_CPU_IPI 54


#ifndef __ASSEMBLY__

typedef int irqreturn_t;

struct irq_ops {
    const char *typename;
    int (*startup)(int irq);
    void (*shutdown)(int irq);
    void (*proack)(int irq);
    void (*epiack)(int irq);
	void (*set_affinity)(int irq, unsigned int dest);
	irqreturn_t (*handler)(int, struct pt_regs *);
};

typedef struct irq_ops irq_ops_t;

/* 
 * This is the "IRQ descriptor", which contains various
 * information about the irq, including what kind of
 * hardware handling it has, whether it is disabled etc.
 * Pad this out to 32 bytes for cache and indexing reasons
 */

typedef struct irq_desc {
	spinlock_t lock;
    unsigned int status;
    irq_ops_t ops;
	unsigned long flags;
	unsigned long mask;
} irq_desc_t;

extern void phoenix_irq_handler(void);
extern void handle_reserved(void);
extern int init_irq(void);

extern char except_vec_tlbrefill[], except_vec_xtlbrefill[];
extern char except_vec_genex[], except_vec_cacheerr[];
extern char except_vec_vecint[];

__asm__ (
	".macro\t__sti\n\t"
	".set\tpush\n\t"
	".set\treorder\n\t"
	".set\tnoat\n\t"
	"mfc0\t$1,$12\n\t"
	"ori\t$1,0x1f\n\t"
	"xori\t$1,0x1e\n\t"
	"mtc0\t$1,$12\n\t"
	".set\tpop\n\t"
	".endm");

static __inline__ void sti(void)
{
	__asm__ __volatile__(
		"__sti"
		: /* no outputs */
		: /* no inputs */
		: "memory");
}

__asm__ (
	".macro\t__cli\n\t"
	".set\tpush\n\t"
	".set\tnoat\n\t"
	"mfc0\t$1,$12\n\t"
	"ori\t$1,1\n\t"
	"xori\t$1,1\n\t"
	".set\tnoreorder\n\t"
	"mtc0\t$1,$12\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	".set\tpop\n\t"
	".endm");

static __inline__ void cli(void)
{
	__asm__ __volatile__(
		"__cli"
		: /* no outputs */
		: /* no inputs */
		: "memory");
}

__asm__ (
	  ".macro\tlocal_irq_enable\n\t"
	  ".set\tpush\n\t"
	  ".set\treorder\n\t"
	  ".set\tnoat\n\t"
	  "mfc0\t$1,$12\n\t"
	  "ori\t$1,0x1f\n\t"
	  "xori\t$1,0x1e\n\t"
	  "mtc0\t$1,$12\n\t"
	  ".set\tpop\n\t"
	  ".endm");

static inline void local_irq_enable(void)
{
	  __asm__ __volatile__(
	      "local_irq_enable"
	      : /* no outputs */
	      : /* no inputs */
	      : "memory");
}

/*
 * For cli() we have to insert nops to make sure that the new value
 * has actually arrived in the status register before the end of this 
 * macro.
 * R4000/R4400 need three nops, the R4600 two nops and the R10000 needs
 * no nops at all. 
 */
__asm__ (
	  ".macro\tlocal_irq_disable\n\t"
	  ".set\tpush\n\t"
	  ".set\tnoat\n\t"
	  "mfc0\t$1,$12\n\t"
	  "ori\t$1,1\n\t"
	  "xori\t$1,1\n\t"
	  ".set\tnoreorder\n\t"
	  "mtc0\t$1,$12\n\t"
	  "sll\t$0, $0, 1\t\t\t# nop\n\t"
	  "sll\t$0, $0, 1\t\t\t# nop\n\t"
	  "sll\t$0, $0, 1\t\t\t# nop\n\t"
	  ".set\tpop\n\t"
	  ".endm");

static inline void local_irq_disable(void)
{
	  __asm__ __volatile__(
	      "local_irq_disable"
	      : /* no outputs */
	      : /* no inputs */
	      : "memory");
}

__asm__ (
    ".macro\t__save_and_cli result\n\t"
    ".set\tpush\n\t"
    ".set\treorder\n\t"
    ".set\tnoat\n\t"
    "mfc0\t\\result, $12\n\t"
    "ori\t$1, \\result, 1\n\t"
    "xori\t$1, 1\n\t"
    ".set\tnoreorder\n\t"
    "mtc0\t$1, $12\n\t"
    "sll\t$0, $0, 1\t\t\t# nop\n\t"
    "sll\t$0, $0, 1\t\t\t# nop\n\t"
    "sll\t$0, $0, 1\t\t\t# nop\n\t"
    ".set\tpop\n\t"
    ".endm");

#define __save_and_cli(x)           \
__asm__ __volatile__(               \
    "__save_and_cli\t%0"            \
    : "=r" (x)                      \
    : /* no inputs */               \
    : "memory")

#define __restore_flags(flags)      \
do {                                \
    unsigned long __tmp1;           \
                                    \
    __asm__ volatile (              \
        "__restore_flags\t%0"       \
        : "=r" (__tmp1)             \
        : "0" (flags)               \
        : "memory");                \
} while(0)

__asm__(".macro\t__restore_flags flags\n\t"
    ".set\tnoreorder\n\t"
    ".set\tnoat\n\t"
    "mfc0\t$1, $12\n\t"
    "andi\t\\flags, 1\n\t"
    "ori\t$1, 1\n\t"
    "xori\t$1, 1\n\t"
    "or\t\\flags, $1\n\t" 
    "mtc0\t\\flags, $12\n\t"
    "sll\t$0, $0, 1\t\t\t# nop\n\t"
    "sll\t$0, $0, 1\t\t\t# nop\n\t"
    "sll\t$0, $0, 1\t\t\t# nop\n\t"
    ".set\tat\n\t"
    ".set\treorder\n\t"
    ".endm");

__asm__ (
	".macro\tlocal_irq_save result\n\t"
	".set\tpush\n\t"
	".set\treorder\n\t"
	".set\tnoat\n\t"
	"mfc0\t\\result, $12\n\t"
	"ori\t$1, \\result, 1\n\t"
	"xori\t$1, 1\n\t"
	".set\tnoreorder\n\t"
	"mtc0\t$1, $12\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	".set\tpop\n\t"
	".endm");

#define local_irq_save(x)					\
__asm__ __volatile__(						\
	"local_irq_save\t%0"					\
	: "=r" (x)								\
	: /* no inputs */						\
	: "memory")

__asm__(".macro\tlocal_irq_restore flags\n\t"
	".set\tnoreorder\n\t"
	".set\tnoat\n\t"
	"mfc0\t$1, $12\n\t"
	"andi\t\\flags, 1\n\t"
	"ori\t$1, 1\n\t"
	"xori\t$1, 1\n\t"
	"or\t\\flags, $1\n\t"
	"mtc0\t\\flags, $12\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	"sll\t$0, $0, 1\t\t\t# nop\n\t"
	".set\tat\n\t"
	".set\treorder\n\t"
	".endm");

#define local_irq_restore(flags)			\
do {										\
	unsigned long __tmp1;					\
											\
	__asm__ __volatile__(					\
		"local_irq_restore\t%0"				\
		: "=r" (__tmp1)						\
		: "0" (flags)						\
		: "memory");						\
} while(0)

#define spin_lock_irqsave(lock, flags) \
    do { \
		local_irq_save(flags); \
		spin_lock(lock); \
	} while (0)

#define spin_unlock_irqrestore(lock, flags) \
    do { \
		spin_unlock(lock); \
		local_irq_restore(flags); \
	} while (0)


#define IRQ_NONE	(0)
#define IRQ_HANDLED	(1)
#define IRQ_RETVAL(x)	((x) != 0)

extern int request_irq(unsigned int,
		       irqreturn_t (*handler)(int, struct pt_regs *),
		       unsigned long, 
		       unsigned int);
extern void free_irq(unsigned int);
extern void set_irqops(int irq, int option, void *function);


extern int register_shutdown_handler(int cpu_id, void (*func)(void *), void *args);
extern void free_shutdown_handler(int cpu_id);

#define EMPTY_IRQ_FLAGS 0
#define IRQ_FLAGS_RESERVED 1

extern void (*netos_shutdown_func[NCPU]) (void *);
extern void *netos_shutdown_args[NCPU];
extern void (*netos_join_func[NCPU]) (void *);
extern void *netos_join_args[NCPU];
	
#endif

#endif
