/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _PCI_H
#define _PCI_H

#include "list.h"
#include "types.h"
#include "iomap.h"

/*
 * Under PCI, each device has 256 bytes of configuration address space,
 * of which the first 64 bytes are standardized as follows:
 */
#define PCI_VENDOR_ID		0x00	/* 16 bits */
#define PCI_DEVICE_ID		0x02	/* 16 bits */
#define PCI_COMMAND		0x04	/* 16 bits */
#define  PCI_COMMAND_IO		0x1	/* Enable response in I/O space */
#define  PCI_COMMAND_MEMORY	0x2	/* Enable response in Memory space */
#define  PCI_COMMAND_MASTER	0x4	/* Enable bus mastering */
#define  PCI_COMMAND_SPECIAL	0x8	/* Enable response to special cycles */
#define  PCI_COMMAND_INVALIDATE	0x10	/* Use memory write and invalidate */
#define  PCI_COMMAND_VGA_PALETTE 0x20	/* Enable palette snooping */
#define  PCI_COMMAND_PARITY	0x40	/* Enable parity checking */
#define  PCI_COMMAND_WAIT 	0x80	/* Enable address/data stepping */
#define  PCI_COMMAND_SERR	0x100	/* Enable SERR */
#define  PCI_COMMAND_FAST_BACK	0x200	/* Enable back-to-back writes */

#define PCI_STATUS		0x06	/* 16 bits */
#define  PCI_STATUS_CAP_LIST	0x10	/* Support Capability List */
#define  PCI_STATUS_66MHZ	0x20	/* Support 66 Mhz PCI 2.1 bus */
#define  PCI_STATUS_UDF		0x40	/* Support User Definable Features [obsolete] */
#define  PCI_STATUS_FAST_BACK	0x80	/* Accept fast-back to back */
#define  PCI_STATUS_PARITY	0x100	/* Detected parity error */
#define  PCI_STATUS_DEVSEL_MASK	0x600	/* DEVSEL timing */
#define  PCI_STATUS_DEVSEL_FAST	0x000	
#define  PCI_STATUS_DEVSEL_MEDIUM 0x200
#define  PCI_STATUS_DEVSEL_SLOW 0x400
#define  PCI_STATUS_SIG_TARGET_ABORT 0x800 /* Set on target abort */
#define  PCI_STATUS_REC_TARGET_ABORT 0x1000 /* Master ack of " */
#define  PCI_STATUS_REC_MASTER_ABORT 0x2000 /* Set on master abort */
#define  PCI_STATUS_SIG_SYSTEM_ERROR 0x4000 /* Set when we drive SERR */
#define  PCI_STATUS_DETECTED_PARITY 0x8000 /* Set on parity error */

#define PCI_CLASS_REVISION	0x08	/* High 24 bits are class, low 8
					   revision */
#define PCI_REVISION_ID         0x08    /* Revision ID */
#define PCI_CLASS_PROG          0x09    /* Reg. Level Programming Interface */
#define PCI_CLASS_DEVICE        0x0a    /* Device class */

#define PCI_CACHE_LINE_SIZE	0x0c	/* 8 bits */
#define PCI_LATENCY_TIMER	0x0d	/* 8 bits */
#define PCI_HEADER_TYPE		0x0e	/* 8 bits */
#define  PCI_HEADER_TYPE_NORMAL	0
#define  PCI_HEADER_TYPE_BRIDGE 1
#define  PCI_HEADER_TYPE_CARDBUS 2

#define PCI_BIST		0x0f	/* 8 bits */
#define PCI_BIST_CODE_MASK	0x0f	/* Return result */
#define PCI_BIST_START		0x40	/* 1 to start BIST, 2 secs or less */
#define PCI_BIST_CAPABLE	0x80	/* 1 if BIST capable */

/*
 * Base addresses specify locations in memory or I/O space.
 * Decoded size can be detenlmned by writing a value of 
 * 0xffffffff to the register, and reading it back.  Only 
 * 1 bits are decoded.
 */
#define PCI_BASE_ADDRESS_0	0x10	/* 32 bits */
#define PCI_BASE_ADDRESS_1	0x14	/* 32 bits [htype 0,1 only] */
#define PCI_BASE_ADDRESS_2	0x18	/* 32 bits [htype 0 only] */
#define PCI_BASE_ADDRESS_3	0x1c	/* 32 bits */
#define PCI_BASE_ADDRESS_4	0x20	/* 32 bits */
#define PCI_BASE_ADDRESS_5	0x24	/* 32 bits */
#define  PCI_BASE_ADDRESS_SPACE	0x01	/* 0 = memory, 1 = I/O */
#define  PCI_BASE_ADDRESS_SPACE_IO 0x01
#define  PCI_BASE_ADDRESS_SPACE_MEMORY 0x00
#define  PCI_BASE_ADDRESS_MEM_TYPE_MASK 0x06
#define  PCI_BASE_ADDRESS_MEM_TYPE_32	0x00	/* 32 bit address */
#define  PCI_BASE_ADDRESS_MEM_TYPE_1M	0x02	/* Below 1M [obsolete] */
#define  PCI_BASE_ADDRESS_MEM_TYPE_64	0x04	/* 64 bit address */
#define  PCI_BASE_ADDRESS_MEM_PREFETCH	0x08	/* prefetchable? */
#define  PCI_BASE_ADDRESS_MEM_MASK	(~0x0fUL)
#define  PCI_BASE_ADDRESS_IO_MASK	(~0x03UL)
/* bit 1 is reserved if address_space = 1 */

/* Header type 0 (normal devices) */
#define PCI_CARDBUS_CIS		0x28
#define PCI_SUBSYSTEM_VENDOR_ID	0x2c
#define PCI_SUBSYSTEM_ID	0x2e  
#define PCI_ROM_ADDRESS		0x30	/* Bits 31..11 are address, 10..1 reserved */
#define  PCI_ROM_ADDRESS_ENABLE	0x01
#define PCI_ROM_ADDRESS_MASK	(~0x7ffUL)

#define PCI_CAPABILITY_LIST	0x34	/* Offset of first capability list entry */

/* 0x35-0x3b are reserved */
#define PCI_INTERRUPT_LINE	0x3c	/* 8 bits */
#define PCI_INTERRUPT_PIN	0x3d	/* 8 bits */
#define PCI_MIN_GNT		0x3e	/* 8 bits */
#define PCI_MAX_LAT		0x3f	/* 8 bits */

#define PCI_VENDOR_ID_PHOENIX   0x182e
#define PCI_DEVICE_ID_PHOENIX   0

#define PCI_DEVFN(slot,func)	((((slot) & 0x1f) << 3) | ((func) & 0x07))
#define PCI_SLOT(devfn)		(((devfn) >> 3) & 0x1f)
#define PCI_FUNC(devfn)		((devfn) & 0x07)

struct pcidev_driver;
struct pci_dev;

struct pci_dev {
  struct list list;

  unsigned int devfn;

  unsigned short vendor;
  unsigned short device;  
  unsigned char hdr_type;

  unsigned char bus_number;

  char         name[90];
  char         slot_name[8];

  struct pcidev_driver *drv;

  unsigned int *mmio;
  unsigned int  mmio_size;
};

struct pcidev_driver {
  char             name[90];
  unsigned short   vendor;
  unsigned short   device;
  void  (*init)(struct pci_dev *);
  struct pcidev_driver *next;
};

#define PCI_CTRL_HOST_MODE   1
#define PCI_CTRL_DEVICE_MODE 0
#define PCI_CTRL_SIM_MODE    2

#define PCI_CTRL_REG_MODE            0
#define PCI_CTRL_REG_DEV_MODE_BASE   1
#define PCI_CTRL_REG_DEV_MODE_SIZE   2

/* PCIX extended reg base starts at byte offset 256 */
#ifdef PHOENIX_LITTLE_ENDIAN
#define PCIX_REG_BASE 64
#else
#define PCIX_REG_BASE (512 + 64)
#endif

#define PCIX_HBAR0_REG ( PCIX_REG_BASE + 0)
#define PCIX_HBAR1_REG ( PCIX_REG_BASE + 1)
#define PCIX_HBAR2_REG ( PCIX_REG_BASE + 2)
#define PCIX_HBAR3_REG ( PCIX_REG_BASE + 3)
#define PCIX_HBAR4_REG ( PCIX_REG_BASE + 4)
#define PCIX_HBAR5_REG ( PCIX_REG_BASE + 5)

#define PCIX_ENBAR0_REG ( PCIX_REG_BASE + 6)
#define PCIX_ENBAR1_REG ( PCIX_REG_BASE + 7)
#define PCIX_ENBAR2_REG ( PCIX_REG_BASE + 8)
#define PCIX_ENBAR3_REG ( PCIX_REG_BASE + 9)
#define PCIX_ENBAR4_REG ( PCIX_REG_BASE + 10)
#define PCIX_ENBAR5_REG ( PCIX_REG_BASE + 11)

#define PCIX_MATCHBIT_ADDR_REG ( PCIX_REG_BASE + 12)
#define PCIX_MATCHBIT_SIZE_REG ( PCIX_REG_BASE + 13)
#define PCIX_PHOENIX_CONTROL_REG ( PCIX_REG_BASE + 14)

#define PCIX_INTRPT_CONTROL_REG ( PCIX_REG_BASE + 15)
#define PCIX_INTRPT_STATUS_REG ( PCIX_REG_BASE + 16)

#define PCIX_ERRSTATUS_REG ( PCIX_REG_BASE + 17)

#define PCIX_INT_MSILADDR0_REG ( PCIX_REG_BASE + 18)
#define PCIX_INT_MSIHADDR0_REG ( PCIX_REG_BASE + 19)
#define PCIX_INT_MSIDATA0_REG ( PCIX_REG_BASE + 20)

#define PCIX_INT_MSILADDR1_REG ( PCIX_REG_BASE + 21)
#define PCIX_INT_MSIHADDR1_REG ( PCIX_REG_BASE + 22)
#define PCIX_INT_MSIDATA1_REG ( PCIX_REG_BASE + 23)

#define PCIX_TARGET_DBG0_REG ( PCIX_REG_BASE + 24)
#define PCIX_TARGET_DBG1_REG ( PCIX_REG_BASE + 25)

#define PCIX_INIT_DBG0_REG ( PCIX_REG_BASE + 26)
#define PCIX_INIT_DBG1_REG ( PCIX_REG_BASE + 27)

#define PCIX_EVENT_EN_REG ( PCIX_REG_BASE + 28)
#define PCIX_EVENT_CNT0_REG ( PCIX_REG_BASE + 29)
#define PCIX_EVENT_CNT1_REG ( PCIX_REG_BASE + 30)

#define PCIX_PCIXMODE_EN_REG ( PCIX_REG_BASE + 31)

#define PCIX_EXPROM_ADDR_REG ( PCIX_REG_BASE + 32)
#define PCIX_EXPROM_SIZE_REG ( PCIX_REG_BASE + 33)

#define PCIX_DMA_TIMEOUT_REG ( PCIX_REG_BASE + 34)

#define PCIX_HOST_MODE_CTRL_STATUS_REG ( PCIX_REG_BASE + 35)

extern int pci_ctrl_mode;
extern unsigned long pci_config_base;
extern unsigned long pci_mmio_base;
extern phoenix_reg_t *pci_ctrl_mmio;

//pcie registers
#define PCIE_EP_MEM_BASE      0x1A
#define PCIE_EP_MEM_LIMIT     0x1B
#define PCIE_MSG_BUCKET0_SIZE 0x320
#define PCIE_MSG_BUCKET1_SIZE 0x321

#endif
