/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <dlfcn.h>
#include <errno.h>

#define MAX_CPUS 32
#define BUFLEN 256

enum {
	MODE_APP_AMP = 1,
	MODE_APP_SMP = 2
};

extern int dom_id;
extern int fdt_get_cpu_online_map(void *fdt, unsigned int *map);
extern void *read_dtb_file(char *dtbfname);

static void usage(void)
{
/*./run -m master/buddy -f ./hello.out -d ./xlp.dtb" */
	fprintf(stderr, "\nUsage:\n\n");
	fprintf(stderr, "  ldapps -m <mode> -f <filename> -d <dtbcfgfile>\n\n");
	fprintf(stderr, "  where:\n");
	fprintf(stderr, "    <mode> is the running mode, it can be ampload/smpload\n");
	fprintf(stderr, "    <filename> is the application file name including the full path\n");
	fprintf(stderr, "    <dtbcfgfile> is the device tree config file name including the full path\n");

}


static unsigned int sys_cpus_map(void)
{
	char buf[BUFLEN];
	FILE *f;
	int len, i;
	unsigned int cpu_map = 0;

	strcpy(buf, "/sys/devices/system/cpu/");
	len = strlen(buf);
	for(i = 0; i < MAX_CPUS; i++) {
		sprintf(&buf[len], "cpu%d", i);
		if((f = fopen(buf, "r")) == NULL)
			continue;
		fclose(f);
		cpu_map |= (1 << i);
	}
	return cpu_map;
}


int main(int argc, char **argv)
{
	char *fname = NULL, *dtbfname = NULL, c;
	int mode = 0;
	char *dtb = NULL;
	unsigned int cpu_up_map = 0, cpu_possible_map;
	char buf[BUFLEN];
	int domain = 1;

	while((c = getopt(argc, argv, "m:f:d:i:h")) != -1)
	{
		switch(c) {
			case 'm':
				if(strcmp(optarg, "ampload") == 0)
					mode = MODE_APP_AMP;
				else if(strcmp(optarg, "smpload") == 0)
					mode = MODE_APP_SMP;
				else {
					printf("Invalid mode \n");
					usage();
					return -1;
				}
				break;
			case 'f':
				fname = optarg;
				break;
			case 'd':
				dtbfname = optarg;
				break;
		        case 'i':
				domain = strtoul(optarg, NULL, 10);
				break;
			default:
				usage();
				return -1;

		}
	}
	if(!fname || !dtbfname || !mode) {
		fprintf(stderr, "Invalid config\n");
		usage();
		return -1;
	}

	/* read dtb file and get 
           - cpu mask 
           - shared mem address and size */
	if((dtb = read_dtb_file(dtbfname)) == NULL)
		return -1;

	/* look up cpu online map from domain@0 */
	dom_id = 0;
	cpu_up_map = sys_cpus_map();
	if(fdt_get_cpu_online_map(dtb, &cpu_possible_map) < 0 || cpu_possible_map == 0)
		goto err_exit;

	if((cpu_up_map & cpu_possible_map) != cpu_possible_map) {
		fprintf(stderr, "cpu map spec in dtb and linux does not matching\n");
		goto err_exit;
	}
	printf("cpu_up_map=%x, cpu_possible_map=%x\n",cpu_up_map, cpu_possible_map);

	/* Set the environment variable */
	snprintf(buf, BUFLEN, "%s", mode == MODE_APP_AMP ? "amp" : "smp");
	if(setenv("ldmode", buf, 0) != 0) {
		fprintf(stderr, "setenv failed for mode\n");
		goto err_exit;
	}

	snprintf(buf, BUFLEN, "%x", cpu_possible_map);
	if(setenv("ldcpupossiblemap", buf, 0) != 0) {
		fprintf(stderr, "setenv failed for cpumap\n");
		goto err_exit;
	}

	if(setenv("ldfname", fname, 0) != 0) {
		fprintf(stderr, "setenv failed for fname\n");
		goto err_exit;
	}
	if(setenv("lddtbfname", dtbfname, 0) != 0) {
		fprintf(stderr, "setenv failed for dtbfname\n");
		goto err_exit;
	}

	snprintf(buf, BUFLEN, "%d", domain);
	if(setenv("lddomain", buf, 0) != 0) {
		fprintf(stderr, "setenv failed for domain id\n");
		goto err_exit;
	}

	free(dtb);


	if(setenv("LD_PRELOAD", "libhpexec.so", 0) != 0) {
		fprintf(stderr, "LD_PRELOAD failed for %s\n", buf);
		return -1;
	}

	/* run the application on me */
	execv(fname, argv);
	return 0;
err_exit:
	if(dtb)
		free(dtb);
	exit(1);
}

