/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sched.h>
#include <unistd.h>
#include <sys/types.h>
#include <dlfcn.h>
#include <syscall.h>
#include <errno.h>
#include <sys/shm.h>
#include <pthread.h>
#include <sys/syscall.h>
#include "spinlock.h"

#define MAX_CPUS 128

extern unsigned int cpu_online_map;

static unsigned int cpu_map = 0;
static int cpu_map_init = 0;
static int ldmode = 0;
static spinlock_t ptlock = { 0 };
static int cpu = 0;
struct wakeup_info
{
	pthread_t pth;
	int (*fn)(void *);
	void *arg;
};

static struct wakeup_info wakeup[MAX_CPUS];

enum {
	MODE_APP_AMP = 1,
	MODE_APP_SMP = 2
};


static int app_set_affinity(int cpu)
{
	cpu_set_t mask;
	int tid;

	CPU_ZERO(&mask);
	CPU_SET(cpu, &mask);
	tid = syscall(SYS_gettid);
	if(sched_setaffinity(tid, sizeof(mask), &mask) != 0) {
		fprintf(stderr, "Set affinity failed for cpu %d\n", cpu);
		return -1;
	}
	return 0;
}

static void *pthread_init(void *arg)
{
	int cpu = (int)(long)arg;
	app_set_affinity(cpu);
	if(wakeup[cpu].fn) {
		wakeup[cpu].pth = pthread_self();
		wakeup[cpu].fn(wakeup[cpu].arg);
	}
	return NULL;
}

static void ld_mode_init()
{
	char *buf;
	if(ldmode == 0) {
		if((buf = getenv("ldmode")) == NULL) {
			fprintf(stderr, "ldmode getenv failed\n");
			return;
		}

		if(strcmp(buf, "amp") == 0)
			ldmode = MODE_APP_AMP;
		else if(strcmp(buf, "smp") == 0)
			ldmode = MODE_APP_SMP;
	}
	return;
}

void exit(int status)
{
	pthread_t t;
	int i;
	void (*f)();

	ld_mode_init();
	if(ldmode == MODE_APP_SMP) {
		t = pthread_self();
		for(i = 0; i <= cpu; i++) {
			if(wakeup[i].pth && (wakeup[i].pth == t)) {
				pthread_exit(0);
			}

		}
		/* Parent wait for all childs to exit only if the exit is normal */
		if(status == 0) {
			for(i = 0; i <= cpu; i++) {
				if((t = wakeup[i].pth) != 0) {
					pthread_join(t, NULL);
					wakeup[i].pth = 0;
				}
			}
		}
	}

	f = dlsym(RTLD_NEXT, "exit");
	if(f != NULL) {
		f(status);
	}
	else {
		fprintf(stderr, "Unable to lookup \'exit\' symbol, aborting!\n");
	}

	abort();
}

pid_t clone(int (*fn)(void *), void *child_stack, int flags, void *arg, ...)
{
	int i;
	pthread_t child;

	ld_mode_init();
	if(ldmode != MODE_APP_SMP)
		return -1;
	if((flags & CLONE_VM) == 0)
		/* In buddy mode VM should be set as all the segments are shared. */
		return -1;

	spin_lock(&ptlock);

	if(!cpu_map_init) {

		cpu_map = cpu_online_map;
		cpu_map_init = 1;

		/* ignore the first cpu, assuming the application is running on that */
		for(i = 0; i < sizeof(cpu_map) * 8; i++) {
			if(cpu_map & (1 << i)) {
				cpu_map &= (~(1 << i));
				break;
			}
		}
	}

	if(cpu_map == 0) {
		spin_unlock(&ptlock);
		return -1;
	}

	for(i = 0; i < sizeof(cpu_map) * 8; i++) {
		if(cpu_map & (1 << i)) {
			cpu_map &= (~(1 << i));
			break;
		}
	}

	cpu++;
	spin_unlock(&ptlock);

	wakeup[cpu].fn = fn;
       	wakeup[cpu].arg = arg;

	if(pthread_create(&child, NULL, (void *)pthread_init, (void *)(long)cpu) != 0)
		return -1;
	wakeup[cpu].pth = child;

	sched_yield();
	return cpu;
}

pid_t getpid()
{
	pthread_t t;
	int i;
	cpu_set_t mask;
	int max_cpus;

	ld_mode_init();
	if(ldmode == MODE_APP_SMP) {
		t = pthread_self();
		for(i = 0; i <= cpu; i++) {
			if(wakeup[i].pth && (wakeup[i].pth == t))
				return i;
		}
	}

	CPU_ZERO(&mask);
	if(sched_getaffinity(0, sizeof(mask), &mask) != 0) {
		fprintf(stderr, "Get affinity failed \n");
		return -1;
	}
	max_cpus = (sizeof(mask) * 8) > MAX_CPUS ? MAX_CPUS : (sizeof(mask) * 8);
	for(i = 0; i < max_cpus; i++) {
		if(CPU_ISSET(i, &mask)) {
			return i;
		}
	}
	return -1;
}
