#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <sys/mman.h>
#include <sys/wait.h>

#include "nlm_soclib.h"
#include "fast_syscall.h"

extern void init_seconds_per_tick(void);
extern void perfctr_env_init(void);
extern void init_cached_proc_id(void);

static cpu_set_t launch_cpumask;

static int nlm_is_big_endian;
static int nlm_is_64bit_ins_enabled;

void launch_cpumask_env_init(void)
{
	const char *name = "NLM_LAUNCH_CPUMASK";
	const char *value = getenv(name);
	unsigned long mask;
	int i;

	CPU_ZERO(&launch_cpumask);

	if (!value) {
		fprintf(stderr, "[%s]: Unable to getenv(%s):%s\n", __FUNCTION__,
			name, strerror(errno));
		return;
	}

	mask = strtoul(value, 0, 16);
	printf("[%s]: Using 0x%08lx as Launch CPUMASK\n", __FUNCTION__, mask);

	for (i = 0; mask; ++i, mask >>= 1)
		if (mask & 0x1)
			CPU_SET(i, &launch_cpumask);
}

static __inline__ void atomic_inc(volatile unsigned int *vaddr)
{
	int value = 1;

	__asm__ __volatile__ (
		".set push\n"
		".set arch=xlp\n"
		"ldaddwu %0, %1\n"
		".set pop\n"
		:
		: "r"(value), "r"(vaddr) : "memory"
		);
}

void nlm_soclib_launch_cpus(void)
{
	int how_many;
	cpu_set_t cpumask;
	volatile unsigned int *mmap_addr = 0;
	int i=0;
	int pid = 0;

	how_many = CPU_COUNT(&launch_cpumask);
	if (how_many == 0) {
		return;
	}

	/* switch to master cpu */
	CPU_ZERO(&cpumask);
	CPU_SET(0, &cpumask);
	if (sched_setaffinity(0, sizeof(cpumask), &cpumask) < 0) {
		perror("sched_setaffinity");
		exit(-1);
	}

	mmap_addr = (volatile unsigned int *)mmap(0, 1024, PROT_READ|PROT_WRITE,
						  MAP_SHARED|MAP_ANONYMOUS, -1, 0);
	if ((void *)mmap_addr == MAP_FAILED) {
		fprintf(stderr, "mmap failed! %s\n", strerror(errno));
		return;
	}
	else {
		mmap_addr[0] = 0;
		printf("mmap_addr = %p, mmap_addr[0]=%d\n", mmap_addr, mmap_addr[0]);
	}

	fflush(stdout);
	fflush(stderr);

	for (i = 31; i >= 0; i--)
	{
		if (!CPU_ISSET(i, &launch_cpumask))
			continue;

		CPU_ZERO(&cpumask);
		CPU_SET(i, &cpumask);

		pid = fork();
		if (pid < 0) {
			fprintf(stderr, "failed to fork! %s\n", strerror(errno));
			return;
		}
		else if (pid == 0) {
			/* set the process group id */
			setpgid(0,0);

			/* child */
			if (sched_setaffinity(0, sizeof(cpumask), &cpumask) < 0) {
				perror("sched_setaffinity");
				exit(-1);
			}

			fflush(stdout);
			fflush(stderr);

			atomic_inc((volatile unsigned int *)mmap_addr);
			break;
		}
		else {
			/* parent */
			setpgid(pid,pid);
		}

	}
	if (i < 0) {
		/* parent */
		unsigned int loop_count = 0;

		while (mmap_addr[0] != how_many) {
			++loop_count;
                       if (loop_count == 1000000000) {
                               printf("ERROR: unable to launch children: giving up "
				      "waiting for the children\n");
                               break;
                        }
                       if ( (loop_count % 100000000) == 0) {
                               printf("DBG: waiting for children to catch up! loop_count=%d, "
				      "mmap_addr[0]=%d how_many=%d\n", loop_count, mmap_addr[0], how_many);
                        }
		}

		wait(NULL);
		/* main process that forked */
		exit(0);
	}

	/* poll for all siblings to catch up */
	while (mmap_addr[0] != how_many) {}
}

static inline void init_endianess(void)
{
	/* determine the endianess of the system */
	nlm_is_big_endian = (fs_is_big_endian() &&
			     !(fs_is_endian_reversed()));

	/* set permission to use 64bit instructions in user space */
	nlm_is_64bit_ins_enabled = fs_uspace_64bit_ins_enabled();

	return;
}

static inline void init_system_info(void)
{
	init_cached_proc_id();

	init_seconds_per_tick();

	init_endianess();
}

void nlm_soclib_init(void)
{
	/* initialize processor id */
	init_system_info();

	perfctr_env_init();

	launch_cpumask_env_init();
}

