/***********************************************************************
Copyright 2003-2010 Netlogic Microsystems (“Netlogic”). All rights
reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in
the documentation and/or other materials provided with the
distribution.
THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
*****************************#NETL_2#********************************/
#ifndef _NLM_CRYPTO_API_H_
#define _NLM_CRYPTO_API_H_

/**
* @file_name nlm_crypto_api.h
*/

/**
* @defgroup crypto Crypto API
* @brief Description about the crypto apis supported in NETOS.
*/

/**
* If user wish to allocate memory for the control and packet descriptors then following macros
* guides for the memory requirement of control and packet descriptors.
*/
#define NLM_MAX_CIPHER_KEY_LEN		(32)
#define NLM_MAX_AUTH_KEY_LEN		(128)
#define NLM_MAX_ARC4_STATE_SIZE		(264)
#define CRYPTO_SIZEOF_CNTRL_DESC	(8 + NLM_MAX_CIPHER_KEY_LEN + NLM_MAX_AUTH_KEY_LEN + NLM_MAX_ARC4_STATE_SIZE)

#define CRYPTO_SIZEOF_PKT_DESC(nr_frags)	(32 + (nr_frags * 2)*8 + 4 /* Designed freeback*/)

#define NLM_CRYPTO_FB_ERR_MASK 0x27bf80

#define NLM_CRYPTO_FB_DATA_ERR_FORCE_BYPASS 		   (1 << 21)
#define NLM_CRYPTO_FB_DATA_ERR_SW_DESC 			   (1 << 18)
#define NLM_CRYPTO_FB_DATA_ERR_INSUFFICIENT_DATA_TO_AUTH   (1 << 17)
#define NLM_CRYPTO_FB_DATA_ERR_INSUFFICIENT_DATA_TO_CIPHER (1 << 16)
#define NLM_CRYPTO_FB_DATA_ERR_NO_PKT_DESC 		   (1 << 15)
#define NLM_CRYPTO_FB_DATA_ERR_ECC_SRC_DST 		   (1 << 13)
#define NLM_CRYPTO_FB_DATA_ERR_ECC_DESC_0123          	   (1 << 12)
#define NLM_CRYPTO_FB_DATA_ERR_ECC_AUTH_KEY 		   (1 << 11)
#define NLM_CRYPTO_FB_DATA_ERR_ECC_CIPHER_KEY 		   (1 << 10)
#define NLM_CRYPTO_FB_DATA_ERR_ECC_CNTL_DESC 		   (1 << 9)
#define NLM_CRYPTO_FB_DATA_ERR_ECC_DATA 		   (1 << 8)
#define NLM_CRYPTO_FB_DATA_ERR_ECC_DESIGNER_FB 		   (1 << 7)

#define NLM_ERR_BASE				0xCAE0000
#define NLM_ERR_CIPHER_ALG_INVALID 		(NLM_ERR_BASE | 1)
#define NLM_ERR_CIPHER_ALG_MODE_INVALID 	(NLM_ERR_BASE | 2)
#define NLM_ERR_AUTH_ALG_INVALID 		(NLM_ERR_BASE | 3)
#define NLM_ERR_AUTH_ALG_MODE_INVALID 		(NLM_ERR_BASE | 4)
#define NLM_ERR_CIPHER_IV_LEN_INVALID 		(NLM_ERR_BASE | 5)
#define NLM_ERR_CIPHER_KEY_LEN_INVALID 		(NLM_ERR_BASE | 6)
#define NLM_ERR_CIPHER_KEY_NULL 		(NLM_ERR_BASE | 7)
#define NLM_ERR_AUTH_KEY_NULL 			(NLM_ERR_BASE | 8)
#define NLM_ERR_AUTH_DST_NULL 			(NLM_ERR_BASE | 9)
#define NLM_ERR_TAG_LEN 			(NLM_ERR_BASE | 10)
#define NLM_ERR_SESSION_NOT_SETUP 		(NLM_ERR_BASE | 11)
#define NLM_ERR_CRYPTO_PARAM_NULL 		(NLM_ERR_BASE | 12)
#define NLM_ERR_INVALID_PARAM 			(NLM_ERR_BASE | 13)
#define NLM_ERR_AUTH_LEN_LESS_THAN_CIPHER_LEN 	(NLM_ERR_BASE | 14)
#define NLM_ERR_CALLBACK_NULL 			(NLM_ERR_BASE | 15)
#define NLM_ERR_NO_MEM 				(NLM_ERR_BASE | 16)
#define NLM_ERR_UNSPECIFIED 			(NLM_ERR_BASE | 17)
#define	NLM_ERR_SEND_TIMEOUT 			(NLM_ERR_BASE | 18)
#define NLM_ERR_IN_PROGRESS 			(NLM_ERR_BASE | 19)
#define NLM_ERR_AUTH_KEY_LEN_INVALID 		(NLM_ERR_BASE | 20)

#define NLM_ERR_FREEBACK_NOT_RECEIVED 		(NLM_ERR_BASE | 256)

struct crypto_iovec
{
  unsigned char *buf;
  uint32_t iov_len;
};

struct crypto_cipher_init_param
{
  uint32_t cipher_alg;
  uint32_t cipher_mode;
  struct crypto_iovec *cipher_key;
  uint8_t arc4_cipher_key_len;
  uint8_t arc4_key_init;
  unsigned char cfb_mask;
};

struct crypto_auth_init_param
{
  uint32_t auth_alg;
  uint32_t auth_mode;
  struct crypto_iovec *auth_key;
  uint8_t hmac;
};

struct crypto_session
{
  void *cntrl_desc;
  void *pkt_desc;
  uint32_t flag;		/* This will indicate the owner of the memory */
  struct crypto_cipher_init_param *cip;
  struct crypto_auth_init_param *aip;

};


struct crypto_param
{
  uint32_t enc;
  struct crypto_iovec *src;	/* 1 or more fragments of the source data */
  struct crypto_iovec *dst;
  uint32_t src_len;
  uint32_t nr_frags;

  uint32_t cipher_len;
  uint32_t iv_len;
  unsigned char *iv;

  uint32_t iv_offset;
  uint32_t cipher_offset;
  uint8_t cipher_bit_count:3;

  uint8_t hash_bit_count:3;
  uint8_t tls_proto:1;
  uint8_t arc4_save_l3_alloc:1;
  uint8_t arc4_save_state:1;
  uint8_t arc4_load_state:1;
  uint8_t hmac_external_pad_key:1;
  uint8_t hash_clobber:1;
  uint8_t cipher_clobber:1;
  uint8_t hash_output_l3_alloc:1;
  uint8_t cipher_output_l3_alloc:1;
  uint8_t hash_source:1;
  uint8_t send_designer_fb:1;

  uint32_t auth_len;
  uint32_t auth_offset;
  uint32_t tag_len;

  uint32_t cipher_key_len;
  uint32_t auth_key_len;
  unsigned char *hash_dst_address;

  uint32_t designer_freeback_id;
  uint32_t designer_freeback_len;
  uint64_t designer_fb[4];
};

/**
 * @brief This API sets up session for performing cipher and authentication
 *         operation. It can be used for both cipher and authentication operation in one pass.
 * 	   It can also be used for either cipher or authentication.
 *
 * @param[in] session  User allocates memory for session context and passes
 *                     the structure pointer. It contains two memory pointer for control and packet descriptors.
 *                     If user wants to own the memory then flag parameter in the session structure needs to be set.
 *                     If flag is not set then driver will allocate the memory for control and packet descriptors.
		       It contains two structure to take input for cipher and auth operation separately.
 *
 * @return Return value of 0 indicates success.
 * Returns error code, in case of failure.
 *
 * @ingroup crypto
 *
 */

extern int crypto_setup_cipher_auth_session (struct crypto_session *session);

/**
 * @brief This API performs cipher and/or authentication operation. It will use preallocated session for algorithm type and mode of operation.
 *	  Authentication output will be written in the hash_dst_address and cipher output will be written in dst of the cprm structure.
 *        Source data can be fragmented and cprm->nr_frags represents number of fragments.
 * @param[in] session  Specifies the session which is setup by calling
 *                     crypto_setup_cipher_auth_session. 
 * @param[in] cprm     This structure contains input parameters for  cipher and authentication. 
 * 
 * @param[out] cprm->hash_dst_address For authentication output.
 * @param[out] cprm->dst For cipher output.
 *
 * @return Return value of 0 indicates success.
 * Returns error code, in case of failure.
 *
 * @ingroup crypto
 *
 */
extern int crypto_cipher_auth_op (struct crypto_session *session,
				  struct crypto_param *cprm);



/**
 * @brief This API resets key for an already configured session.
 *
 * @param[in] session    Specifies the session which is setup by calling
 *                       crypto_setup_cipher_auth_session.
 *
 * @param[in] cipher_key New cipher key. For auth only sessions, this parameter
 *                       is ignored.
 * @param[in] auth_key   New auth key. For cipher only sessions, this parameter
 *                       is ignored.
 *
 * @return Return value of 0 indicates success.
 * Returns error code, in case of failure.
 *
 * @ingroup crypto
 */

extern int crypto_session_reset_key (struct crypto_session *session,
				     unsigned char *cipher_key, unsigned
				     char *auth_key);

/**
 * @brief This API cleans up internal driver memory associated with a session.
 *
 * @param[in] session Specifies the session which is setup by calling
 *                     crypto_setup_cipher_auth_session. 
 *
 * @return Return value of 0 indicates success.
 * Returns error code, in case of failure.
 *
 * @ingroup crypto
 */
extern int crypto_cleanup_session (struct crypto_session *session);

extern int nlm_crypto_init (void);

#endif
