/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "spinlock.h"
#include "shmparam.h"
#include "contig_alloc.h"
#include "sys_entry.h"
#include "devices.h"

/**
* @file_name shmapi.c
*/

/**
* @defgroup shm Shared Memory API
* @brief Description about the shared memory apis supported in NETOS
*/

/**
* @defgroup mem Memory Management API
* @brief Description about the memory management apis supported in NETOS
*/


struct shm_vma {
	void *addr;
	size_t len;
	void *obj;
	struct shm_vma *next;
};

struct shm_vma_head {
	struct shm_vma *head;
	spinlock_t lock;
} shm_vma_head;

int shm_vma_add(void *addr, size_t len, void *obj)
{
	struct shm_vma *s, *t;
	if((s = calloc(1, sizeof(struct shm_vma))) == NULL)
		return -1;
	s->addr = addr;
	s->len = len;
	s->obj = obj;
	spin_lock(&shm_vma_head.lock);
	if(shm_vma_head.head == NULL) {
		shm_vma_head.head = s;
		s->next = NULL;
	} else {
		t = shm_vma_head.head;
		shm_vma_head.head = s;
		s->next = t;
	}
	spin_unlock(&shm_vma_head.lock);
	return 0;
}

void *shm_vma_remove(void *addr, size_t len) 
{
	struct shm_vma *cur;
	struct shm_vma **loc;
	void *obj = NULL;

	spin_lock(&shm_vma_head.lock);
        loc = &shm_vma_head.head;
        cur = *loc;
        while (cur != NULL) {
		if(cur->addr == addr && cur->len == len)
                        break;
                loc = &(cur->next);
                cur = *loc;
        }
        if (cur != NULL) {
                *loc = cur->next;
		obj = cur->obj;
        }
	spin_unlock(&shm_vma_head.lock);
	free(cur);
	return obj;
}
static inline void *dev_object_get(int fd)
{
	struct device_ops *dev;
	dev = dev_get(fd);
	if(dev)
		return dev->obj;
	return NULL;
}

int shm_truncate(struct _reent *reent, int fd, off_t length)
{
	struct shm_open_entry *shm;
	void *mem;

	if((shm = dev_object_get(fd)) == NULL) {
		printf("trunate failed fd %d\n", fd);
		return -1;
	}
	spin_lock(&shm->lock);
	if(shm->len >= length) {
		spin_unlock(&shm->lock);
		return 0;
	}

	mem = contig_realloc(shm->mem, length);
	if(mem == NULL) {
		reent->_errno = ENOSPC;
		spin_unlock(&shm->lock);
		return -1;
	}
	memset(mem + shm->len, 0, length - shm->len); 
	shm->mem = mem;
	shm->len = length;
	spin_unlock(&shm->lock);
	return 0;
};

void *shm_mmap(struct _reent *reent, int fd, void *addr, size_t len, int prot, int flags, off_t off)
{

	struct shm_open_entry *shm;
	void *raddr = (void *)(long)MAP_FAILED;

	if((shm = dev_object_get(fd)) == NULL)
		return MAP_FAILED;

	if(addr != NULL) {
		reent->_errno = EINVAL;
		return MAP_FAILED;
	}
	if(flags & MAP_FIXED || flags & MAP_PRIVATE) {
		reent->_errno = EINVAL;
		return MAP_FAILED;
	}

	spin_lock(&shm->lock);
	if((len + off) > shm->len) {
		reent->_errno = EINVAL;
		goto err_exit;
	}

	raddr  = shm->mem + off;
	if(shm_vma_add(raddr, len, shm) != 0) {
		reent->_errno = EINVAL;
		goto err_exit;
	}
	shm->refcnt++;

err_exit:
	spin_unlock(&shm->lock);
	return raddr;
};

int shm_munmap(struct _reent *reent, void *addr, size_t len)
{
	struct shm_open_entry *shm;
	struct netl_shmdestroy *d;
	struct netl_shmcall shmcall;

	shm = shm_vma_remove(addr, len);
	if(shm == NULL) {
		reent->_errno = EINVAL;
		return -1;
	}

	spin_lock(&shm->lock);
	shm->refcnt--;
	if(shm->unlink && shm->refcnt == 0) {
		/* give the lock before calling destroy */
		spin_unlock(&shm->lock);

		d = &shmcall.shmdestroy;
		d->obj = (long)shm;
		nlm_sys_call(NLM_SHM_SYS_CALL, (void *)&shmcall);
		reent->_errno = shmcall.err;
		return shmcall.rv;
	}

	spin_unlock(&shm->lock);
	return 0;
}

int shm_close(struct _reent *reent, int fd)
{
	struct shm_open_entry *shm;
	struct netl_shmdestroy *d;
	struct netl_shmcall shmcall;

	if((shm = dev_object_get(fd)) == NULL)
		return -1;

	spin_lock(&shm->lock);
	shm->refcnt--;
	if(shm->unlink && shm->refcnt == 0) {
		/* give the lock before calling destroy */
		spin_unlock(&shm->lock);

		d = &shmcall.shmdestroy;
		d->obj = (long)shm;
		nlm_sys_call(NLM_SHM_SYS_CALL, (void *)&shmcall);
		reent->_errno = shmcall.err;
		return shmcall.rv;
	}
	spin_unlock(&shm->lock);
	return 0;
}

/**
* @brief The shm_open function creates and opens a new, or opens an existing shared memory object
* @param[in] name Name of the shared memory object. For portable use, name 
*                  should have an initial slash(/) and contain no embeddes slashes
* @param[in] oflag Operation flags. Vald values are
*  - O_RDWR  : Memory object is created for read write access
*  - O_CREAT : Create the shared memory object if it does not exist.  
*  - O_EXCL  : Used along with O_CREAT. Returns error if the shared memory object already exist. 
*  - O_TRUNC : If the shared memory object already exists, truncate it to zero bytes. 
* @param[in] mode Permission to use. Valid values are
*  - S_IRUSR | S_IWUSR : User has read and write permissions.
*
* @return 
*  - On sucess,  returns a non-negative file descriptor. 
*  - On failure, returns -1. errno is set to indicate the cause of the error.
*	  
* @sa    shm_unlink
*
* Example usage:
* @code
*
* #define LEN 100
* struct mymem {
*    char buf[LEN];
*    int len;
* };
* 
* struct mymem *ptr;
* int fd, rv;
* 
* Create a shared memory object and set its size 
* fd = shm_open("/mymem", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
* if(fd < 0)
*    handle error 
* 
* Set the size  
* rv = ftruncate(fd, sizeof(struct mymem));
* if(rv < 0)
*    handle error
* 
* Map the shared memory object
* ptr = mmap(NULL, sizeof(struct mymem), PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
* if(ptr == MAP_FAILED)
*    handle error
*
* Now we can refer the mymem fields using ptr
* 
* 	
* 
* Once we have done with the mymem, free the shared object
* 
* Unmap the shared memory object
* rv = munmap(ptr, sizeof(struct mymem));
* if(rv < 0)
*     handle error 
* 
* Close the file descriptor 
* rv = close(fd);
* if(rv < 0)
*     handle error
* 
* Unlink the shared memory object 
* rv = shm_unlink("/mymem");
* if(rv < 0)
*     handle error 
* 
* @endcode
* @ingroup shm
*/
int shm_open(const char *name, int oflag, mode_t mode)
{
	struct netl_shmopen *s;
	struct netl_shmcall shmcall;
	struct device_ops *dev;
	struct _reent *reent = __getreent();
	int fd;

	s = &shmcall.shmopen;

	shmcall.cmd = NETL_SHMOPEN_CALL;
	strncpy(s->name, name, PATH_MAX);
	s->oflag = oflag;
	s->mode = mode; 
	nlm_sys_call(NLM_SHM_SYS_CALL, (void *)&shmcall);
	reent->_errno = shmcall.err;

	if(shmcall.rv != 0)
		return shmcall.rv;
	
	if((dev = malloc(sizeof(struct device_ops))) == NULL)
		return -1;

	if((dev->name = malloc(PATH_MAX)) == NULL)
		return -1;
	strncpy((char *) dev->name, name, PATH_MAX);

	dev->ftrunc = shm_truncate;
	dev->mmap = shm_mmap;
	dev->munmap = shm_munmap;
	dev->close_r = shm_close;
	dev->obj = (void *)(long)s->obj;
	if(dev->obj == NULL) 
		return -1;

	fd = fdget();
	if(dev_add(dev, fd) < 0) {
		free(dev);
		fdput(fd);
		return -1;
	}
	
	return fd;
}

/**
* @brief The shm_unlink function removes the shared memory object referenced by name
* @param[in] name Name of the shared memory object 
*
* @return 
*  - On sucess,  returns zero. 
*  - On failure, returns -1. errno is set to indicate the cause of the error.
*	  
* @sa    shm_open
* @ingroup shm
*/
int shm_unlink(const char *name)
{
	struct netl_shmunlink *s;
	struct netl_shmcall shmcall;
	struct _reent *reent = __getreent();
	s = &shmcall.shmunlink;

	shmcall.cmd = NETL_SHMUNLINK_CALL;
	strncpy(s->name, name, PATH_MAX);
	nlm_sys_call(NLM_SHM_SYS_CALL, (void *)&shmcall);
	reent->_errno = shmcall.err;
	return shmcall.rv;
}


