/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
/**
* @file_name timer_api.c
*/

/**
* @defgroup timer Timer API
* @brief Description about the timer apis supported in hyperexec
*/

#include <timer_api.h>
#include <irq.h>
#include <sys/signal.h>
#include <sys/time.h>
#define NR_TIMERS 1
#define NR_CPUS  32
struct netl_timer
{
    struct itimerval timer;
    spinlock_t lock;
    void (*func)(int);
    volatile int cpu_owner;
    volatile int enabled;
    uint32_t last_intr_cycles;
    uint64_t remaining_cycles;    /*interrupt after `cycles`*/
    int timer_updated;
};

extern unsigned int nlm_cpu_frequency;
#define CYCLES_PER_SECOND   (nlm_cpu_frequency)
#define CYCLES_PER_MICROSEC (CYCLES_PER_SECOND/1000000)

static struct netl_timer g_netl_timer[NR_TIMERS][NR_CPUS];
static struct netl_timer *netl_timer[NR_TIMERS][NR_CPUS];

static void reinit_timer(struct netl_timer *nl_timer)
{
    nl_timer->cpu_owner = -1;
    nl_timer->enabled = 0;
    nl_timer->timer_updated = 0;
    nl_timer->remaining_cycles = 0;
    nl_timer->last_intr_cycles = 0;
}

static uint64_t convert_sec_to_cycles(time_t sec, suseconds_t usec)
{
    uint64_t cycles;
    cycles = (uint64_t)(CYCLES_PER_MICROSEC*((uint64_t)usec)) +
                (uint64_t)(CYCLES_PER_SECOND*((uint64_t)sec));
    return cycles;
}

static void update_compare_reg(struct netl_timer *nl_timer, uint64_t cycles)
{
    nl_timer->remaining_cycles   = cycles;
    if(cycles > CYCLES_PER_SECOND){
        write_c0_compare(read_c0_count() + CYCLES_PER_SECOND);
        nl_timer->last_intr_cycles = CYCLES_PER_SECOND;
    }else{
        write_c0_compare(read_c0_count() + cycles);
        nl_timer->last_intr_cycles = cycles;
    }
}


static irqreturn_t netl_timer_handler(int32_t irq, struct pt_regs *regs)
{
    int32_t cpu = processor_id();
    struct netl_timer *nl_timer = netl_timer[ITIMER_REAL][cpu];
    uint64_t cycles = 0;
    struct itimerval *itimer = &nl_timer->timer;
    unsigned long flags = 0;
    uint64_t remaining_cycles = 0;
    
    spin_lock_irqsave(&nl_timer->lock, flags);
    if(nl_timer->timer_updated){
        cycles = convert_sec_to_cycles(itimer->it_value.tv_sec, 
                itimer->it_value.tv_usec);
        if(cycles){
            update_compare_reg(nl_timer, cycles);
        }else{
            /*disable eimr.*/
            write_phnx_eimr(read_phnx_eimr() & ~(1ULL << irq));
            reinit_timer(nl_timer);
        }
        nl_timer->timer_updated = 0;
        spin_unlock_irqrestore(&nl_timer->lock,flags);
        return 0;
    }
    spin_unlock_irqrestore(&nl_timer->lock,flags);
    
    nl_timer->remaining_cycles -= nl_timer->last_intr_cycles; 
    
    if(nl_timer->remaining_cycles > 0){
        remaining_cycles = nl_timer->remaining_cycles;
        if(remaining_cycles > CYCLES_PER_SECOND){
            write_c0_compare(read_c0_count() + CYCLES_PER_SECOND);
            nl_timer->last_intr_cycles = CYCLES_PER_SECOND;
        }else{
            write_c0_compare(read_c0_count() + remaining_cycles);
            nl_timer->last_intr_cycles = remaining_cycles;
        }
        return 0;
    }
    nl_timer->remaining_cycles = 0;
    nl_timer->last_intr_cycles = 0;

    /*reset the timer..*/
    itimer->it_value = itimer->it_interval;
    cycles = convert_sec_to_cycles(itimer->it_value.tv_sec, 
                itimer->it_value.tv_usec);
    
    if(cycles){
        if(nl_timer->func)
             nl_timer->func(SIGALRM);
        update_compare_reg(nl_timer, cycles);
    }else{
        /*disable eimr.*/
        write_phnx_eimr(read_phnx_eimr() & ~(1ULL << irq));
        spin_lock_irqsave(&nl_timer->lock, flags);
        reinit_timer(nl_timer);
        spin_unlock_irqrestore(&nl_timer->lock, flags);
        if(nl_timer->func)
             nl_timer->func(SIGALRM);
        cli();
    }
    return 0;
}

/**
* @brief Get value of the indexed timer.
*
* @param[in] index Timer index to read. Valid value is,
*  - ITIMER_REAL
*
* @param[in] value Value of the index timer is stored in the structure pointed
* by ovalue.
*
* @return 
*  - On sucess,  returns 0. 
*  - On failure, returns -1. 
*	  
* @sa    setitimer, signal 
*
* @ingroup timer
*/
int getitimer(int32_t index, struct itimerval *value)
{
    int32_t cpu = processor_id();
    struct netl_timer *nl_timer = netl_timer[ITIMER_REAL][cpu];
    unsigned long flags;

    if(index != ITIMER_REAL)
        return -1;

    spin_lock_irqsave(&nl_timer->lock, flags);
    /*copy `index` timer to `value`.*/
    *value = nl_timer->timer;
    spin_unlock_irqrestore(&nl_timer->lock, flags); 
    return 0;
}

/**
* @brief Set value of interval timer.
*
* @param[in] index Timer index to be used. Valid value is,
*  - ITIMER_REAL : A SIGALRM siganl is deliverd when this timer expires.
*
* @param[in] value Value of the timer.

* @param[in] ovalue If ovalue is not a null pointer, previous value of the timer
* is stored in the structure pointed to by ovalue.
*
* @return 
*  - On sucess,  returns 0. 
*  - On failure, returns -1. 
*	  
* @sa    getitimer, signal 
*
* @ingroup timer
*/

int setitimer(int32_t index, const struct itimerval *value,
                         struct itimerval *ovalue)
{
    int32_t cpu = processor_id();
    struct netl_timer *nl_timer = netl_timer[ITIMER_REAL][cpu];
    unsigned long flags;
    uint64_t cycles;
    struct itimerval *itimer = &nl_timer->timer;
    int irq = IRQ_TIMER;

    if(index != ITIMER_REAL)
        return -1;

    spin_lock_irqsave(&nl_timer->lock, flags);
    
    if(ovalue){
        /*copy old value.*/
        *ovalue = nl_timer->timer;
    }

    *itimer = *value;
    
    if(nl_timer->enabled && nl_timer->cpu_owner != cpu){
        /*master/buddy mode.*/
        /*interrupts are already enabled on other cpu.*/
        nl_timer->timer_updated = 1;
        spin_unlock_irqrestore(&nl_timer->lock,flags);
        return 0;
    }

    cycles = convert_sec_to_cycles(itimer->it_value.tv_sec, 
                            itimer->it_value.tv_usec);

    if(!cycles){
        /*Disable eimr*/
        reinit_timer(nl_timer);
        write_phnx_eimr(read_phnx_eimr() & ~(1ULL << irq));
        spin_unlock_irqrestore(&nl_timer->lock,flags);
        return 0;
    }

    update_compare_reg(nl_timer, cycles);

    if(!nl_timer->enabled){
        request_irq(irq, netl_timer_handler, EMPTY_IRQ_FLAGS,1 << cpu);
    }

    nl_timer->cpu_owner = cpu;
    nl_timer->enabled = 1;
    spin_unlock_irqrestore(&nl_timer->lock,flags);
    sti();
    return 0;
}

/**
* @brief Set a callback function for the specified signal index.
*
* @param[in] sig Signal index. Valid value is,
*  - SIGALRM
*
* @param[in] func Call back function pointer.
*
* @return 
*  - returns func. 
*	  
* @ingroup timer
*/
_sig_func_ptr signal(int sig, _sig_func_ptr func)
{
    int32_t cpu = processor_id();
    switch(sig)
    {
        case SIGALRM:
            netl_timer[ITIMER_REAL][cpu]->func = func;
            break;
        default:
            break;
    }

	return func;
}

void init_timers(void)
{
    int32_t i = 0, j = 0;
    int32_t cpu = processor_id();
    extern int32_t master_cpu;
    extern uint32_t buddy_cpu_mask;

    if(buddy_cpu_mask){
         /*Master Buddy mode use only one timerval structure.*/
        for(i=0; i<NR_TIMERS; i++){
            for(j=0; j<NR_CPUS; j++){
                netl_timer[i][j] = &g_netl_timer[i][master_cpu];
                netl_timer[i][j]->cpu_owner = -1;
            }
        }
    }else{
        for(i=0; i<NR_TIMERS; i++){
                netl_timer[i][cpu] = &g_netl_timer[i][cpu];
                netl_timer[i][cpu]->cpu_owner = -1;
        }
    }
}
